@@ -20,12 +20,15 @@
 package client_test
 
 import (
+	"encoding/json"
 	"errors"
 	"fmt"
+	"io"
 	"io/ioutil"
 	"net"
 	"net/http"
 	"net/http/httptest"
+	"net/url"
 	"os"
 	"path/filepath"
 	"strings"
@@ -36,27 +39,35 @@ import (
 
 	"github.com/snapcore/snapd/client"
 	"github.com/snapcore/snapd/dirs"
+	"github.com/snapcore/snapd/testutil"
 )
 
 // Hook up check.v1 into the "go test" runner
 func Test(t *testing.T) { TestingT(t) }
 
 type clientSuite struct {
-	cli     *client.Client
-	req     *http.Request
-	reqs    []*http.Request
-	rsp     string
-	rsps    []string
-	err     error
-	doCalls int
-	header  http.Header
-	status  int
+	testutil.BaseTest
+
+	cli           *client.Client
+	req           *http.Request
+	reqs          []*http.Request
+	rsp           string
+	rsps          []string
+	err           error
+	doCalls       int
+	header        http.Header
+	status        int
+	contentLength int64
+
+	countingCloser *countingCloser
 }
 
 var _ = Suite(&clientSuite{})
 
 func (cs *clientSuite) SetUpTest(c *C) {
 	os.Setenv(client.TestAuthFileEnvKey, filepath.Join(c.MkDir(), "auth.json"))
+	cs.AddCleanup(func() { os.Unsetenv(client.TestAuthFileEnvKey) })
+
 	cs.cli = client.New(nil)
 	cs.cli.SetDoer(cs)
 	cs.err = nil
@@ -68,12 +79,23 @@ func (cs *clientSuite) SetUpTest(c *C) {
 	cs.header = nil
 	cs.status = 200
 	cs.doCalls = 0
+	cs.contentLength = 0
+	cs.countingCloser = nil
 
 	dirs.SetRootDir(c.MkDir())
+	cs.AddCleanup(func() { dirs.SetRootDir("") })
+
+	cs.AddCleanup(client.MockDoTimings(time.Millisecond, 100*time.Millisecond))
 }
 
-func (cs *clientSuite) TearDownTest(c *C) {
-	os.Unsetenv(client.TestAuthFileEnvKey)
+type countingCloser struct {
+	io.Reader
+	closeCalled int
+}
+
+func (n *countingCloser) Close() error {
+	n.closeCalled++
+	return nil
 }
 
 func (cs *clientSuite) Do(req *http.Request) (*http.Response, error) {
@@ -83,10 +105,12 @@ func (cs *clientSuite) Do(req *http.Request) (*http.Response, error) {
 	if cs.doCalls < len(cs.rsps) {
 		body = cs.rsps[cs.doCalls]
 	}
+	cs.countingCloser = &countingCloser{Reader: strings.NewReader(body)}
 	rsp := &http.Response{
-		Body:       ioutil.NopCloser(strings.NewReader(body)),
-		Header:     cs.header,
-		StatusCode: cs.status,
+		Body:          cs.countingCloser,
+		Header:        cs.header,
+		StatusCode:    cs.status,
+		ContentLength: cs.contentLength,
 	}
 	cs.doCalls++
 	return rsp, cs.err
@@ -95,14 +119,12 @@ func (cs *clientSuite) Do(req *http.Request) (*http.Response, error) {
 func (cs *clientSuite) TestNewPanics(c *C) {
 	c.Assert(func() {
 		client.New(&client.Config{BaseURL: ":"})
-	}, PanicMatches, `cannot parse server base URL: ":" \(parse :: missing protocol scheme\)`)
+	}, PanicMatches, `cannot parse server base URL: ":" \(parse \"?:\"?: missing protocol scheme\)`)
 }
 
 func (cs *clientSuite) TestClientDoReportsErrors(c *C) {
-	restore := client.MockDoRetry(10*time.Millisecond, 100*time.Millisecond)
-	defer restore()
 	cs.err = errors.New("ouchie")
-	err := cs.cli.Do("GET", "/", nil, nil, nil)
+	_, err := cs.cli.Do("GET", "/", nil, nil, nil, nil)
 	c.Check(err, ErrorMatches, "cannot communicate with server: ouchie")
 	if cs.doCalls < 2 {
 		c.Fatalf("do did not retry")
@@ -113,8 +135,104 @@ func (cs *clientSuite) TestClientWorks(c *C) {
 	var v []int
 	cs.rsp = `[1,2]`
 	reqBody := ioutil.NopCloser(strings.NewReader(""))
-	err := cs.cli.Do("GET", "/this", nil, reqBody, &v)
+	statusCode, err := cs.cli.Do("GET", "/this", nil, reqBody, &v, nil)
 	c.Check(err, IsNil)
+	c.Check(statusCode, Equals, 200)
+	c.Check(v, DeepEquals, []int{1, 2})
+	c.Assert(cs.req, NotNil)
+	c.Assert(cs.req.URL, NotNil)
+	c.Check(cs.req.Method, Equals, "GET")
+	c.Check(cs.req.Body, Equals, reqBody)
+	c.Check(cs.req.URL.Path, Equals, "/this")
+}
+
+func makeMaintenanceFile(c *C, b []byte) {
+	c.Assert(os.MkdirAll(filepath.Dir(dirs.SnapdMaintenanceFile), 0755), IsNil)
+	c.Assert(ioutil.WriteFile(dirs.SnapdMaintenanceFile, b, 0644), IsNil)
+}
+
+func (cs *clientSuite) TestClientSetMaintenanceForMaintenanceJSON(c *C) {
+	// write a maintenance.json that says snapd is down for a restart
+	maintErr := &client.Error{
+		Kind:    client.ErrorKindSystemRestart,
+		Message: "system is restarting",
+	}
+	b, err := json.Marshal(maintErr)
+	c.Assert(err, IsNil)
+	makeMaintenanceFile(c, b)
+
+	// now after a Do(), we will have maintenance set to what we wrote
+	// originally
+	_, err = cs.cli.Do("GET", "/this", nil, nil, nil, nil)
+	c.Check(err, IsNil)
+
+	returnedErr := cs.cli.Maintenance()
+	c.Assert(returnedErr, DeepEquals, maintErr)
+}
+
+func (cs *clientSuite) TestClientIgnoresGarbageMaintenanceJSON(c *C) {
+	// write a garbage maintenance.json that can't be unmarshalled
+	makeMaintenanceFile(c, []byte("blah blah blah not json"))
+
+	// after a Do(), no maintenance set and also no error returned from Do()
+	_, err := cs.cli.Do("GET", "/this", nil, nil, nil, nil)
+	c.Check(err, IsNil)
+
+	returnedErr := cs.cli.Maintenance()
+	c.Assert(returnedErr, IsNil)
+}
+
+func (cs *clientSuite) TestClientDoNoTimeoutIgnoresRetry(c *C) {
+	var v []int
+	cs.rsp = `[1,2]`
+	cs.err = fmt.Errorf("borken")
+	reqBody := ioutil.NopCloser(strings.NewReader(""))
+	doOpts := &client.DoOptions{
+		// Timeout is unset, thus 0, and thus we ignore the retry and only run
+		// once even though there is an error
+		Retry: time.Duration(time.Second),
+	}
+	_, err := cs.cli.Do("GET", "/this", nil, reqBody, &v, doOpts)
+	c.Check(err, ErrorMatches, "cannot communicate with server: borken")
+	c.Assert(cs.doCalls, Equals, 1)
+}
+
+func (cs *clientSuite) TestClientDoRetryValidation(c *C) {
+	var v []int
+	cs.rsp = `[1,2]`
+	reqBody := ioutil.NopCloser(strings.NewReader(""))
+	doOpts := &client.DoOptions{
+		Retry:   time.Duration(-1),
+		Timeout: time.Duration(time.Minute),
+	}
+	_, err := cs.cli.Do("GET", "/this", nil, reqBody, &v, doOpts)
+	c.Check(err, ErrorMatches, "internal error: retry setting.*invalid")
+	c.Assert(cs.req, IsNil)
+}
+
+func (cs *clientSuite) TestClientDoRetryWorks(c *C) {
+	reqBody := ioutil.NopCloser(strings.NewReader(""))
+	cs.err = fmt.Errorf("borken")
+	doOpts := &client.DoOptions{
+		Retry:   time.Duration(time.Millisecond),
+		Timeout: time.Duration(time.Second),
+	}
+	_, err := cs.cli.Do("GET", "/this", nil, reqBody, nil, doOpts)
+	c.Check(err, ErrorMatches, "cannot communicate with server: borken")
+	// best effort checking given that execution could be slow
+	// on some machines
+	c.Assert(cs.doCalls > 100, Equals, true, Commentf("got only %v calls", cs.doCalls))
+	c.Assert(cs.doCalls < 1100, Equals, true, Commentf("got %v calls", cs.doCalls))
+}
+
+func (cs *clientSuite) TestClientUnderstandsStatusCode(c *C) {
+	var v []int
+	cs.status = 202
+	cs.rsp = `[1,2]`
+	reqBody := ioutil.NopCloser(strings.NewReader(""))
+	statusCode, err := cs.cli.Do("GET", "/this", nil, reqBody, &v, nil)
+	c.Check(err, IsNil)
+	c.Check(statusCode, Equals, 202)
 	c.Check(v, DeepEquals, []int{1, 2})
 	c.Assert(cs.req, NotNil)
 	c.Assert(cs.req.URL, NotNil)
@@ -128,7 +246,7 @@ func (cs *clientSuite) TestClientDefaultsToNoAuthorization(c *C) {
 	defer os.Unsetenv(client.TestAuthFileEnvKey)
 
 	var v string
-	_ = cs.cli.Do("GET", "/this", nil, nil, &v)
+	_, _ = cs.cli.Do("GET", "/this", nil, nil, &v, nil)
 	c.Assert(cs.req, NotNil)
 	authorization := cs.req.Header.Get("Authorization")
 	c.Check(authorization, Equals, "")
@@ -146,7 +264,7 @@ func (cs *clientSuite) TestClientSetsAuthorization(c *C) {
 	c.Assert(err, IsNil)
 
 	var v string
-	_ = cs.cli.Do("GET", "/this", nil, nil, &v)
+	_, _ = cs.cli.Do("GET", "/this", nil, nil, &v, nil)
 	authorization := cs.req.Header.Get("Authorization")
 	c.Check(authorization, Equals, `Macaroon root="macaroon", discharge="discharge"`)
 }
@@ -165,7 +283,7 @@ func (cs *clientSuite) TestClientHonorsDisableAuth(c *C) {
 	var v string
 	cli := client.New(&client.Config{DisableAuth: true})
 	cli.SetDoer(cs)
-	_ = cli.Do("GET", "/this", nil, nil, &v)
+	_, _ = cli.Do("GET", "/this", nil, nil, &v, nil)
 	authorization := cs.req.Header.Get("Authorization")
 	c.Check(authorization, Equals, "")
 }
@@ -174,13 +292,13 @@ func (cs *clientSuite) TestClientHonorsInteractive(c *C) {
 	var v string
 	cli := client.New(&client.Config{Interactive: false})
 	cli.SetDoer(cs)
-	_ = cli.Do("GET", "/this", nil, nil, &v)
+	_, _ = cli.Do("GET", "/this", nil, nil, &v, nil)
 	interactive := cs.req.Header.Get(client.AllowInteractionHeader)
 	c.Check(interactive, Equals, "")
 
 	cli = client.New(&client.Config{Interactive: true})
 	cli.SetDoer(cs)
-	_ = cli.Do("GET", "/this", nil, nil, &v)
+	_, _ = cli.Do("GET", "/this", nil, nil, &v, nil)
 	interactive = cs.req.Header.Get(client.AllowInteractionHeader)
 	c.Check(interactive, Equals, "true")
 }
@@ -218,6 +336,8 @@ func (cs *clientSuite) TestClientSysInfo(c *C) {
                       "on-classic": true,
                       "build-id": "1234",
                       "confinement": "strict",
+                      "architecture": "TI-99/4A",
+                      "virtualization": "MESS",
                       "sandbox-features": {"backend": ["feature-1", "feature-2"]}}}`
 	sysInfo, err := cs.cli.SysInfo()
 	c.Check(err, IsNil)
@@ -233,7 +353,9 @@ func (cs *clientSuite) TestClientSysInfo(c *C) {
 		SandboxFeatures: map[string][]string{
 			"backend": {"feature-1", "feature-2"},
 		},
-		BuildID: "1234",
+		BuildID:        "1234",
+		Architecture:   "TI-99/4A",
+		Virtualization: "MESS",
 	})
 }
 
@@ -241,14 +363,19 @@ func (cs *clientSuite) TestServerVersion(c *C) {
 	cs.rsp = `{"type": "sync", "result":
                      {"series": "16",
                       "version": "2",
-                      "os-release": {"id": "zyggy", "version-id": "123"}}}`
+                      "os-release": {"id": "zyggy", "version-id": "123"},
+                      "architecture": "m32",
+                      "virtualization": "qemu"
+}}}`
 	version, err := cs.cli.ServerVersion()
 	c.Check(err, IsNil)
 	c.Check(version, DeepEquals, &client.ServerVersion{
-		Version:     "2",
-		Series:      "16",
-		OSID:        "zyggy",
-		OSVersionID: "123",
+		Version:        "2",
+		Series:         "16",
+		OSID:           "zyggy",
+		OSVersionID:    "123",
+		Architecture:   "m32",
+		Virtualization: "qemu",
 	})
 }
 
@@ -275,7 +402,7 @@ func (cs *clientSuite) TestSnapdClientIntegration(c *C) {
 
 	cli := client.New(nil)
 	si, err := cli.SysInfo()
-	c.Check(err, IsNil)
+	c.Assert(err, IsNil)
 	c.Check(si.Series, Equals, "42")
 }
 
@@ -308,19 +435,21 @@ func (cs *clientSuite) TestSnapClientIntegration(c *C) {
 		Args:      []string{"bar", "--baz"},
 	}
 
-	stdout, stderr, err := cli.RunSnapctl(options)
+	stdout, stderr, err := cli.RunSnapctl(options, nil)
 	c.Check(err, IsNil)
 	c.Check(string(stdout), Equals, "test stdout")
 	c.Check(string(stderr), Equals, "test stderr")
 }
 
 func (cs *clientSuite) TestClientReportsOpError(c *C) {
-	cs.rsp = `{"type": "error", "status": "potatoes"}`
+	cs.status = 500
+	cs.rsp = `{"type": "error"}`
 	_, err := cs.cli.SysInfo()
-	c.Check(err, ErrorMatches, `.*server error: "potatoes"`)
+	c.Check(err, ErrorMatches, `.*server error: "Internal Server Error"`)
 }
 
 func (cs *clientSuite) TestClientReportsOpErrorStr(c *C) {
+	cs.status = 400
 	cs.rsp = `{
 		"result": {},
 		"status": "Bad Request",
@@ -365,6 +494,7 @@ func (cs *clientSuite) TestClientMaintenance(c *C) {
 }
 
 func (cs *clientSuite) TestClientAsyncOpMaintenance(c *C) {
+	cs.status = 202
 	cs.rsp = `{"type":"async", "status-code": 202, "change": "42", "maintenance": {"kind": "system-restart", "message": "system is restarting"}}`
 	_, err := cs.cli.Install("foo", nil)
 	c.Assert(err, IsNil)
@@ -420,105 +550,23 @@ func (cs *clientSuite) TestIsTwoFactor(c *C) {
 	c.Check(client.IsTwoFactorError((*client.Error)(nil)), Equals, false)
 }
 
-func (cs *clientSuite) TestClientCreateUser(c *C) {
-	_, err := cs.cli.CreateUser(&client.CreateUserOptions{})
-	c.Assert(err, ErrorMatches, "cannot create a user without providing an email")
-
-	cs.rsp = `{
-		"type": "sync",
-		"result": {
-                        "username": "karl",
-                        "ssh-keys": ["one", "two"]
-		}
-	}`
-	rsp, err := cs.cli.CreateUser(&client.CreateUserOptions{Email: "one@email.com", Sudoer: true, Known: true})
-	c.Assert(cs.req.Method, Equals, "POST")
-	c.Assert(cs.req.URL.Path, Equals, "/v2/create-user")
-	c.Assert(err, IsNil)
-
-	body, err := ioutil.ReadAll(cs.req.Body)
-	c.Assert(err, IsNil)
-	c.Assert(string(body), Equals, `{"email":"one@email.com","sudoer":true,"known":true}`)
-
-	c.Assert(rsp, DeepEquals, &client.CreateUserResult{
-		Username: "karl",
-		SSHKeys:  []string{"one", "two"},
-	})
+func (cs *clientSuite) TestIsRetryable(c *C) {
+	// unhappy
+	c.Check(client.IsRetryable(nil), Equals, false)
+	c.Check(client.IsRetryable(errors.New("some-error")), Equals, false)
+	c.Check(client.IsRetryable(&client.Error{Kind: "something-else"}), Equals, false)
+	// happy
+	c.Check(client.IsRetryable(&client.Error{Kind: client.ErrorKindSnapChangeConflict}), Equals, true)
 }
 
-var createUsersTests = []struct {
-	options   []*client.CreateUserOptions
-	bodies    []string
-	responses []string
-	results   []*client.CreateUserResult
-	error     string
-}{{
-	options: []*client.CreateUserOptions{{}},
-	error:   "cannot create user from store details without an email to query for",
-}, {
-	options: []*client.CreateUserOptions{{
-		Email:  "one@example.com",
-		Sudoer: true,
-	}, {
-		Known: true,
-	}},
-	bodies: []string{
-		`{"email":"one@example.com","sudoer":true}`,
-		`{"known":true}`,
-	},
-	responses: []string{
-		`{"type": "sync", "result": {"username": "one", "ssh-keys":["a", "b"]}}`,
-		`{"type": "sync", "result": [{"username": "two"}, {"username": "three"}]}`,
-	},
-	results: []*client.CreateUserResult{{
-		Username: "one",
-		SSHKeys:  []string{"a", "b"},
-	}, {
-		Username: "two",
-	}, {
-		Username: "three",
-	}},
-}}
+func (cs *clientSuite) TestUserAgent(c *C) {
+	cli := client.New(&client.Config{UserAgent: "some-agent/9.87"})
+	cli.SetDoer(cs)
 
-func (cs *clientSuite) TestClientCreateUsers(c *C) {
-	for _, test := range createUsersTests {
-		cs.rsps = test.responses
-
-		results, err := cs.cli.CreateUsers(test.options)
-		if test.error != "" {
-			c.Assert(err, ErrorMatches, test.error)
-		}
-		c.Assert(results, DeepEquals, test.results)
-
-		var bodies []string
-		for _, req := range cs.reqs {
-			c.Assert(req.Method, Equals, "POST")
-			c.Assert(req.URL.Path, Equals, "/v2/create-user")
-			data, err := ioutil.ReadAll(req.Body)
-			c.Assert(err, IsNil)
-			bodies = append(bodies, string(data))
-		}
-
-		c.Assert(bodies, DeepEquals, test.bodies)
-	}
-}
-
-func (cs *clientSuite) TestClientJSONError(c *C) {
-	cs.rsp = `some non-json error message`
-	_, err := cs.cli.SysInfo()
-	c.Assert(err, ErrorMatches, `cannot obtain system details: cannot decode "some non-json error message": invalid char.*`)
-}
-
-func (cs *clientSuite) TestUsers(c *C) {
-	cs.rsp = `{"type": "sync", "result":
-                     [{"username": "foo","email":"foo@example.com"},
-                      {"username": "bar","email":"bar@example.com"}]}`
-	users, err := cs.cli.Users()
-	c.Check(err, IsNil)
-	c.Check(users, DeepEquals, []*client.User{
-		{Username: "foo", Email: "foo@example.com"},
-		{Username: "bar", Email: "bar@example.com"},
-	})
+	var v string
+	_, _ = cli.Do("GET", "/", nil, nil, &v, nil)
+	c.Assert(cs.req, NotNil)
+	c.Check(cs.req.Header.Get("User-Agent"), Equals, "some-agent/9.87")
 }
 
 func (cs *clientSuite) TestDebugEnsureStateSoon(c *C) {
@@ -547,3 +595,49 @@ func (cs *clientSuite) TestDebugGeneric(c *C) {
 	c.Assert(err, IsNil)
 	c.Check(string(data), DeepEquals, `{"action":"do-something","params":["param1","param2"]}`)
 }
+
+func (cs *clientSuite) TestDebugGet(c *C) {
+	cs.rsp = `{"type": "sync", "result":["res1","res2"]}`
+
+	var result []string
+	err := cs.cli.DebugGet("do-something", &result, map[string]string{"foo": "bar"})
+	c.Check(err, IsNil)
+	c.Check(result, DeepEquals, []string{"res1", "res2"})
+	c.Check(cs.reqs, HasLen, 1)
+	c.Check(cs.reqs[0].Method, Equals, "GET")
+	c.Check(cs.reqs[0].URL.Path, Equals, "/v2/debug")
+	c.Check(cs.reqs[0].URL.Query(), DeepEquals, url.Values{"aspect": []string{"do-something"}, "foo": []string{"bar"}})
+}
+
+type integrationSuite struct{}
+
+var _ = Suite(&integrationSuite{})
+
+func (cs *integrationSuite) TestClientTimeoutLP1837804(c *C) {
+	restore := client.MockDoTimings(time.Millisecond, 5*time.Millisecond)
+	defer restore()
+
+	testServer := httptest.NewServer(http.HandlerFunc(func(res http.ResponseWriter, req *http.Request) {
+		time.Sleep(25 * time.Millisecond)
+	}))
+	defer func() { testServer.Close() }()
+
+	cli := client.New(&client.Config{BaseURL: testServer.URL})
+	_, err := cli.Do("GET", "/", nil, nil, nil, nil)
+	c.Assert(err, ErrorMatches, `.* timeout exceeded while waiting for response`)
+
+	_, err = cli.Do("POST", "/", nil, nil, nil, nil)
+	c.Assert(err, ErrorMatches, `.* timeout exceeded while waiting for response`)
+}
+
+func (cs *clientSuite) TestClientSystemRecoveryKeys(c *C) {
+	cs.rsp = `{"type":"sync", "result":{"recovery-key":"42"}}`
+
+	var key client.SystemRecoveryKeysResponse
+	err := cs.cli.SystemRecoveryKeys(&key)
+	c.Assert(err, IsNil)
+	c.Check(cs.reqs, HasLen, 1)
+	c.Check(cs.reqs[0].Method, Equals, "GET")
+	c.Check(cs.reqs[0].URL.Path, Equals, "/v2/system-recovery-keys")
+	c.Check(key.RecoveryKey, Equals, "42")
+}
