@@ -14,7 +14,10 @@
  * along with this program.  If not, see <http://www.gnu.org/licenses/>.
  *
  */
+#ifdef HAVE_CONFIG_H
 #include "config.h"
+#endif
+
 #include "mount-support.h"
 
 #include <errno.h>
@@ -30,83 +33,103 @@
 #include <sys/stat.h>
 #include <sys/syscall.h>
 #include <sys/types.h>
-#include <sys/types.h>
 #include <sys/wait.h>
 #include <unistd.h>
-#include <libgen.h>
 
+#include "../libsnap-confine-private/apparmor-support.h"
 #include "../libsnap-confine-private/classic.h"
 #include "../libsnap-confine-private/cleanup-funcs.h"
 #include "../libsnap-confine-private/mount-opt.h"
 #include "../libsnap-confine-private/mountinfo.h"
 #include "../libsnap-confine-private/snap.h"
 #include "../libsnap-confine-private/string-utils.h"
+#include "../libsnap-confine-private/tool.h"
 #include "../libsnap-confine-private/utils.h"
 #include "mount-support-nvidia.h"
-#include "apparmor-support.h"
-#include "quirks.h"
 
 #define MAX_BUF 1000
 
-/*!
- * The void directory.
- *
- * Snap confine moves to that directory in case it cannot retain the current
- * working directory across the pivot_root call.
- **/
-#define SC_VOID_DIR "/var/lib/snapd/void"
+static void sc_detach_views_of_writable(sc_distro distro, bool normal_mode);
 
 // TODO: simplify this, after all it is just a tmpfs
 // TODO: fold this into bootstrap
 static void setup_private_mount(const char *snap_name)
 {
-	uid_t uid = getuid();
-	gid_t gid = getgid();
-	char tmpdir[MAX_BUF] = { 0 };
-
-	// Create a 0700 base directory, this is the base dir that is
-	// protected from other users.
+	// Create a 0700 base directory. This is the "base" directory that is
+	// protected from other users. This directory name is NOT randomly
+	// generated. This has several properties:
 	//
-	// Under that basedir, we put a 1777 /tmp dir that is then bind
-	// mounted for the applications to use
-	sc_must_snprintf(tmpdir, sizeof(tmpdir), "/tmp/snap.%d_%s_XXXXXX", uid,
-			 snap_name);
-	if (mkdtemp(tmpdir) == NULL) {
-		die("cannot create temporary directory essential for private /tmp");
-	}
-	// now we create a 1777 /tmp inside our private dir
-	mode_t old_mask = umask(0);
-	char *d = strdup(tmpdir);
-	if (!d) {
-		die("cannot allocate memory for string copy");
-	}
-	sc_must_snprintf(tmpdir, sizeof(tmpdir), "%s/tmp", d);
-	free(d);
+	// Users can relate to the name and can find the temporary directory as
+	// visible from within the snap. If this directory was random it would be
+	// harder to find because there may be situations in which multiple
+	// directories related to the same snap name would exist.
+	//
+	// Snapd can partially manage the directory. Specifically on snap remove
+	// snapd could remove the directory and everything in it, potentially
+	// avoiding runaway disk use on a machine that either never reboots or uses
+	// persistent /tmp directory.
+	//
+	// Underneath the base directory there is a "tmp" sub-directory that has
+	// mode 1777 and behaves as a typical /tmp directory would. That directory
+	// is used as a bind-mounted /tmp directory.
+	//
+	// Because the directories are reused across invocations by distinct users
+	// and because the directories are trivially guessable, each invocation
+	// unconditionally chowns/chmods them to appropriate values.
+	char base_dir[MAX_BUF] = { 0 };
+	char tmp_dir[MAX_BUF] = { 0 };
+	int base_dir_fd SC_CLEANUP(sc_cleanup_close) = -1;
+	int tmp_dir_fd SC_CLEANUP(sc_cleanup_close) = -1;
+	sc_must_snprintf(base_dir, sizeof(base_dir), "/tmp/snap.%s", snap_name);
+	sc_must_snprintf(tmp_dir, sizeof(tmp_dir), "%s/tmp", base_dir);
 
-	if (mkdir(tmpdir, 01777) != 0) {
-		die("cannot create temporary directory for private /tmp");
+	/* Switch to root group so that mkdir and open calls below create filesystem
+	 * elements that are not owned by the user calling into snap-confine. */
+	sc_identity old = sc_set_effective_identity(sc_root_group_identity());
+	// Create /tmp/snap.$SNAP_NAME/ 0700 root.root. Ignore EEXIST since we want
+	// to reuse and we will open with O_NOFOLLOW, below.
+	if (mkdir(base_dir, 0700) < 0 && errno != EEXIST) {
+		die("cannot create base directory %s", base_dir);
 	}
-	umask(old_mask);
-
-	// chdir to '/' since the mount won't apply to the current directory
-	char *pwd = get_current_dir_name();
-	if (pwd == NULL)
-		die("cannot get current working directory");
-	if (chdir("/") != 0)
-		die("cannot change directory to '/'");
-
-	// MS_BIND is there from linux 2.4
-	sc_do_mount(tmpdir, "/tmp", NULL, MS_BIND, NULL);
-	// MS_PRIVATE needs linux > 2.6.11
+	base_dir_fd = open(base_dir,
+			   O_RDONLY | O_DIRECTORY | O_CLOEXEC | O_NOFOLLOW);
+	if (base_dir_fd < 0) {
+		die("cannot open base directory %s", base_dir);
+	}
+	/* This seems redundant on first read but it has the non-obvious
+	 * property of changing existing directories  that have already existed
+	 * but had incorrect ownership or permission. This is possible due to
+	 * earlier bugs in snap-confine and due to the fact that some systems
+	 * use persistent /tmp directory and may not clean up leftover files
+	 * for arbitrarily long. This comment applies the following two pairs
+	 * of fchmod and fchown. */
+	if (fchmod(base_dir_fd, 0700) < 0) {
+		die("cannot chmod base directory %s to 0700", base_dir);
+	}
+	if (fchown(base_dir_fd, 0, 0) < 0) {
+		die("cannot chown base directory %s to root.root", base_dir);
+	}
+	// Create /tmp/snap.$SNAP_NAME/tmp 01777 root.root Ignore EEXIST since we
+	// want to reuse and we will open with O_NOFOLLOW, below.
+	if (mkdirat(base_dir_fd, "tmp", 01777) < 0 && errno != EEXIST) {
+		die("cannot create private tmp directory %s/tmp", base_dir);
+	}
+	(void)sc_set_effective_identity(old);
+	tmp_dir_fd = openat(base_dir_fd, "tmp",
+			    O_RDONLY | O_DIRECTORY | O_CLOEXEC | O_NOFOLLOW);
+	if (tmp_dir_fd < 0) {
+		die("cannot open private tmp directory %s/tmp", base_dir);
+	}
+	if (fchmod(tmp_dir_fd, 01777) < 0) {
+		die("cannot chmod private tmp directory %s/tmp to 01777",
+		    base_dir);
+	}
+	if (fchown(tmp_dir_fd, 0, 0) < 0) {
+		die("cannot chown private tmp directory %s/tmp to root.root",
+		    base_dir);
+	}
+	sc_do_mount(tmp_dir, "/tmp", NULL, MS_BIND, NULL);
 	sc_do_mount("none", "/tmp", NULL, MS_PRIVATE, NULL);
-	// do the chown after the bind mount to avoid potential shenanigans
-	if (chown("/tmp/", uid, gid) < 0) {
-		die("cannot change ownership of /tmp");
-	}
-	// chdir to original directory
-	if (chdir(pwd) != 0)
-		die("cannot change current working directory to the original directory");
-	free(pwd);
 }
 
 // TODO: fold this into bootstrap
@@ -137,63 +160,7 @@ static void setup_private_pts(void)
 	// copied from /etc/default/devpts
 	sc_do_mount("devpts", "/dev/pts", "devpts", MS_MGC_VAL,
 		    "newinstance,ptmxmode=0666,mode=0620,gid=5");
-	sc_do_mount("/dev/pts/ptmx", "/dev/ptmx", "none", MS_BIND, 0);
-}
-
-/**
- * Setup mount profiles by running snap-update-ns.
- *
- * The first argument is an open file descriptor (though opened with O_PATH, so
- * not as powerful), to a copy of snap-update-ns. The program is opened before
- * the root filesystem is pivoted so that it is easier to pick the right copy.
- **/
-static void sc_setup_mount_profiles(struct sc_apparmor *apparmor,
-				    int snap_update_ns_fd,
-				    const char *snap_name)
-{
-	debug("calling snap-update-ns to initialize mount namespace");
-	pid_t child = fork();
-	if (child < 0) {
-		die("cannot fork to run snap-update-ns");
-	}
-	if (child == 0) {
-		// We are the child, execute snap-update-ns under a dedicated profile.
-		char profile[PATH_MAX] = { 0 };
-		sc_must_snprintf(profile, sizeof profile, "snap-update-ns.%s",
-				 snap_name);
-		debug("launching snap-update-ns under per-snap profile %s",
-		      profile);
-		sc_maybe_aa_change_onexec(apparmor, profile);
-		char *snap_name_copy SC_CLEANUP(sc_cleanup_string) = NULL;
-		snap_name_copy = strdup(snap_name);
-		if (snap_name_copy == NULL) {
-			die("cannot copy snap name");
-		}
-		char *argv[] = {
-			"snap-update-ns", "--from-snap-confine", snap_name_copy,
-			NULL
-		};
-		char *envp[3] = { NULL };
-		if (sc_is_debug_enabled()) {
-			envp[0] = "SNAPD_DEBUG=1";
-		}
-		debug("fexecv(%d (snap-update-ns), %s %s %s,)",
-		      snap_update_ns_fd, argv[0], argv[1], argv[2]);
-		fexecve(snap_update_ns_fd, argv, envp);
-		die("cannot execute snap-update-ns");
-	}
-	// We are the parent, so wait for snap-update-ns to finish.
-	int status = 0;
-	debug("waiting for snap-update-ns to finish...");
-	if (waitpid(child, &status, 0) < 0) {
-		die("waitpid() failed for snap-update-ns process");
-	}
-	if (WIFEXITED(status) && WEXITSTATUS(status) != 0) {
-		die("snap-update-ns failed with code %i", WEXITSTATUS(status));
-	} else if (WIFSIGNALED(status)) {
-		die("snap-update-ns killed by signal %i", WTERMSIG(status));
-	}
-	debug("snap-update-ns finished successfully");
+	sc_do_mount("/dev/pts/ptmx", "/dev/ptmx", "none", MS_BIND, NULL);
 }
 
 struct sc_mount {
@@ -202,14 +169,18 @@ struct sc_mount {
 	// Alternate path defines the rbind mount "alternative" of path.
 	// It exists so that we can make /media on systems that use /run/media.
 	const char *altpath;
+	// Optional mount points are not processed unless the source and
+	// destination both exist.
+	bool is_optional;
 };
 
 struct sc_mount_config {
 	const char *rootfs_dir;
 	// The struct is terminated with an entry with NULL path.
 	const struct sc_mount *mounts;
-	bool on_classic_distro;
-	bool uses_base_snap;
+	sc_distro distro;
+	bool normal_mode;
+	const char *base_snap_name;
 };
 
 /**
@@ -282,10 +253,10 @@ static void sc_bootstrap_mount_namespace(const struct sc_mount_config *config)
 	// replicate the state of the root filesystem into the scratch directory.
 	sc_do_mount(config->rootfs_dir, scratch_dir, NULL, MS_REC | MS_BIND,
 		    NULL);
-	// Make the scratch directory recursively private. Nothing done there will
-	// be shared with any peer group, This effectively detaches us from the
-	// original namespace and coupled with pivot_root below serves as the
-	// foundation of the mount sandbox.
+	// Make the scratch directory recursively slave. Nothing done there will be
+	// shared with the initial mount namespace. This effectively detaches us,
+	// in one way, from the original namespace and coupled with pivot_root
+	// below serves as the foundation of the mount sandbox.
 	sc_do_mount("none", scratch_dir, NULL, MS_REC | MS_SLAVE, NULL);
 	// Bind mount certain directories from the host filesystem to the scratch
 	// directory. By default mount events will propagate in both into and out
@@ -294,13 +265,28 @@ static void sc_bootstrap_mount_namespace(const struct sc_mount_config *config)
 	// disabling the "is_bidirectional" flag as can be seen below.
 	for (const struct sc_mount * mnt = config->mounts; mnt->path != NULL;
 	     mnt++) {
-		if (mnt->is_bidirectional && mkdir(mnt->path, 0755) < 0 &&
-		    errno != EEXIST) {
-			die("cannot create %s", mnt->path);
+
+		if (mnt->is_bidirectional) {
+			sc_identity old =
+			    sc_set_effective_identity(sc_root_group_identity());
+			if (mkdir(mnt->path, 0755) < 0 && errno != EEXIST) {
+				die("cannot create %s", mnt->path);
+			}
+			(void)sc_set_effective_identity(old);
 		}
 		sc_must_snprintf(dst, sizeof dst, "%s/%s", scratch_dir,
 				 mnt->path);
-		sc_do_mount(mnt->path, dst, NULL, MS_REC | MS_BIND, NULL);
+		if (mnt->is_optional) {
+			bool ok = sc_do_optional_mount(mnt->path, dst, NULL,
+						       MS_REC | MS_BIND, NULL);
+			if (!ok) {
+				// If we cannot mount it, just continue.
+				continue;
+			}
+		} else {
+			sc_do_mount(mnt->path, dst, NULL, MS_REC | MS_BIND,
+				    NULL);
+		}
 		if (!mnt->is_bidirectional) {
 			// Mount events will only propagate inwards to the namespace. This
 			// way the running application cannot alter any global state apart
@@ -326,7 +312,7 @@ static void sc_bootstrap_mount_namespace(const struct sc_mount_config *config)
 			sc_do_mount("none", dst, NULL, MS_REC | MS_SLAVE, NULL);
 		}
 	}
-	if (config->on_classic_distro) {
+	if (config->normal_mode) {
 		// Since we mounted /etc from the host filesystem to the scratch directory,
 		// we may need to put certain directories from the desired root filesystem
 		// (e.g. the core snap) back. This way the behavior of running snaps is not
@@ -335,29 +321,92 @@ static void sc_bootstrap_mount_namespace(const struct sc_mount_config *config)
 		// Fixes the following bugs:
 		//  - https://bugs.launchpad.net/snap-confine/+bug/1580018
 		//  - https://bugzilla.opensuse.org/show_bug.cgi?id=1028568
-		const char *dirs_from_core[] =
-		    { "/etc/alternatives", "/etc/ssl", "/etc/nsswitch.conf",
+		const char *dirs_from_core[] = {
+			"/etc/alternatives", "/etc/nsswitch.conf",
+			// Some specific and privileged interfaces (e.g docker-support) give
+			// access to apparmor_parser from the base snap which at a minimum
+			// needs to use matching configuration from the base snap instead
+			// of from the users host system.
+			"/etc/apparmor", "/etc/apparmor.d",
+			// Use ssl certs from the base by default unless
+			// using Debian/Ubuntu classic (see below)
+			"/etc/ssl",
 			NULL
 		};
+
 		for (const char **dirs = dirs_from_core; *dirs != NULL; dirs++) {
 			const char *dir = *dirs;
-			struct stat buf;
-			if (access(dir, F_OK) == 0) {
-				sc_must_snprintf(src, sizeof src, "%s%s",
-						 config->rootfs_dir, dir);
-				sc_must_snprintf(dst, sizeof dst, "%s%s",
-						 scratch_dir, dir);
-				if (lstat(src, &buf) == 0
-				    && lstat(dst, &buf) == 0) {
-					sc_do_mount(src, dst, NULL, MS_BIND,
-						    NULL);
-					sc_do_mount("none", dst, NULL, MS_SLAVE,
-						    NULL);
-				}
+
+			// Special case for ubuntu/debian based
+			// classic distros that use the core* snap:
+			// here we use the host /etc/ssl
+			// to support custom ca-cert setups
+			if (sc_streq(dir, "/etc/ssl") &&
+			    config->distro == SC_DISTRO_CLASSIC &&
+			    sc_is_debian_like() &&
+			    sc_startswith(config->base_snap_name, "core")) {
+				continue;
 			}
+
+			if (access(dir, F_OK) != 0) {
+				continue;
+			}
+			struct stat dst_stat;
+			struct stat src_stat;
+			sc_must_snprintf(src, sizeof src, "%s%s",
+					 config->rootfs_dir, dir);
+			sc_must_snprintf(dst, sizeof dst, "%s%s",
+					 scratch_dir, dir);
+			if (lstat(src, &src_stat) != 0) {
+				if (errno == ENOENT) {
+					continue;
+				}
+				die("cannot stat %s from desired rootfs", src);
+			}
+			if (!S_ISREG(src_stat.st_mode)
+			    && !S_ISDIR(src_stat.st_mode)) {
+				debug
+				    ("entry %s from the desired rootfs is not a file or directory, skipping mount",
+				     src);
+				continue;
+			}
+
+			if (lstat(dst, &dst_stat) != 0) {
+				if (errno == ENOENT) {
+					continue;
+				}
+				die("cannot stat %s from host", src);
+			}
+			if (!S_ISREG(dst_stat.st_mode)
+			    && !S_ISDIR(dst_stat.st_mode)) {
+				debug
+				    ("entry %s from the host is not a file or directory, skipping mount",
+				     src);
+				continue;
+			}
+
+			if ((dst_stat.st_mode & S_IFMT) !=
+			    (src_stat.st_mode & S_IFMT)) {
+				debug
+				    ("entries %s and %s are of different types, skipping mount",
+				     dst, src);
+				continue;
+			}
+			// both source and destination exist where both are either files
+			// or both are directories
+			sc_do_mount(src, dst, NULL, MS_BIND, NULL);
+			sc_do_mount("none", dst, NULL, MS_SLAVE, NULL);
 		}
 	}
-	if (config->uses_base_snap) {
+	// The "core" base snap is special as it contains snapd and friends.
+	// Other base snaps do not, so whenever a base snap other than core is
+	// in use we need extra provisions for setting up internal tooling to
+	// be available.
+	//
+	// However on a core18 (and similar) system the core snap is not
+	// a special base anymore and we should map our own tooling in.
+	if (config->distro == SC_DISTRO_CORE_OTHER
+	    || !sc_streq(config->base_snap_name, "core")) {
 		// when bases are used we need to bind-mount the libexecdir
 		// (that contains snap-exec) into /usr/lib/snapd of the
 		// base snap so that snap-exec is available for the snaps
@@ -369,16 +418,23 @@ static void sc_bootstrap_mount_namespace(const struct sc_mount_config *config)
 				 scratch_dir);
 
 		// bind mount the current $ROOT/usr/lib/snapd path,
-		// where $ROOT is either "/" or the "/snap/core/current"
+		// where $ROOT is either "/" or the "/snap/{core,snapd}/current"
 		// that we are re-execing from
 		char *src = NULL;
 		char self[PATH_MAX + 1] = { 0 };
-		if (readlink("/proc/self/exe", self, sizeof(self) - 1) < 0) {
+		ssize_t nread;
+		nread = readlink("/proc/self/exe", self, sizeof self - 1);
+		if (nread < 0) {
 			die("cannot read /proc/self/exe");
 		}
+		// Though we initialized self to NULs and passed one less to
+		// readlink, therefore guaranteeing that self is
+		// zero-terminated, perform an explicit assignment to make
+		// Coverity happy.
+		self[nread] = '\0';
 		// this cannot happen except when the kernel is buggy
 		if (strstr(self, "/snap-confine") == NULL) {
-			die("cannot use result from readlink: %s", src);
+			die("cannot use result from readlink: %s", self);
 		}
 		src = dirname(self);
 		// dirname(path) might return '.' depending on path.
@@ -390,36 +446,28 @@ static void sc_bootstrap_mount_namespace(const struct sc_mount_config *config)
 
 		sc_do_mount(src, dst, NULL, MS_BIND | MS_RDONLY, NULL);
 		sc_do_mount("none", dst, NULL, MS_SLAVE, NULL);
-
-		// FIXME: snapctl tool - our apparmor policy wants it in
-		//        /usr/bin/snapctl, we will need an empty file
-		//        here from the base snap or we need to move it
-		//        into a different location and just symlink it
-		//        (/usr/lib/snapd/snapctl -> /usr/bin/snapctl)
-		//        and in the base snap case adjust PATH
-		//src = "/usr/bin/snapctl";
-		//sc_must_snprintf(dst, sizeof dst, "%s%s", scratch_dir, src);
-		//sc_do_mount(src, dst, NULL, MS_REC | MS_BIND, NULL);
-		//sc_do_mount("none", dst, NULL, MS_REC | MS_SLAVE, NULL);
 	}
 	// Bind mount the directory where all snaps are mounted. The location of
 	// the this directory on the host filesystem may not match the location in
 	// the desired root filesystem. In the "core" and "ubuntu-core" snaps the
 	// directory is always /snap. On the host it is a build-time configuration
-	// option stored in SNAP_MOUNT_DIR.
-	sc_must_snprintf(dst, sizeof dst, "%s/snap", scratch_dir);
-	sc_do_mount(SNAP_MOUNT_DIR, dst, NULL, MS_BIND | MS_REC | MS_SLAVE,
-		    NULL);
-	sc_do_mount("none", dst, NULL, MS_REC | MS_SLAVE, NULL);
+	// option stored in SNAP_MOUNT_DIR. In legacy mode (or in other words, not
+	// in normal mode), we don't need to do this because /snap is fixed and
+	// already contains the correct view of the mounted snaps.
+	if (config->normal_mode) {
+		sc_must_snprintf(dst, sizeof dst, "%s/snap", scratch_dir);
+		sc_do_mount(SNAP_MOUNT_DIR, dst, NULL, MS_BIND | MS_REC, NULL);
+		sc_do_mount("none", dst, NULL, MS_REC | MS_SLAVE, NULL);
+	}
 	// Create the hostfs directory if one is missing. This directory is a part
 	// of packaging now so perhaps this code can be removed later.
-	if (access(SC_HOSTFS_DIR, F_OK) != 0) {
-		debug("creating missing hostfs directory");
-		if (mkdir(SC_HOSTFS_DIR, 0755) != 0) {
-			die("cannot perform operation: mkdir %s",
-			    SC_HOSTFS_DIR);
+	sc_identity old = sc_set_effective_identity(sc_root_group_identity());
+	if (mkdir(SC_HOSTFS_DIR, 0755) < 0) {
+		if (errno != EEXIST) {
+			die("cannot perform operation: mkdir %s", SC_HOSTFS_DIR);
 		}
 	}
+	(void)sc_set_effective_identity(old);
 	// Ensure that hostfs isgroup owned by root. We may have (now or earlier)
 	// created the directory as the user who first ran a snap on a given
 	// system and the group identity of that user is visilbe on disk.
@@ -445,7 +493,7 @@ static void sc_bootstrap_mount_namespace(const struct sc_mount_config *config)
 	// uniform way after pivot_root but this is good enough and requires less
 	// code changes the nvidia code assumes it has access to the existing
 	// pre-pivot filesystem.
-	if (config->on_classic_distro) {
+	if (config->distro == SC_DISTRO_CLASSIC) {
 		sc_mount_nvidia_driver(scratch_dir);
 	}
 	// XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
@@ -480,7 +528,7 @@ static void sc_bootstrap_mount_namespace(const struct sc_mount_config *config)
 	// This way we can remove the temporary directory we created and "clean up"
 	// after ourselves nicely.
 	sc_must_snprintf(dst, sizeof dst, "%s/%s", SC_HOSTFS_DIR, scratch_dir);
-	sc_do_umount(dst, 0);
+	sc_do_umount(dst, UMOUNT_NOFOLLOW);
 	// Remove the scratch directory. Note that we are using the path that is
 	// based on the old root filesystem as after pivot_root we cannot guarantee
 	// what is present at the same location normally. (It is probably an empty
@@ -506,6 +554,43 @@ static void sc_bootstrap_mount_namespace(const struct sc_mount_config *config)
 	// mount table and software inspecting the mount table may become confused.
 	sc_must_snprintf(src, sizeof src, "%s/proc", SC_HOSTFS_DIR);
 	sc_do_umount(src, UMOUNT_NOFOLLOW | MNT_DETACH);
+	// Detach both views of /writable: the one from hostfs and the one directly
+	// visible in /writable. Interfaces don't grant access to this directory
+	// and it has a large duplicated view of many mount points.  Note that this
+	// is only applicable to ubuntu-core systems.
+	sc_detach_views_of_writable(config->distro, config->normal_mode);
+}
+
+static void sc_detach_views_of_writable(sc_distro distro, bool normal_mode)
+{
+	// Note that prior to detaching either mount point we switch the
+	// propagation to private to both limit the change to just this view and to
+	// prevent otherwise occurring event propagation from self-conflicting and
+	// returning EBUSY. A similar approach is used by snap-update-ns and is
+	// documented in umount(2).
+	const char *writable_dir = "/writable";
+	const char *hostfs_writable_dir = "/var/lib/snapd/hostfs/writable";
+
+	// Writable only exists on ubuntu-core.
+	if (distro == SC_DISTRO_CLASSIC) {
+		return;
+	}
+	// On all core distributions we see /var/lib/snapd/hostfs/writable that
+	// exposes writable, with a structure specific to ubuntu-core.
+	debug("detaching %s", hostfs_writable_dir);
+	sc_do_mount("none", hostfs_writable_dir, NULL,
+		    MS_REC | MS_PRIVATE, NULL);
+	sc_do_umount(hostfs_writable_dir, UMOUNT_NOFOLLOW | MNT_DETACH);
+
+	// On ubuntu-core 16, when the executed snap uses core as base we also see
+	// the /writable that we directly inherited from the initial mount
+	// namespace.
+	if (distro == SC_DISTRO_CORE16 && !normal_mode) {
+		debug("detaching %s", writable_dir);
+		sc_do_mount("none", writable_dir, NULL, MS_REC | MS_PRIVATE,
+			    NULL);
+		sc_do_umount(writable_dir, UMOUNT_NOFOLLOW | MNT_DETACH);
+	}
 }
 
 /**
@@ -515,8 +600,8 @@ static void sc_bootstrap_mount_namespace(const struct sc_mount_config *config)
  * @fulllen: full original path length.
  * Returns a pointer to the next path segment, or NULL if done.
  */
-static char * __attribute__ ((used))
-    get_nextpath(char *path, size_t * offsetp, size_t fulllen)
+static char * __attribute__((used))
+    get_nextpath(char *path, size_t *offsetp, size_t fulllen)
 {
 	size_t offset = *offsetp;
 
@@ -535,7 +620,7 @@ static char * __attribute__ ((used))
 /**
  * Check that @subdir is a subdir of @dir.
 **/
-static bool __attribute__ ((used))
+static bool __attribute__((used))
     is_subdir(const char *subdir, const char *dir)
 {
 	size_t dirlen = strlen(dir);
@@ -564,50 +649,16 @@ static bool __attribute__ ((used))
 	return false;
 }
 
-int sc_open_snap_update_ns(void)
-{
-	// +1 is for the case where the link is exactly PATH_MAX long but we also
-	// want to store the terminating '\0'. The readlink system call doesn't add
-	// terminating null, but our initialization of buf handles this for us.
-	char buf[PATH_MAX + 1] = { 0 };
-	if (readlink("/proc/self/exe", buf, sizeof buf) < 0) {
-		die("cannot readlink /proc/self/exe");
-	}
-	if (buf[0] != '/') {	// this shouldn't happen, but make sure have absolute path
-		die("readlink /proc/self/exe returned relative path");
-	}
-	char *bufcopy SC_CLEANUP(sc_cleanup_string) = NULL;
-	bufcopy = strdup(buf);
-	if (bufcopy == NULL) {
-		die("cannot copy buffer");
-	}
-	char *dname = dirname(bufcopy);
-	sc_must_snprintf(buf, sizeof buf, "%s/%s", dname, "snap-update-ns");
-	debug("snap-update-ns executable: %s", buf);
-	int fd = open(buf, O_PATH | O_RDONLY | O_NOFOLLOW | O_CLOEXEC);
-	if (fd < 0) {
-		die("cannot open snap-update-ns executable");
-	}
-	debug("opened snap-update-ns executable as file descriptor %d", fd);
-	return fd;
-}
-
 void sc_populate_mount_ns(struct sc_apparmor *apparmor, int snap_update_ns_fd,
-			  const char *base_snap_name, const char *snap_name)
+			  const sc_invocation * inv, const gid_t real_gid,
+			  const gid_t saved_gid)
 {
-	// Get the current working directory before we start fiddling with
-	// mounts and possibly pivot_root.  At the end of the whole process, we
-	// will try to re-locate to the same directory (if possible).
-	char *vanilla_cwd SC_CLEANUP(sc_cleanup_string) = NULL;
-	vanilla_cwd = get_current_dir_name();
-	if (vanilla_cwd == NULL) {
-		die("cannot get the current working directory");
-	}
+	// Classify the current distribution, as claimed by /etc/os-release.
+	sc_distro distro = sc_classify_distro();
 
-	bool on_classic_distro = is_running_on_classic_distribution();
-	// on classic or with alternative base snaps we need to setup
-	// a different confinement
-	if (on_classic_distro || !sc_streq(base_snap_name, "core")) {
+	// Check which mode we should run in, normal or legacy.
+	if (inv->is_normal_mode) {
+		// In normal mode we use the base snap as / and set up several bind mounts.
 		const struct sc_mount mounts[] = {
 			{"/dev"},	// because it contains devices on host OS
 			{"/etc"},	// because that's where /etc/resolv.conf lives, perhaps a bad idea
@@ -620,7 +671,8 @@ void sc_populate_mount_ns(struct sc_apparmor *apparmor, int snap_update_ns_fd,
 			{"/var/lib/snapd"},	// to get access to snapd state and seccomp profiles
 			{"/var/tmp"},	// to get access to the other temporary directory
 			{"/run"},	// to get /run with sockets and what not
-			{"/lib/modules"},	// access to the modules of the running kernel
+			{"/lib/modules",.is_optional = true},	// access to the modules of the running kernel
+			{"/lib/firmware",.is_optional = true},	// access to the firmware of the running kernel
 			{"/usr/src"},	// FIXME: move to SecurityMounts in system-trace interface
 			{"/var/log"},	// FIXME: move to SecurityMounts in log-observe interface
 #ifdef MERGED_USR
@@ -629,93 +681,60 @@ void sc_populate_mount_ns(struct sc_apparmor *apparmor, int snap_update_ns_fd,
 			{"/media", true},	// access to the users removable devices
 #endif				// MERGED_USR
 			{"/run/netns", true},	// access to the 'ip netns' network namespaces
+			// The /mnt directory is optional in base snaps to ensure backwards
+			// compatibility with the first version of base snaps that was
+			// released.
+			{"/mnt",.is_optional = true},	// to support the removable-media interface
+			{"/var/lib/extrausers",.is_optional = true},	// access to UID/GID of extrausers (if available)
 			{},
 		};
-		char rootfs_dir[PATH_MAX] = { 0 };
-		sc_must_snprintf(rootfs_dir, sizeof rootfs_dir,
-				 "%s/%s/current/", SNAP_MOUNT_DIR,
-				 base_snap_name);
-		if (access(rootfs_dir, F_OK) != 0) {
-			if (sc_streq(base_snap_name, "core")) {
-				// As a special fallback, allow the
-				// base snap to degrade from "core" to
-				// "ubuntu-core". This is needed for
-				// the migration tests.
-				base_snap_name = "ubuntu-core";
-				sc_must_snprintf(rootfs_dir, sizeof rootfs_dir,
-						 "%s/%s/current/",
-						 SNAP_MOUNT_DIR,
-						 base_snap_name);
-				if (access(rootfs_dir, F_OK) != 0) {
-					die("cannot locate the core or legacy core snap (current symlink missing?)");
-				}
-			}
-			die("cannot locate the base snap: %s", base_snap_name);
-		}
-		struct sc_mount_config classic_config = {
-			.rootfs_dir = rootfs_dir,
+		struct sc_mount_config normal_config = {
+			.rootfs_dir = inv->rootfs_dir,
 			.mounts = mounts,
-			.on_classic_distro = true,
-			.uses_base_snap = !sc_streq(base_snap_name, "core"),
+			.distro = distro,
+			.normal_mode = true,
+			.base_snap_name = inv->base_snap_name,
 		};
-		sc_bootstrap_mount_namespace(&classic_config);
+		sc_bootstrap_mount_namespace(&normal_config);
 	} else {
-		// This is what happens on an all-snap system. The rootfs we start with
-		// is the real outer rootfs.  There are no unidirectional bind mounts
-		// needed because everything is already OK. We still keep the
-		// bidirectional /media mount point so that snaps designed for mounting
-		// filesystems can use that space for whatever they need.
+		// In legacy mode we don't pivot and instead just arrange bi-
+		// directional mount propagation for two directories.
 		const struct sc_mount mounts[] = {
 			{"/media", true},
 			{"/run/netns", true},
 			{},
 		};
-		struct sc_mount_config all_snap_config = {
+		struct sc_mount_config legacy_config = {
 			.rootfs_dir = "/",
 			.mounts = mounts,
-			.uses_base_snap = !sc_streq(base_snap_name, "core"),
+			.distro = distro,
+			.normal_mode = false,
+			.base_snap_name = inv->base_snap_name,
 		};
-		sc_bootstrap_mount_namespace(&all_snap_config);
+		sc_bootstrap_mount_namespace(&legacy_config);
 	}
 
-	// set up private mounts
 	// TODO: rename this and fold it into bootstrap
-	setup_private_mount(snap_name);
-
+	setup_private_mount(inv->snap_instance);
 	// set up private /dev/pts
 	// TODO: fold this into bootstrap
 	setup_private_pts();
 
-	// setup quirks for specific snaps
-	if (on_classic_distro) {
-		sc_setup_quirks();
-	}
 	// setup the security backend bind mounts
-	sc_setup_mount_profiles(apparmor, snap_update_ns_fd, snap_name);
-
-	// Try to re-locate back to vanilla working directory. This can fail
-	// because that directory is no longer present.
-	if (chdir(vanilla_cwd) != 0) {
-		debug("cannot remain in %s, moving to the void directory",
-		      vanilla_cwd);
-		if (chdir(SC_VOID_DIR) != 0) {
-			die("cannot change directory to %s", SC_VOID_DIR);
-		}
-		debug("successfully moved to %s", SC_VOID_DIR);
-	}
+	sc_call_snap_update_ns(snap_update_ns_fd, inv->snap_instance, apparmor);
 }
 
 static bool is_mounted_with_shared_option(const char *dir)
-    __attribute__ ((nonnull(1)));
+    __attribute__((nonnull(1)));
 
 static bool is_mounted_with_shared_option(const char *dir)
 {
-	struct sc_mountinfo *sm SC_CLEANUP(sc_cleanup_mountinfo) = NULL;
+	sc_mountinfo *sm SC_CLEANUP(sc_cleanup_mountinfo) = NULL;
 	sm = sc_parse_mountinfo(NULL);
 	if (sm == NULL) {
 		die("cannot parse /proc/self/mountinfo");
 	}
-	struct sc_mountinfo_entry *entry = sc_first_mountinfo_entry(sm);
+	sc_mountinfo_entry *entry = sc_first_mountinfo_entry(sm);
 	while (entry != NULL) {
 		const char *mount_dir = entry->mount_dir;
 		if (sc_streq(mount_dir, dir)) {
@@ -739,25 +758,14 @@ void sc_ensure_shared_snap_mount(void)
 		// removed once we have a measurement and feedback mechanism that lets
 		// us decide based on measurable data.
 		sc_do_mount(SNAP_MOUNT_DIR, SNAP_MOUNT_DIR, "none",
-			    MS_BIND | MS_REC, 0);
+			    MS_BIND | MS_REC, NULL);
 		sc_do_mount("none", SNAP_MOUNT_DIR, NULL, MS_SHARED | MS_REC,
 			    NULL);
 	}
 }
 
-static void sc_make_slave_mount_ns(void)
-{
-	if (unshare(CLONE_NEWNS) < 0) {
-		die("can not unshare mount namespace");
-	}
-	// In our new mount namespace, recursively change all mounts
-	// to slave mode, so we see changes from the parent namespace
-	// but don't propagate our own changes.
-	sc_do_mount("none", "/", NULL, MS_REC | MS_SLAVE, NULL);
-}
-
 void sc_setup_user_mounts(struct sc_apparmor *apparmor, int snap_update_ns_fd,
-                          const char *snap_name)
+			  const char *snap_name)
 {
 	debug("%s: %s", __FUNCTION__, snap_name);
 
@@ -771,59 +779,58 @@ void sc_setup_user_mounts(struct sc_apparmor *apparmor, int snap_update_ns_fd,
 		return;
 	}
 
-	sc_make_slave_mount_ns();
+	// In our new mount namespace, recursively change all mounts
+	// to slave mode, so we see changes from the parent namespace
+	// but don't propagate our own changes.
+	sc_do_mount("none", "/", NULL, MS_REC | MS_SLAVE, NULL);
+	sc_identity old = sc_set_effective_identity(sc_root_group_identity());
+	sc_call_snap_update_ns_as_user(snap_update_ns_fd, snap_name, apparmor);
+	(void)sc_set_effective_identity(old);
+}
 
-	debug("calling snap-update-ns to initialize user mounts");
-	pid_t child = fork();
-	if (child < 0) {
-		die("cannot fork to run snap-update-ns");
+void sc_ensure_snap_dir_shared_mounts(void)
+{
+	const char *dirs[] = { SNAP_MOUNT_DIR, "/var/snap", NULL };
+	for (int i = 0; dirs[i] != NULL; i++) {
+		const char *dir = dirs[i];
+		if (!is_mounted_with_shared_option(dir)) {
+			/* Since this directory isn't yet shared (but it should be),
+			 * recursively bind mount it, then recursively share it so that
+			 * changes to the host are seen in the snap and vice-versa. This
+			 * allows us to fine-tune propagation events elsewhere for this new
+			 * mountpoint.
+			 *
+			 * Not using MS_SLAVE because it's too late for SNAP_MOUNT_DIR,
+			 * since snaps are already mounted, and it's not needed for
+			 * /var/snap.
+			 */
+			sc_do_mount(dir, dir, "none", MS_BIND | MS_REC, NULL);
+			sc_do_mount("none", dir, NULL, MS_REC | MS_SHARED,
+				    NULL);
+		}
 	}
-	if (child == 0) {
-		// We are the child, execute snap-update-ns under a dedicated profile.
-		char profile[PATH_MAX] = { 0 };
-		sc_must_snprintf(profile, sizeof profile, "snap-update-ns.%s",
-				 snap_name);
-		debug("launching snap-update-ns under per-snap profile %s",
-		      profile);
-		sc_maybe_aa_change_onexec(apparmor, profile);
-		char *snap_name_copy SC_CLEANUP(sc_cleanup_string) = NULL;
-		snap_name_copy = strdup(snap_name);
-		if (snap_name_copy == NULL) {
-			die("cannot allocate memory for snap name");
-		}
-		char *argv[] = {
-			"snap-update-ns", "--user-mounts", snap_name_copy,
-			NULL
-		};
-		char *envp[3] = { NULL };
-		int last_env = 0;
-		if (sc_is_debug_enabled()) {
-			envp[last_env++] = "SNAPD_DEBUG=1";
-		}
-		const char *xdg_runtime_dir = getenv("XDG_RUNTIME_DIR");
-		char xdg_runtime_dir_env[PATH_MAX];
-		if (xdg_runtime_dir != NULL) {
-			sc_must_snprintf(xdg_runtime_dir_env,
-					 sizeof(xdg_runtime_dir_env),
-					 "XDG_RUNTIME_DIR=%s", xdg_runtime_dir);
-			envp[last_env++] = xdg_runtime_dir_env;
-		}
+}
+
+void sc_setup_parallel_instance_classic_mounts(const char *snap_name,
+					       const char *snap_instance_name)
+{
+	char src[PATH_MAX] = { 0 };
+	char dst[PATH_MAX] = { 0 };
 
-		debug("fexecv(%d (snap-update-ns), %s %s %s,)",
-		      snap_update_ns_fd, argv[0], argv[1], argv[2]);
-		fexecve(snap_update_ns_fd, argv, envp);
-		die("cannot execute snap-update-ns");
+	const char *dirs[] = { SNAP_MOUNT_DIR, "/var/snap", NULL };
+	for (int i = 0; dirs[i] != NULL; i++) {
+		const char *dir = dirs[i];
+		sc_do_mount("none", dir, NULL, MS_REC | MS_SLAVE, NULL);
 	}
-	// We are the parent, so wait for snap-update-ns to finish.
-	int status = 0;
-	debug("waiting for snap-update-ns to finish...");
-	if (waitpid(child, &status, 0) < 0) {
-		die("waitpid() failed for snap-update-ns process");
-	}
-	if (WIFEXITED(status) && WEXITSTATUS(status) != 0) {
-		die("snap-update-ns failed with code %i", WEXITSTATUS(status));
-	} else if (WIFSIGNALED(status)) {
-		die("snap-update-ns killed by signal %i", WTERMSIG(status));
-	}
-	debug("snap-update-ns finished successfully");
+
+	/* Mount SNAP_MOUNT_DIR/<snap>_<key> on SNAP_MOUNT_DIR/<snap> */
+	sc_must_snprintf(src, sizeof src, "%s/%s", SNAP_MOUNT_DIR,
+			 snap_instance_name);
+	sc_must_snprintf(dst, sizeof dst, "%s/%s", SNAP_MOUNT_DIR, snap_name);
+	sc_do_mount(src, dst, "none", MS_BIND | MS_REC, NULL);
+
+	/* Mount /var/snap/<snap>_<key> on /var/snap/<snap> */
+	sc_must_snprintf(src, sizeof src, "/var/snap/%s", snap_instance_name);
+	sc_must_snprintf(dst, sizeof dst, "/var/snap/%s", snap_name);
+	sc_do_mount(src, dst, "none", MS_BIND | MS_REC, NULL);
 }
