@@ -27,43 +27,57 @@ import (
 	"sort"
 	"strings"
 	"time"
+	"unicode/utf8"
 
 	"github.com/jessevdk/go-flags"
 
 	"github.com/snapcore/snapd/client"
+	"github.com/snapcore/snapd/dirs"
 	"github.com/snapcore/snapd/i18n"
 	"github.com/snapcore/snapd/osutil"
+	"github.com/snapcore/snapd/release"
+	"github.com/snapcore/snapd/snap/channel"
+	"github.com/snapcore/snapd/strutil"
 )
 
 var (
-	shortInstallHelp = i18n.G("Install a snap to the system")
-	shortRemoveHelp  = i18n.G("Remove a snap from the system")
-	shortRefreshHelp = i18n.G("Refresh a snap in the system")
-	shortTryHelp     = i18n.G("Test a snap in the system")
+	shortInstallHelp = i18n.G("Install snaps on the system")
+	shortRemoveHelp  = i18n.G("Remove snaps from the system")
+	shortRefreshHelp = i18n.G("Refresh snaps in the system")
+	shortTryHelp     = i18n.G("Test an unpacked snap in the system")
 	shortEnableHelp  = i18n.G("Enable a snap in the system")
 	shortDisableHelp = i18n.G("Disable a snap in the system")
 )
 
 var longInstallHelp = i18n.G(`
-The install command installs the named snaps in the system.
+The install command installs the named snaps on the system.
+
+To install multiple instances of the same snap, append an underscore and a
+unique identifier (for each instance) to a snap's name.
 
 With no further options, the snaps are installed tracking the stable channel,
 with strict security confinement.
 
-Revision choice via the --revision override requires the the user to
+Revision choice via the --revision override requires the user to
 have developer access to the snap, either directly or through the
 store's collaboration feature, and to be logged in (see 'snap help login').
 
-Note a later refresh will typically undo a revision override, taking the snap
+Note that a later refresh will typically undo a revision override, taking the snap
 back to the current revision of the channel it's tracking.
+
+Use --name to set the instance name when installing from snap file.
 `)
 
 var longRemoveHelp = i18n.G(`
-The remove command removes the named snap from the system.
+The remove command removes the named snap instance from the system.
 
 By default all the snap revisions are removed, including their data and the
 common data directory. When a --revision option is passed only the specified
 revision is removed.
+
+Unless automatic snapshots are disabled, a snapshot of all data for the snap is 
+saved upon removal, which is then available for future restoration with snap
+restore. The --purge option disables automatically creating snapshots.
 `)
 
 var longRefreshHelp = i18n.G(`
@@ -73,7 +87,7 @@ none are specified.
 With no further options, the snaps are refreshed to the current revision of the
 channel they're tracking, preserving their confinement options.
 
-Revision choice via the --revision override requires the the user to
+Revision choice via the --revision override requires the user to
 have developer access to the snap, either directly or through the
 store's collaboration feature, and to be logged in (see 'snap help login').
 
@@ -105,6 +119,7 @@ type cmdRemove struct {
 	waitMixin
 
 	Revision   string `long:"revision"`
+	Purge      bool   `long:"purge"`
 	Positional struct {
 		Snaps []installedSnapName `positional-arg-name:"<snap>" required:"1"`
 	} `positional-args:"yes" required:"yes"`
@@ -113,8 +128,7 @@ type cmdRemove struct {
 func (x *cmdRemove) removeOne(opts *client.SnapOptions) error {
 	name := string(x.Positional.Snaps[0])
 
-	cli := Client()
-	changeID, err := cli.Remove(name, opts)
+	changeID, err := x.client.Remove(name, opts)
 	if err != nil {
 		msg, err := errorToCmdMessage(name, err, opts)
 		if err != nil {
@@ -124,26 +138,29 @@ func (x *cmdRemove) removeOne(opts *client.SnapOptions) error {
 		return nil
 	}
 
-	if _, err := x.wait(cli, changeID); err != nil {
+	if _, err := x.wait(changeID); err != nil {
 		if err == noWait {
 			return nil
 		}
 		return err
 	}
 
-	fmt.Fprintf(Stdout, i18n.G("%s removed\n"), name)
+	if opts.Revision != "" {
+		fmt.Fprintf(Stdout, i18n.G("%s (revision %s) removed\n"), name, opts.Revision)
+	} else {
+		fmt.Fprintf(Stdout, i18n.G("%s removed\n"), name)
+	}
 	return nil
 }
 
 func (x *cmdRemove) removeMany(opts *client.SnapOptions) error {
 	names := installedSnapNames(x.Positional.Snaps)
-	cli := Client()
-	changeID, err := cli.RemoveMany(names, opts)
+	changeID, err := x.client.RemoveMany(names, opts)
 	if err != nil {
 		return err
 	}
 
-	chg, err := x.wait(cli, changeID)
+	chg, err := x.wait(changeID)
 	if err != nil {
 		if err == noWait {
 			return nil
@@ -174,13 +191,13 @@ func (x *cmdRemove) removeMany(opts *client.SnapOptions) error {
 }
 
 func (x *cmdRemove) Execute([]string) error {
-	opts := &client.SnapOptions{Revision: x.Revision}
+	opts := &client.SnapOptions{Revision: x.Revision, Purge: x.Purge}
 	if len(x.Positional.Snaps) == 1 {
 		return x.removeOne(opts)
 	}
 
-	if x.Revision != "" {
-		return errors.New(i18n.G("a single snap name is needed to specify the revision"))
+	if x.Purge || x.Revision != "" {
+		return errors.New(i18n.G("a single snap name is needed to specify options"))
 	}
 	return x.removeMany(nil)
 }
@@ -209,11 +226,16 @@ func (mxd mixinDescs) also(m map[string]string) mixinDescs {
 }
 
 var channelDescs = mixinDescs{
-	"channel":   i18n.G("Use this channel instead of stable"),
-	"beta":      i18n.G("Install from the beta channel"),
-	"edge":      i18n.G("Install from the edge channel"),
+	// TRANSLATORS: This should not start with a lowercase letter.
+	"channel": i18n.G("Use this channel instead of stable"),
+	// TRANSLATORS: This should not start with a lowercase letter.
+	"beta": i18n.G("Install from the beta channel"),
+	// TRANSLATORS: This should not start with a lowercase letter.
+	"edge": i18n.G("Install from the edge channel"),
+	// TRANSLATORS: This should not start with a lowercase letter.
 	"candidate": i18n.G("Install from the candidate channel"),
-	"stable":    i18n.G("Install from the stable channel"),
+	// TRANSLATORS: This should not start with a lowercase letter.
+	"stable": i18n.G("Install from the stable channel"),
 }
 
 func (mx *channelMixin) setChannelFromCommandline() error {
@@ -235,41 +257,121 @@ func (mx *channelMixin) setChannelFromCommandline() error {
 		mx.Channel = ch.chName
 	}
 
-	if !strings.Contains(mx.Channel, "/") && mx.Channel != "" && mx.Channel != "edge" && mx.Channel != "beta" && mx.Channel != "candidate" && mx.Channel != "stable" {
-		// shortcut to jump to a different track, e.g.
-		// snap install foo --channel=3.4 # implies 3.4/stable
-		mx.Channel += "/stable"
+	if mx.Channel != "" {
+		if _, err := channel.Parse(mx.Channel, ""); err != nil {
+			full, er := channel.Full(mx.Channel)
+			if er != nil {
+				// the parse error has more detailed info
+				return err
+			}
+
+			// TODO: get escapes in here so we can bold the Warning
+			head := i18n.G("Warning:")
+			msg := i18n.G("Specifying a channel %q is relying on undefined behaviour. Interpreting it as %q for now, but this will be an error later.\n")
+			warn := fill(fmt.Sprintf(msg, mx.Channel, full), utf8.RuneCountInString(head)+1) // +1 for the space
+			fmt.Fprint(Stderr, head, " ", warn, "\n\n")
+			mx.Channel = full // so a malformed-but-eh channel will always be full, i.e. //stable// -> latest/stable
+		}
 	}
 
 	return nil
 }
 
+// isSnapInPath checks whether the snap binaries dir (e.g. /snap/bin)
+// is in $PATH.
+//
+// TODO: consider symlinks
+func isSnapInPath() bool {
+	paths := filepath.SplitList(os.Getenv("PATH"))
+	for _, path := range paths {
+		if filepath.Clean(path) == dirs.SnapBinariesDir {
+			return true
+		}
+	}
+	return false
+}
+
+func isSameRisk(tracking, current string) (bool, error) {
+	if tracking == current {
+		return true, nil
+	}
+	var trackingRisk, currentRisk string
+	if tracking != "" {
+		traCh, err := channel.Parse(tracking, "")
+		if err != nil {
+			return false, err
+		}
+		trackingRisk = traCh.Risk
+	}
+	if current != "" {
+		curCh, err := channel.Parse(current, "")
+		if err != nil {
+			return false, err
+		}
+		currentRisk = curCh.Risk
+	}
+	return trackingRisk == currentRisk, nil
+}
+
+func maybeWithSudoSecurePath() bool {
+	// Some distributions set secure_path to a known list of paths in
+	// /etc/sudoers. The snapd package currently has no means of overwriting
+	// or extending that setting.
+	if _, isSet := os.LookupEnv("SUDO_UID"); !isSet {
+		return false
+	}
+	// Known distros setting secure_path that does not include
+	// $SNAP_MOUNT_DIR/bin:
+	return release.DistroLike("fedora", "opensuse", "debian")
+}
+
 // show what has been done
-func showDone(names []string, op string) error {
-	cli := Client()
+func showDone(cli *client.Client, names []string, op string, opts *client.SnapOptions, esc *escapes) error {
 	snaps, err := cli.List(names, nil)
 	if err != nil {
 		return err
 	}
 
+	needsPathWarning := !isSnapInPath() && !maybeWithSudoSecurePath()
 	for _, snap := range snaps {
 		channelStr := ""
-		if snap.Channel != "" && snap.Channel != "stable" {
-			channelStr = fmt.Sprintf(" (%s)", snap.Channel)
+		if snap.Channel != "" {
+			ch, err := channel.Parse(snap.Channel, "")
+			if err != nil {
+				return err
+			}
+			if ch.Name != "stable" {
+				channelStr = fmt.Sprintf(" (%s)", ch.Name)
+			}
 		}
 		switch op {
 		case "install":
-			if snap.Developer != "" {
-				// TRANSLATORS: the args are a snap name optionally followed by a channel, then a version, then the developer name (e.g. "some-snap (beta) 1.3 from 'alice' installed")
-				fmt.Fprintf(Stdout, i18n.G("%s%s %s from '%s' installed\n"), snap.Name, channelStr, snap.Version, snap.Developer)
+			if needsPathWarning {
+				head := i18n.G("Warning:")
+				warn := fill(fmt.Sprintf(i18n.G("%s was not found in your $PATH. If you've not restarted your session since you installed snapd, try doing that. Please see https://forum.snapcraft.io/t/9469 for more details."), dirs.SnapBinariesDir), utf8.RuneCountInString(head)+1) // +1 for the space
+				fmt.Fprint(Stderr, esc.bold, head, esc.end, " ", warn, "\n\n")
+				needsPathWarning = false
+			}
+
+			if opts != nil && opts.Classic && snap.Confinement != client.ClassicConfinement {
+				// requested classic but the snap is not classic
+				head := i18n.G("Warning:")
+				// TRANSLATORS: the arg is a snap name (e.g. "some-snap")
+				warn := fill(fmt.Sprintf(i18n.G("flag --classic ignored for strictly confined snap %s"), snap.Name), utf8.RuneCountInString(head)+1) // +1 for the space
+				fmt.Fprint(Stderr, esc.bold, head, esc.end, " ", warn, "\n\n")
+			}
+
+			if snap.Publisher != nil {
+				// TRANSLATORS: the args are a snap name optionally followed by a channel, then a version, then the developer name (e.g. "some-snap (beta) 1.3 from Alice installed")
+				fmt.Fprintf(Stdout, i18n.G("%s%s %s from %s installed\n"), snap.Name, channelStr, snap.Version, longPublisher(esc, snap.Publisher))
 			} else {
 				// TRANSLATORS: the args are a snap name optionally followed by a channel, then a version (e.g. "some-snap (beta) 1.3 installed")
 				fmt.Fprintf(Stdout, i18n.G("%s%s %s installed\n"), snap.Name, channelStr, snap.Version)
 			}
 		case "refresh":
-			if snap.Developer != "" {
-				// TRANSLATORS: the args are a snap name optionally followed by a channel, then a version, then the developer name (e.g. "some-snap (beta) 1.3 from 'alice' refreshed")
-				fmt.Fprintf(Stdout, i18n.G("%s%s %s from '%s' refreshed\n"), snap.Name, channelStr, snap.Version, snap.Developer)
+			if snap.Publisher != nil {
+				// TRANSLATORS: the args are a snap name optionally followed by a channel, then a version, then the developer name (e.g. "some-snap (beta) 1.3 from Alice refreshed")
+				fmt.Fprintf(Stdout, i18n.G("%s%s %s from %s refreshed\n"), snap.Name, channelStr, snap.Version, longPublisher(esc, snap.Publisher))
 			} else {
 				// TRANSLATORS: the args are a snap name optionally followed by a channel, then a version (e.g. "some-snap (beta) 1.3 refreshed")
 				fmt.Fprintf(Stdout, i18n.G("%s%s %s refreshed\n"), snap.Name, channelStr, snap.Version)
@@ -277,12 +379,40 @@ func showDone(names []string, op string) error {
 		case "revert":
 			// TRANSLATORS: first %s is a snap name, second %s is a revision
 			fmt.Fprintf(Stdout, i18n.G("%s reverted to %s\n"), snap.Name, snap.Version)
+		case "switch":
+			switchCohort := opts.CohortKey != ""
+			switchChannel := opts.Channel != ""
+			var msg string
+			// we have three boolean things to check, meaning 2³=8 possibilities,
+			// minus 3 error cases which are handled before the call to showDone.
+			switch {
+			case switchCohort && !opts.LeaveCohort && !switchChannel:
+				// TRANSLATORS: the first %q will be the (quoted) snap name, the second an ellipted cohort string
+				msg = fmt.Sprintf(i18n.G("%q switched to the %q cohort\n"), snap.Name, strutil.ElliptLeft(opts.CohortKey, 10))
+			case switchCohort && !opts.LeaveCohort && switchChannel:
+				// TRANSLATORS: the first %q will be the (quoted) snap name, the second a channel, the third an ellipted cohort string
+				msg = fmt.Sprintf(i18n.G("%q switched to the %q channel and the %q cohort\n"), snap.Name, snap.TrackingChannel, strutil.ElliptLeft(opts.CohortKey, 10))
+			case !switchCohort && !opts.LeaveCohort && switchChannel:
+				// TRANSLATORS: the first %q will be the (quoted) snap name, the second a channel
+				msg = fmt.Sprintf(i18n.G("%q switched to the %q channel\n"), snap.Name, snap.TrackingChannel)
+			case !switchCohort && opts.LeaveCohort && switchChannel:
+				// TRANSLATORS: the first %q will be the (quoted) snap name, the second a channel
+				msg = fmt.Sprintf(i18n.G("%q left the cohort, and switched to the %q channel"), snap.Name, snap.TrackingChannel)
+			case !switchCohort && opts.LeaveCohort && !switchChannel:
+				// TRANSLATORS: %q will be the (quoted) snap name
+				msg = fmt.Sprintf(i18n.G("%q left the cohort"), snap.Name)
+			}
+			fmt.Fprintln(Stdout, msg)
 		default:
 			fmt.Fprintf(Stdout, "internal error: unknown op %q", op)
 		}
-		if snap.TrackingChannel != snap.Channel && snap.Channel != "" {
-			// TRANSLATORS: first %s is a channel name, following %s is a snap name, last %s is a channel name again.
-			fmt.Fprintf(Stdout, i18n.G("Channel %s for %s is closed; temporarily forwarding to %s.\n"), snap.TrackingChannel, snap.Name, snap.Channel)
+		if op == "install" || op == "refresh" {
+			if snap.TrackingChannel != snap.Channel && snap.Channel != "" {
+				if sameRisk, err := isSameRisk(snap.TrackingChannel, snap.Channel); err == nil && !sameRisk {
+					// TRANSLATORS: first %s is a channel name, following %s is a snap name, last %s is a channel name again.
+					fmt.Fprintf(Stdout, i18n.G("Channel %s for %s is closed; temporarily forwarding to %s.\n"), snap.TrackingChannel, snap.Name, snap.Channel)
+				}
+			}
 		}
 	}
 
@@ -300,8 +430,11 @@ type modeMixin struct {
 }
 
 var modeDescs = mixinDescs{
-	"classic":  i18n.G("Put snap in classic mode and disable security confinement"),
-	"devmode":  i18n.G("Put snap in development mode and disable security confinement"),
+	// TRANSLATORS: This should not start with a lowercase letter.
+	"classic": i18n.G("Put snap in classic mode and disable security confinement"),
+	// TRANSLATORS: This should not start with a lowercase letter.
+	"devmode": i18n.G("Put snap in development mode and disable security confinement"),
+	// TRANSLATORS: This should not start with a lowercase letter.
 	"jailmode": i18n.G("Put snap in enforced confinement mode"),
 }
 
@@ -325,6 +458,7 @@ func (mx modeMixin) setModes(opts *client.SnapOptions) {
 }
 
 type cmdInstall struct {
+	colorMixin
 	waitMixin
 
 	channelMixin
@@ -338,25 +472,33 @@ type cmdInstall struct {
 
 	Unaliased bool `long:"unaliased"`
 
-	Positional struct {
+	Name string `long:"name"`
+
+	Cohort        string `long:"cohort"`
+	IgnoreRunning bool   `long:"ignore-running" hidden:"yes"`
+	Positional    struct {
 		Snaps []remoteSnapName `positional-arg-name:"<snap>"`
 	} `positional-args:"yes" required:"yes"`
 }
 
-func (x *cmdInstall) installOne(name string, opts *client.SnapOptions) error {
+func (x *cmdInstall) installOne(nameOrPath, desiredName string, opts *client.SnapOptions) error {
 	var err error
-	var installFromFile bool
 	var changeID string
+	var snapName string
+	var path string
 
-	cli := Client()
-	if strings.Contains(name, "/") || strings.HasSuffix(name, ".snap") || strings.Contains(name, ".snap.") {
-		installFromFile = true
-		changeID, err = cli.InstallPath(name, opts)
+	if strings.Contains(nameOrPath, "/") || strings.HasSuffix(nameOrPath, ".snap") || strings.Contains(nameOrPath, ".snap.") {
+		path = nameOrPath
+		changeID, err = x.client.InstallPath(path, x.Name, opts)
 	} else {
-		changeID, err = cli.Install(name, opts)
+		snapName = nameOrPath
+		if desiredName != "" {
+			return errors.New(i18n.G("cannot use explicit name when installing from store"))
+		}
+		changeID, err = x.client.Install(snapName, opts)
 	}
 	if err != nil {
-		msg, err := errorToCmdMessage(name, err, opts)
+		msg, err := errorToCmdMessage(nameOrPath, err, opts)
 		if err != nil {
 			return err
 		}
@@ -364,7 +506,7 @@ func (x *cmdInstall) installOne(name string, opts *client.SnapOptions) error {
 		return nil
 	}
 
-	chg, err := x.wait(cli, changeID)
+	chg, err := x.wait(changeID)
 	if err != nil {
 		if err == noWait {
 			return nil
@@ -373,16 +515,14 @@ func (x *cmdInstall) installOne(name string, opts *client.SnapOptions) error {
 	}
 
 	// extract the snapName from the change, important for sideloaded
-	var snapName string
-
-	if installFromFile {
+	if path != "" {
 		if err := chg.Get("snap-name", &snapName); err != nil {
-			return fmt.Errorf("cannot extract the snap-name from local file %q: %s", name, err)
+			return fmt.Errorf("cannot extract the snap-name from local file %q: %s", nameOrPath, err)
 		}
-		name = snapName
 	}
 
-	return showDone([]string{name}, "install")
+	// TODO: mention details of the install (e.g. like switch does)
+	return showDone(x.client, []string{snapName}, "install", opts, x.getEscapes())
 }
 
 func (x *cmdInstall) installMany(names []string, opts *client.SnapOptions) error {
@@ -393,8 +533,7 @@ func (x *cmdInstall) installMany(names []string, opts *client.SnapOptions) error
 		}
 	}
 
-	cli := Client()
-	changeID, err := cli.InstallMany(names, opts)
+	changeID, err := x.client.InstallMany(names, opts)
 	if err != nil {
 		var snapName string
 		if err, ok := err.(*client.Error); ok {
@@ -408,7 +547,7 @@ func (x *cmdInstall) installMany(names []string, opts *client.SnapOptions) error
 		return nil
 	}
 
-	chg, err := x.wait(cli, changeID)
+	chg, err := x.wait(changeID)
 	if err != nil {
 		if err == noWait {
 			return nil
@@ -422,7 +561,7 @@ func (x *cmdInstall) installMany(names []string, opts *client.SnapOptions) error
 	}
 
 	if len(installed) > 0 {
-		if err := showDone(installed, "install"); err != nil {
+		if err := showDone(x.client, installed, "install", opts, x.getEscapes()); err != nil {
 			return err
 		}
 	}
@@ -453,26 +592,41 @@ func (x *cmdInstall) Execute([]string) error {
 
 	dangerous := x.Dangerous || x.ForceDangerous
 	opts := &client.SnapOptions{
-		Channel:   x.Channel,
-		Revision:  x.Revision,
-		Dangerous: dangerous,
-		Unaliased: x.Unaliased,
+		Channel:       x.Channel,
+		Revision:      x.Revision,
+		Dangerous:     dangerous,
+		Unaliased:     x.Unaliased,
+		CohortKey:     x.Cohort,
+		IgnoreRunning: x.IgnoreRunning,
 	}
 	x.setModes(opts)
 
 	names := remoteSnapNames(x.Positional.Snaps)
+	if len(names) == 0 {
+		return errors.New(i18n.G("cannot install zero snaps"))
+	}
+	for _, name := range names {
+		if len(name) == 0 {
+			return errors.New(i18n.G("cannot install snap with empty name"))
+		}
+	}
+
 	if len(names) == 1 {
-		return x.installOne(names[0], opts)
+		return x.installOne(names[0], x.Name, opts)
 	}
 
 	if x.asksForMode() || x.asksForChannel() {
 		return errors.New(i18n.G("a single snap name is needed to specify mode or channel flags"))
 	}
 
+	if x.Name != "" {
+		return errors.New(i18n.G("cannot use instance name when installing multiple snaps"))
+	}
 	return x.installMany(names, nil)
 }
 
 type cmdRefresh struct {
+	colorMixin
 	timeMixin
 	waitMixin
 	channelMixin
@@ -480,22 +634,24 @@ type cmdRefresh struct {
 
 	Amend            bool   `long:"amend"`
 	Revision         string `long:"revision"`
+	Cohort           string `long:"cohort"`
+	LeaveCohort      bool   `long:"leave-cohort"`
 	List             bool   `long:"list"`
 	Time             bool   `long:"time"`
 	IgnoreValidation bool   `long:"ignore-validation"`
+	IgnoreRunning    bool   `long:"ignore-running" hidden:"yes"`
 	Positional       struct {
 		Snaps []installedSnapName `positional-arg-name:"<snap>"`
 	} `positional-args:"yes"`
 }
 
 func (x *cmdRefresh) refreshMany(snaps []string, opts *client.SnapOptions) error {
-	cli := Client()
-	changeID, err := cli.RefreshMany(snaps, opts)
+	changeID, err := x.client.RefreshMany(snaps, opts)
 	if err != nil {
 		return err
 	}
 
-	chg, err := x.wait(cli, changeID)
+	chg, err := x.wait(changeID)
 	if err != nil {
 		if err == noWait {
 			return nil
@@ -509,7 +665,7 @@ func (x *cmdRefresh) refreshMany(snaps []string, opts *client.SnapOptions) error
 	}
 
 	if len(refreshed) > 0 {
-		return showDone(refreshed, "refresh")
+		return showDone(x.client, refreshed, "refresh", opts, x.getEscapes())
 	}
 
 	fmt.Fprintln(Stderr, i18n.G("All snaps up to date."))
@@ -518,8 +674,7 @@ func (x *cmdRefresh) refreshMany(snaps []string, opts *client.SnapOptions) error
 }
 
 func (x *cmdRefresh) refreshOne(name string, opts *client.SnapOptions) error {
-	cli := Client()
-	changeID, err := cli.Refresh(name, opts)
+	changeID, err := x.client.Refresh(name, opts)
 	if err != nil {
 		msg, err := errorToCmdMessage(name, err, opts)
 		if err != nil {
@@ -529,19 +684,28 @@ func (x *cmdRefresh) refreshOne(name string, opts *client.SnapOptions) error {
 		return nil
 	}
 
-	if _, err := x.wait(cli, changeID); err != nil {
+	if _, err := x.wait(changeID); err != nil {
 		if err == noWait {
 			return nil
 		}
 		return err
 	}
 
-	return showDone([]string{name}, "refresh")
+	// TODO: this doesn't really tell about all the things you
+	// could set while refreshing (something switch does)
+	return showDone(x.client, []string{name}, "refresh", opts, x.getEscapes())
+}
+
+func parseSysinfoTime(s string) time.Time {
+	t, err := time.Parse(time.RFC3339, s)
+	if err != nil {
+		return time.Time{}
+	}
+	return t
 }
 
 func (x *cmdRefresh) showRefreshTimes() error {
-	cli := Client()
-	sysinfo, err := cli.SysInfo()
+	sysinfo, err := x.client.SysInfo()
 	if err != nil {
 		return err
 	}
@@ -553,17 +717,27 @@ func (x *cmdRefresh) showRefreshTimes() error {
 	} else {
 		return errors.New("internal error: both refresh.timer and refresh.schedule are empty")
 	}
-	if t, err := time.Parse(time.RFC3339, sysinfo.Refresh.Last); err == nil {
-		fmt.Fprintf(Stdout, "last: %s\n", x.fmtTime(t))
+	last := parseSysinfoTime(sysinfo.Refresh.Last)
+	hold := parseSysinfoTime(sysinfo.Refresh.Hold)
+	next := parseSysinfoTime(sysinfo.Refresh.Next)
+
+	if !last.IsZero() {
+		fmt.Fprintf(Stdout, "last: %s\n", x.fmtTime(last))
 	} else {
 		fmt.Fprintf(Stdout, "last: n/a\n")
 	}
-
-	if t, err := time.Parse(time.RFC3339, sysinfo.Refresh.Hold); err == nil {
-		fmt.Fprintf(Stdout, "hold: %s\n", x.fmtTime(t))
+	if !hold.IsZero() {
+		fmt.Fprintf(Stdout, "hold: %s\n", x.fmtTime(hold))
 	}
-	if t, err := time.Parse(time.RFC3339, sysinfo.Refresh.Next); err == nil {
-		fmt.Fprintf(Stdout, "next: %s\n", x.fmtTime(t))
+	// only show "next" if its after "hold" to not confuse users
+	if !next.IsZero() {
+		// Snapstate checks for holdTime.After(limitTime) so we need
+		// to check for before or equal here to be fully correct.
+		if next.Before(hold) || next.Equal(hold) {
+			fmt.Fprintf(Stdout, "next: %s (but held)\n", x.fmtTime(next))
+		} else {
+			fmt.Fprintf(Stdout, "next: %s\n", x.fmtTime(next))
+		}
 	} else {
 		fmt.Fprintf(Stdout, "next: n/a\n")
 	}
@@ -571,8 +745,7 @@ func (x *cmdRefresh) showRefreshTimes() error {
 }
 
 func (x *cmdRefresh) listRefresh() error {
-	cli := Client()
-	snaps, _, err := cli.Find(&client.FindOptions{
+	snaps, _, err := x.client.Find(&client.FindOptions{
 		Refresh: true,
 	})
 	if err != nil {
@@ -585,12 +758,14 @@ func (x *cmdRefresh) listRefresh() error {
 
 	sort.Sort(snapsByName(snaps))
 
+	esc := x.getEscapes()
 	w := tabWriter()
 	defer w.Flush()
 
-	fmt.Fprintln(w, i18n.G("Name\tVersion\tRev\tDeveloper\tNotes"))
+	// TRANSLATORS: the %s is to insert a filler escape sequence (please keep it flush to the column header, with no extra spaces)
+	fmt.Fprintf(w, i18n.G("Name\tVersion\tRev\tSize\tPublisher%s\tNotes\n"), fillerPublisher(esc))
 	for _, snap := range snaps {
-		fmt.Fprintf(w, "%s\t%s\t%s\t%s\t%s\n", snap.Name, snap.Version, snap.Revision, snap.Developer, NotesFromRemote(snap, nil))
+		fmt.Fprintf(w, "%s\t%s\t%s\t%s\t%s\t%s\n", snap.Name, snap.Version, snap.Revision, strutil.SizeToStr(snap.DownloadSize), shortPublisher(esc, snap.Publisher), NotesFromRemote(snap, nil))
 	}
 
 	return nil
@@ -606,14 +781,14 @@ func (x *cmdRefresh) Execute([]string) error {
 
 	if x.Time {
 		if x.asksForMode() || x.asksForChannel() {
-			return errors.New(i18n.G("--time does not take mode nor channel flags"))
+			return errors.New(i18n.G("--time does not take mode or channel flags"))
 		}
 		return x.showRefreshTimes()
 	}
 
 	if x.List {
-		if x.asksForMode() || x.asksForChannel() {
-			return errors.New(i18n.G("--list does not take mode nor channel flags"))
+		if len(x.Positional.Snaps) > 0 || x.asksForMode() || x.asksForChannel() {
+			return errors.New(i18n.G("--list does not accept additional arguments"))
 		}
 
 		return x.listRefresh()
@@ -630,7 +805,10 @@ func (x *cmdRefresh) Execute([]string) error {
 			Amend:            x.Amend,
 			Channel:          x.Channel,
 			IgnoreValidation: x.IgnoreValidation,
+			IgnoreRunning:    x.IgnoreRunning,
 			Revision:         x.Revision,
+			CohortKey:        x.Cohort,
+			LeaveCohort:      x.LeaveCohort,
 		}
 		x.setModes(opts)
 		return x.refreshOne(names[0], opts)
@@ -643,6 +821,9 @@ func (x *cmdRefresh) Execute([]string) error {
 	if x.IgnoreValidation {
 		return errors.New(i18n.G("a single snap name must be specified when ignoring validation"))
 	}
+	if x.IgnoreRunning {
+		return errors.New(i18n.G("a single snap name must be specified when ignoring running apps and hooks"))
+	}
 
 	return x.refreshMany(names, nil)
 }
@@ -674,7 +855,6 @@ func (x *cmdTry) Execute([]string) error {
 	if err := x.validateMode(); err != nil {
 		return err
 	}
-	cli := Client()
 	name := x.Positional.SnapDir
 	opts := &client.SnapOptions{}
 	x.setModes(opts)
@@ -698,17 +878,17 @@ func (x *cmdTry) Execute([]string) error {
 		return fmt.Errorf(i18n.G("cannot get full path for %q: %v"), name, err)
 	}
 
-	changeID, err := cli.Try(path, opts)
-	if e, ok := err.(*client.Error); ok && e.Kind == client.ErrorKindNotSnap {
-		return fmt.Errorf(i18n.G(`%q does not contain an unpacked snap.
-
-Try 'snapcraft prime' in your project directory, then 'snap try' again.`), path)
-	}
+	changeID, err := x.client.Try(path, opts)
 	if err != nil {
-		return err
+		msg, err := errorToCmdMessage(name, err, opts)
+		if err != nil {
+			return err
+		}
+		fmt.Fprintln(Stderr, msg)
+		return nil
 	}
 
-	chg, err := x.wait(cli, changeID)
+	chg, err := x.wait(changeID)
 	if err != nil {
 		if err == noWait {
 			return nil
@@ -725,7 +905,7 @@ Try 'snapcraft prime' in your project directory, then 'snap try' again.`), path)
 	name = snapName
 
 	// show output as speced
-	snaps, err := cli.List([]string{name}, nil)
+	snaps, err := x.client.List([]string{name}, nil)
 	if err != nil {
 		return err
 	}
@@ -748,15 +928,14 @@ type cmdEnable struct {
 }
 
 func (x *cmdEnable) Execute([]string) error {
-	cli := Client()
 	name := string(x.Positional.Snap)
 	opts := &client.SnapOptions{}
-	changeID, err := cli.Enable(name, opts)
+	changeID, err := x.client.Enable(name, opts)
 	if err != nil {
 		return err
 	}
 
-	if _, err := x.wait(cli, changeID); err != nil {
+	if _, err := x.wait(changeID); err != nil {
 		if err == noWait {
 			return nil
 		}
@@ -776,15 +955,14 @@ type cmdDisable struct {
 }
 
 func (x *cmdDisable) Execute([]string) error {
-	cli := Client()
 	name := string(x.Positional.Snap)
 	opts := &client.SnapOptions{}
-	changeID, err := cli.Disable(name, opts)
+	changeID, err := x.client.Disable(name, opts)
 	if err != nil {
 		return err
 	}
 
-	if _, err := x.wait(cli, changeID); err != nil {
+	if _, err := x.wait(changeID); err != nil {
 		if err == noWait {
 			return nil
 		}
@@ -799,8 +977,9 @@ type cmdRevert struct {
 	waitMixin
 
 	modeMixin
-	Revision   string `long:"revision"`
-	Positional struct {
+	Revision      string `long:"revision"`
+	IgnoreRunning bool   `long:"ignore-running" hidden:"yes"`
+	Positional    struct {
 		Snap installedSnapName `positional-arg-name:"<snap>"`
 	} `positional-args:"yes" required:"yes"`
 }
@@ -824,23 +1003,25 @@ func (x *cmdRevert) Execute(args []string) error {
 		return err
 	}
 
-	cli := Client()
 	name := string(x.Positional.Snap)
-	opts := &client.SnapOptions{Revision: x.Revision}
+	opts := &client.SnapOptions{
+		Revision:      x.Revision,
+		IgnoreRunning: x.IgnoreRunning,
+	}
 	x.setModes(opts)
-	changeID, err := cli.Revert(name, opts)
+	changeID, err := x.client.Revert(name, opts)
 	if err != nil {
 		return err
 	}
 
-	if _, err := x.wait(cli, changeID); err != nil {
+	if _, err := x.wait(changeID); err != nil {
 		if err == noWait {
 			return nil
 		}
 		return err
 	}
 
-	return showDone([]string{name}, "revert")
+	return showDone(x.client, []string{name}, "revert", nil, nil)
 }
 
 var shortSwitchHelp = i18n.G("Switches snap to a different channel")
@@ -853,6 +1034,9 @@ type cmdSwitch struct {
 	waitMixin
 	channelMixin
 
+	Cohort      string `long:"cohort"`
+	LeaveCohort bool   `long:"leave-cohort"`
+
 	Positional struct {
 		Snap installedSnapName `positional-arg-name:"<snap>" required:"1"`
 	} `positional-args:"yes" required:"yes"`
@@ -862,55 +1046,101 @@ func (x cmdSwitch) Execute(args []string) error {
 	if err := x.setChannelFromCommandline(); err != nil {
 		return err
 	}
-	if x.Channel == "" {
-		return fmt.Errorf("missing --channel=<channel-name> parameter")
-	}
 
-	cli := Client()
 	name := string(x.Positional.Snap)
 	channel := string(x.Channel)
-	opts := &client.SnapOptions{
-		Channel: channel,
+
+	switchCohort := x.Cohort != ""
+	switchChannel := x.Channel != ""
+
+	// we have three boolean things to check, meaning 2³=8 possibilities
+	// of which 3 are errors (which is why we look at the errors first).
+	// the 5 valid cases are handled by showDone.
+	if switchCohort && x.LeaveCohort {
+		// this one counts as two (no channel filter)
+		return fmt.Errorf(i18n.G("cannot specify both --cohort and --leave-cohort"))
 	}
-	changeID, err := cli.Switch(name, opts)
+	if !switchCohort && !x.LeaveCohort && !switchChannel {
+		return fmt.Errorf(i18n.G("nothing to switch; specify --channel (and/or one of --cohort/--leave-cohort)"))
+	}
+
+	opts := &client.SnapOptions{
+		Channel:     channel,
+		CohortKey:   x.Cohort,
+		LeaveCohort: x.LeaveCohort,
+	}
+	changeID, err := x.client.Switch(name, opts)
 	if err != nil {
 		return err
 	}
 
-	if _, err := x.wait(cli, changeID); err != nil {
+	if _, err := x.wait(changeID); err != nil {
 		if err == noWait {
 			return nil
 		}
 		return err
 	}
 
-	fmt.Fprintf(Stdout, i18n.G("%q switched to the %q channel\n"), name, channel)
-	return nil
+	return showDone(x.client, []string{name}, "switch", opts, nil)
 }
 
 func init() {
 	addCommand("remove", shortRemoveHelp, longRemoveHelp, func() flags.Commander { return &cmdRemove{} },
-		waitDescs.also(map[string]string{"revision": i18n.G("Remove only the given revision")}), nil)
+		waitDescs.also(map[string]string{
+			// TRANSLATORS: This should not start with a lowercase letter.
+			"revision": i18n.G("Remove only the given revision"),
+			// TRANSLATORS: This should not start with a lowercase letter.
+			"purge": i18n.G("Remove the snap without saving a snapshot of its data"),
+		}), nil)
 	addCommand("install", shortInstallHelp, longInstallHelp, func() flags.Commander { return &cmdInstall{} },
-		waitDescs.also(channelDescs).also(modeDescs).also(map[string]string{
-			"revision":        i18n.G("Install the given revision of a snap, to which you must have developer access"),
-			"dangerous":       i18n.G("Install the given snap file even if there are no pre-acknowledged signatures for it, meaning it was not verified and could be dangerous (--devmode implies this)"),
+		colorDescs.also(waitDescs).also(channelDescs).also(modeDescs).also(map[string]string{
+			// TRANSLATORS: This should not start with a lowercase letter.
+			"revision": i18n.G("Install the given revision of a snap, to which you must have developer access"),
+			// TRANSLATORS: This should not start with a lowercase letter.
+			"dangerous": i18n.G("Install the given snap file even if there are no pre-acknowledged signatures for it, meaning it was not verified and could be dangerous (--devmode implies this)"),
+			// TRANSLATORS: This should not start with a lowercase letter.
 			"force-dangerous": i18n.G("Alias for --dangerous (DEPRECATED)"),
-			"unaliased":       i18n.G("Install the given snap without enabling its automatic aliases"),
+			// TRANSLATORS: This should not start with a lowercase letter.
+			"unaliased": i18n.G("Install the given snap without enabling its automatic aliases"),
+			// TRANSLATORS: This should not start with a lowercase letter.
+			"name": i18n.G("Install the snap file under the given instance name"),
+			// TRANSLATORS: This should not start with a lowercase letter.
+			"cohort": i18n.G("Install the snap in the given cohort"),
+			// TRANSLATORS: This should not start with a lowercase letter.
+			"ignore-running": i18n.G("Ignore running hooks or applications blocking the installation"),
 		}), nil)
 	addCommand("refresh", shortRefreshHelp, longRefreshHelp, func() flags.Commander { return &cmdRefresh{} },
-		waitDescs.also(channelDescs).also(modeDescs).also(timeDescs).also(map[string]string{
-			"amend":             i18n.G("Allow refresh attempt on snap unknown to the store"),
-			"revision":          i18n.G("Refresh to the given revision, to which you must have developer access"),
-			"list":              i18n.G("Show available snaps for refresh but do not perform a refresh"),
-			"time":              i18n.G("Show auto refresh information but do not perform a refresh"),
+		colorDescs.also(waitDescs).also(channelDescs).also(modeDescs).also(timeDescs).also(map[string]string{
+			// TRANSLATORS: This should not start with a lowercase letter.
+			"amend": i18n.G("Allow refresh attempt on snap unknown to the store"),
+			// TRANSLATORS: This should not start with a lowercase letter.
+			"revision": i18n.G("Refresh to the given revision, to which you must have developer access"),
+			// TRANSLATORS: This should not start with a lowercase letter.
+			"list": i18n.G("Show the new versions of snaps that would be updated with the next refresh"),
+			// TRANSLATORS: This should not start with a lowercase letter.
+			"time": i18n.G("Show auto refresh information but do not perform a refresh"),
+			// TRANSLATORS: This should not start with a lowercase letter.
 			"ignore-validation": i18n.G("Ignore validation by other snaps blocking the refresh"),
+			// TRANSLATORS: This should not start with a lowercase letter.
+			"ignore-running": i18n.G("Ignore running hooks or applications blocking the refresh"),
+			// TRANSLATORS: This should not start with a lowercase letter.
+			"cohort": i18n.G("Refresh the snap into the given cohort"),
+			// TRANSLATORS: This should not start with a lowercase letter.
+			"leave-cohort": i18n.G("Refresh the snap out of its cohort"),
 		}), nil)
 	addCommand("try", shortTryHelp, longTryHelp, func() flags.Commander { return &cmdTry{} }, waitDescs.also(modeDescs), nil)
 	addCommand("enable", shortEnableHelp, longEnableHelp, func() flags.Commander { return &cmdEnable{} }, waitDescs, nil)
 	addCommand("disable", shortDisableHelp, longDisableHelp, func() flags.Commander { return &cmdDisable{} }, waitDescs, nil)
 	addCommand("revert", shortRevertHelp, longRevertHelp, func() flags.Commander { return &cmdRevert{} }, waitDescs.also(modeDescs).also(map[string]string{
-		"revision": "Revert to the given revision",
+		// TRANSLATORS: This should not start with a lowercase letter.
+		"revision": i18n.G("Revert to the given revision"),
+		// TRANSLATORS: This should not start with a lowercase letter.
+		"ignore-running": i18n.G("Ignore running hooks or applications blocking the revert"),
+	}), nil)
+	addCommand("switch", shortSwitchHelp, longSwitchHelp, func() flags.Commander { return &cmdSwitch{} }, waitDescs.also(channelDescs).also(map[string]string{
+		// TRANSLATORS: This should not start with a lowercase letter.
+		"cohort": i18n.G("Switch the snap into the given cohort"),
+		// TRANSLATORS: This should not start with a lowercase letter.
+		"leave-cohort": i18n.G("Switch the snap out of its cohort"),
 	}), nil)
-	addCommand("switch", shortSwitchHelp, longSwitchHelp, func() flags.Commander { return &cmdSwitch{} }, waitDescs.also(channelDescs), nil)
 }
