@@ -61,7 +61,7 @@ type appsSuite struct {
 	serviceControlError error
 	serviceControlCalls []serviceControlArgs
 
-	infoA, infoB, infoC, infoD *snap.Info
+	infoA, infoB, infoC, infoD, infoE *snap.Info
 }
 
 func (s *appsSuite) journalctl(svcs []string, n int, follow bool) (rc io.ReadCloser, err error) {
@@ -142,10 +142,15 @@ func (s *appsSuite) SetUpTest(c *check.C) {
 	restoreServicestateCtrl := daemon.MockServicestateControl(s.fakeServiceControl)
 	s.AddCleanup(restoreServicestateCtrl)
 
+	// turn off ensuring snap services which will call systemctl automatically
+	r := servicestate.MockEnsuredSnapServices(s.d.Overlord().ServiceManager(), true)
+	s.AddCleanup(r)
+
 	s.infoA = s.mkInstalledInState(c, s.d, "snap-a", "dev", "v1", snap.R(1), true, "apps: {svc1: {daemon: simple}, svc2: {daemon: simple, reload-command: x}}")
 	s.infoB = s.mkInstalledInState(c, s.d, "snap-b", "dev", "v1", snap.R(1), false, "apps: {svc3: {daemon: simple}, cmd1: {}}")
 	s.infoC = s.mkInstalledInState(c, s.d, "snap-c", "dev", "v1", snap.R(1), true, "")
 	s.infoD = s.mkInstalledInState(c, s.d, "snap-d", "dev", "v1", snap.R(1), true, "apps: {cmd2: {}, cmd3: {}}")
+	s.infoE = s.mkInstalledInState(c, s.d, "snap-e", "dev", "v1", snap.R(1), true, "apps: {svc4: {daemon: simple, daemon-scope: user}}")
 
 	d.Overlord().Loop()
 	s.AddCleanup(func() { d.Overlord().Stop() })
@@ -171,7 +176,8 @@ func (s *appsSuite) TestSplitAppName(c *check.C) {
 }
 
 func (s *appsSuite) TestGetAppsInfo(c *check.C) {
-	svcNames := []string{"snap-a.svc1", "snap-a.svc2", "snap-b.svc3"}
+	// System services from active snaps
+	svcNames := []string{"snap-a.svc1", "snap-a.svc2"}
 	for _, name := range svcNames {
 		s.SysctlBufs = append(s.SysctlBufs, []byte(fmt.Sprintf(`
 Id=snap.%s.service
@@ -180,29 +186,39 @@ ActiveState=active
 UnitFileState=enabled
 `[1:], name)))
 	}
+	// System services from inactive snaps
+	svcNames = append(svcNames, "snap-b.svc3")
+	// User services from active snaps
+	svcNames = append(svcNames, "snap-e.svc4")
+	s.SysctlBufs = append(s.SysctlBufs, []byte("enabled\n"))
 
 	req, err := http.NewRequest("GET", "/v2/apps", nil)
 	c.Assert(err, check.IsNil)
 
-	rsp := s.req(c, req, nil).(*daemon.Resp)
+	rsp := s.syncReq(c, req, nil)
 	c.Assert(rsp.Status, check.Equals, 200)
-	c.Assert(rsp.Type, check.Equals, daemon.ResponseTypeSync)
 	c.Assert(rsp.Result, check.FitsTypeOf, []client.AppInfo{})
 	apps := rsp.Result.([]client.AppInfo)
-	c.Assert(apps, check.HasLen, 6)
+	c.Assert(apps, check.HasLen, 7)
 
 	for _, name := range svcNames {
-		snap, app := daemon.SplitAppName(name)
+		snapName, app := daemon.SplitAppName(name)
 		needle := client.AppInfo{
-			Snap:   snap,
-			Name:   app,
-			Daemon: "simple",
+			Snap:        snapName,
+			Name:        app,
+			Daemon:      "simple",
+			DaemonScope: snap.SystemDaemon,
 		}
-		if snap != "snap-b" {
+		if snapName != "snap-b" {
 			// snap-b is not active (all the others are)
 			needle.Active = true
 			needle.Enabled = true
 		}
+		if snapName == "snap-e" {
+			// snap-e contains user services
+			needle.DaemonScope = snap.UserDaemon
+			needle.Active = false
+		}
 		c.Check(apps, testutil.DeepContains, needle)
 	}
 
@@ -226,9 +242,8 @@ func (s *appsSuite) TestGetAppsInfoNames(c *check.C) {
 	req, err := http.NewRequest("GET", "/v2/apps?names=snap-d", nil)
 	c.Assert(err, check.IsNil)
 
-	rsp := s.req(c, req, nil).(*daemon.Resp)
+	rsp := s.syncReq(c, req, nil)
 	c.Assert(rsp.Status, check.Equals, 200)
-	c.Assert(rsp.Type, check.Equals, daemon.ResponseTypeSync)
 	c.Assert(rsp.Result, check.FitsTypeOf, []client.AppInfo{})
 	apps := rsp.Result.([]client.AppInfo)
 	c.Assert(apps, check.HasLen, 2)
@@ -249,7 +264,8 @@ func (s *appsSuite) TestGetAppsInfoNames(c *check.C) {
 }
 
 func (s *appsSuite) TestGetAppsInfoServices(c *check.C) {
-	svcNames := []string{"snap-a.svc1", "snap-a.svc2", "snap-b.svc3"}
+	// System services from active snaps
+	svcNames := []string{"snap-a.svc1", "snap-a.svc2"}
 	for _, name := range svcNames {
 		s.SysctlBufs = append(s.SysctlBufs, []byte(fmt.Sprintf(`
 Id=snap.%s.service
@@ -258,29 +274,39 @@ ActiveState=active
 UnitFileState=enabled
 `[1:], name)))
 	}
+	// System services from inactive snaps
+	svcNames = append(svcNames, "snap-b.svc3")
+	// User services from active snaps
+	svcNames = append(svcNames, "snap-e.svc4")
+	s.SysctlBufs = append(s.SysctlBufs, []byte("enabled\n"))
 
 	req, err := http.NewRequest("GET", "/v2/apps?select=service", nil)
 	c.Assert(err, check.IsNil)
 
-	rsp := s.req(c, req, nil).(*daemon.Resp)
+	rsp := s.syncReq(c, req, nil)
 	c.Assert(rsp.Status, check.Equals, 200)
-	c.Assert(rsp.Type, check.Equals, daemon.ResponseTypeSync)
 	c.Assert(rsp.Result, check.FitsTypeOf, []client.AppInfo{})
 	svcs := rsp.Result.([]client.AppInfo)
-	c.Assert(svcs, check.HasLen, 3)
+	c.Assert(svcs, check.HasLen, 4)
 
 	for _, name := range svcNames {
-		snap, app := daemon.SplitAppName(name)
+		snapName, app := daemon.SplitAppName(name)
 		needle := client.AppInfo{
-			Snap:   snap,
-			Name:   app,
-			Daemon: "simple",
+			Snap:        snapName,
+			Name:        app,
+			Daemon:      "simple",
+			DaemonScope: snap.SystemDaemon,
 		}
-		if snap != "snap-b" {
+		if snapName != "snap-b" {
 			// snap-b is not active (all the others are)
 			needle.Active = true
 			needle.Enabled = true
 		}
+		if snapName == "snap-e" {
+			// snap-e contains user services
+			needle.DaemonScope = snap.UserDaemon
+			needle.Active = false
+		}
 		c.Check(svcs, testutil.DeepContains, needle)
 	}
 
@@ -295,24 +321,22 @@ func (s *appsSuite) TestGetAppsInfoBadSelect(c *check.C) {
 	req, err := http.NewRequest("GET", "/v2/apps?select=potato", nil)
 	c.Assert(err, check.IsNil)
 
-	rsp := s.req(c, req, nil).(*daemon.Resp)
-	c.Assert(rsp.Status, check.Equals, 400)
-	c.Assert(rsp.Type, check.Equals, daemon.ResponseTypeError)
+	rspe := s.errorReq(c, req, nil)
+	c.Assert(rspe.Status, check.Equals, 400)
 }
 
 func (s *appsSuite) TestGetAppsInfoBadName(c *check.C) {
 	req, err := http.NewRequest("GET", "/v2/apps?names=potato", nil)
 	c.Assert(err, check.IsNil)
 
-	rsp := s.req(c, req, nil).(*daemon.Resp)
-	c.Assert(rsp.Status, check.Equals, 404)
-	c.Assert(rsp.Type, check.Equals, daemon.ResponseTypeError)
+	rspe := s.errorReq(c, req, nil)
+	c.Assert(rspe.Status, check.Equals, 404)
 }
 
 func (s *appsSuite) TestAppInfosForOne(c *check.C) {
 	st := s.d.Overlord().State()
-	appInfos, rsp := daemon.AppInfosFor(st, []string{"snap-a.svc1"}, daemon.AppInfoServiceTrue)
-	c.Assert(rsp, check.IsNil)
+	appInfos, rspe := daemon.AppInfosFor(st, []string{"snap-a.svc1"}, daemon.AppInfoServiceTrue)
+	c.Assert(rspe, check.IsNil)
 	c.Assert(appInfos, check.HasLen, 1)
 	c.Check(appInfos[0].Snap, check.DeepEquals, s.infoA)
 	c.Check(appInfos[0].Name, check.Equals, "svc1")
@@ -328,20 +352,20 @@ func (s *appsSuite) TestAppInfosForAll(c *check.C) {
 	for _, t := range []T{
 		{
 			opts:  daemon.AppInfoServiceTrue,
-			names: []string{"svc1", "svc2", "svc3"},
-			snaps: []*snap.Info{s.infoA, s.infoA, s.infoB},
+			names: []string{"svc1", "svc2", "svc3", "svc4"},
+			snaps: []*snap.Info{s.infoA, s.infoA, s.infoB, s.infoE},
 		},
 		{
 			opts:  daemon.AppInfoServiceFalse,
-			names: []string{"svc1", "svc2", "cmd1", "svc3", "cmd2", "cmd3"},
-			snaps: []*snap.Info{s.infoA, s.infoA, s.infoB, s.infoB, s.infoD, s.infoD},
+			names: []string{"svc1", "svc2", "cmd1", "svc3", "cmd2", "cmd3", "svc4"},
+			snaps: []*snap.Info{s.infoA, s.infoA, s.infoB, s.infoB, s.infoD, s.infoD, s.infoE},
 		},
 	} {
 		c.Assert(len(t.names), check.Equals, len(t.snaps), check.Commentf("%s", t.opts))
 
 		st := s.d.Overlord().State()
-		appInfos, rsp := daemon.AppInfosFor(st, nil, t.opts)
-		c.Assert(rsp, check.IsNil, check.Commentf("%s", t.opts))
+		appInfos, rspe := daemon.AppInfosFor(st, nil, t.opts)
+		c.Assert(rspe, check.IsNil, check.Commentf("%s", t.opts))
 		names := make([]string, len(appInfos))
 		for i, appInfo := range appInfos {
 			names[i] = appInfo.Name
@@ -356,8 +380,8 @@ func (s *appsSuite) TestAppInfosForAll(c *check.C) {
 
 func (s *appsSuite) TestAppInfosForOneSnap(c *check.C) {
 	st := s.d.Overlord().State()
-	appInfos, rsp := daemon.AppInfosFor(st, []string{"snap-a"}, daemon.AppInfoServiceTrue)
-	c.Assert(rsp, check.IsNil)
+	appInfos, rspe := daemon.AppInfosFor(st, []string{"snap-a"}, daemon.AppInfoServiceTrue)
+	c.Assert(rspe, check.IsNil)
 	c.Assert(appInfos, check.HasLen, 2)
 	sort.Sort(snap.AppInfoBySnapApp(appInfos))
 
@@ -369,8 +393,8 @@ func (s *appsSuite) TestAppInfosForOneSnap(c *check.C) {
 
 func (s *appsSuite) TestAppInfosForMixedArgs(c *check.C) {
 	st := s.d.Overlord().State()
-	appInfos, rsp := daemon.AppInfosFor(st, []string{"snap-a", "snap-a.svc1"}, daemon.AppInfoServiceTrue)
-	c.Assert(rsp, check.IsNil)
+	appInfos, rspe := daemon.AppInfosFor(st, []string{"snap-a", "snap-a.svc1"}, daemon.AppInfoServiceTrue)
+	c.Assert(rspe, check.IsNil)
 	c.Assert(appInfos, check.HasLen, 2)
 	sort.Sort(snap.AppInfoBySnapApp(appInfos))
 
@@ -382,7 +406,7 @@ func (s *appsSuite) TestAppInfosForMixedArgs(c *check.C) {
 
 func (s *appsSuite) TestAppInfosCleanupAndSorted(c *check.C) {
 	st := s.d.Overlord().State()
-	appInfos, rsp := daemon.AppInfosFor(st, []string{
+	appInfos, rspe := daemon.AppInfosFor(st, []string{
 		"snap-b.svc3",
 		"snap-a.svc2",
 		"snap-a.svc1",
@@ -392,7 +416,7 @@ func (s *appsSuite) TestAppInfosCleanupAndSorted(c *check.C) {
 		"snap-b",
 		"snap-a",
 	}, daemon.AppInfoServiceTrue)
-	c.Assert(rsp, check.IsNil)
+	c.Assert(rspe, check.IsNil)
 	c.Assert(appInfos, check.HasLen, 3)
 	sort.Sort(snap.AppInfoBySnapApp(appInfos))
 
@@ -406,28 +430,28 @@ func (s *appsSuite) TestAppInfosCleanupAndSorted(c *check.C) {
 
 func (s *appsSuite) TestAppInfosForAppless(c *check.C) {
 	st := s.d.Overlord().State()
-	appInfos, rsp := daemon.AppInfosFor(st, []string{"snap-c"}, daemon.AppInfoServiceTrue)
-	c.Assert(rsp, check.FitsTypeOf, &daemon.Resp{})
-	c.Check(rsp.(*daemon.Resp).Status, check.Equals, 404)
-	c.Check(rsp.(*daemon.Resp).Result.(*daemon.ErrorResult).Kind, check.Equals, client.ErrorKindAppNotFound)
+	appInfos, rspe := daemon.AppInfosFor(st, []string{"snap-c"}, daemon.AppInfoServiceTrue)
+	c.Assert(rspe, check.NotNil)
+	c.Check(rspe.Status, check.Equals, 404)
+	c.Check(rspe.Kind, check.Equals, client.ErrorKindAppNotFound)
 	c.Assert(appInfos, check.IsNil)
 }
 
 func (s *appsSuite) TestAppInfosForMissingApp(c *check.C) {
 	st := s.d.Overlord().State()
-	appInfos, rsp := daemon.AppInfosFor(st, []string{"snap-c.whatever"}, daemon.AppInfoServiceTrue)
-	c.Assert(rsp, check.FitsTypeOf, &daemon.Resp{})
-	c.Check(rsp.(*daemon.Resp).Status, check.Equals, 404)
-	c.Check(rsp.(*daemon.Resp).Result.(*daemon.ErrorResult).Kind, check.Equals, client.ErrorKindAppNotFound)
+	appInfos, rspe := daemon.AppInfosFor(st, []string{"snap-c.whatever"}, daemon.AppInfoServiceTrue)
+	c.Assert(rspe, check.NotNil)
+	c.Check(rspe.Status, check.Equals, 404)
+	c.Check(rspe.Kind, check.Equals, client.ErrorKindAppNotFound)
 	c.Assert(appInfos, check.IsNil)
 }
 
 func (s *appsSuite) TestAppInfosForMissingSnap(c *check.C) {
 	st := s.d.Overlord().State()
-	appInfos, rsp := daemon.AppInfosFor(st, []string{"snap-x"}, daemon.AppInfoServiceTrue)
-	c.Assert(rsp, check.FitsTypeOf, &daemon.Resp{})
-	c.Check(rsp.(*daemon.Resp).Status, check.Equals, 404)
-	c.Check(rsp.(*daemon.Resp).Result.(*daemon.ErrorResult).Kind, check.Equals, client.ErrorKindSnapNotFound)
+	appInfos, rspe := daemon.AppInfosFor(st, []string{"snap-x"}, daemon.AppInfoServiceTrue)
+	c.Assert(rspe, check.NotNil)
+	c.Check(rspe.Status, check.Equals, 404)
+	c.Check(rspe.Kind, check.Equals, client.ErrorKindSnapNotFound)
 	c.Assert(appInfos, check.IsNil)
 }
 
@@ -438,9 +462,8 @@ func (s *appsSuite) testPostApps(c *check.C, inst servicestate.Instruction, serv
 	req, err := http.NewRequest("POST", "/v2/apps", bytes.NewBuffer(postBody))
 	c.Assert(err, check.IsNil)
 
-	rsp := s.req(c, req, nil).(*daemon.Resp)
+	rsp := s.asyncReq(c, req, nil)
 	c.Assert(rsp.Status, check.Equals, 202)
-	c.Assert(rsp.Type, check.Equals, daemon.ResponseTypeAsync)
 	c.Check(rsp.Change, check.Matches, `[0-9]+`)
 
 	st := s.d.Overlord().State()
@@ -552,60 +575,60 @@ func (s *appsSuite) TestPostAppsDisableNow(c *check.C) {
 func (s *appsSuite) TestPostAppsBadJSON(c *check.C) {
 	req, err := http.NewRequest("POST", "/v2/apps", bytes.NewBufferString(`'junk`))
 	c.Assert(err, check.IsNil)
-	rsp := s.req(c, req, nil).(*daemon.Resp)
-	c.Check(rsp.Status, check.Equals, 400)
-	c.Check(rsp.Type, check.Equals, daemon.ResponseTypeError)
-	c.Check(rsp.Result.(*daemon.ErrorResult).Message, check.Matches, ".*cannot decode request body.*")
+	rspe := s.errorReq(c, req, nil)
+	c.Check(rspe.Status, check.Equals, 400)
+	c.Check(rspe.Message, check.Matches, ".*cannot decode request body.*")
 }
 
 func (s *appsSuite) TestPostAppsBadOp(c *check.C) {
 	req, err := http.NewRequest("POST", "/v2/apps", bytes.NewBufferString(`{"random": "json"}`))
 	c.Assert(err, check.IsNil)
-	rsp := s.req(c, req, nil).(*daemon.Resp)
-	c.Check(rsp.Status, check.Equals, 400)
-	c.Check(rsp.Type, check.Equals, daemon.ResponseTypeError)
-	c.Check(rsp.Result.(*daemon.ErrorResult).Message, check.Matches, ".*cannot perform operation on services without a list of services.*")
+	rspe := s.errorReq(c, req, nil)
+	c.Check(rspe.Status, check.Equals, 400)
+	c.Check(rspe.Message, check.Matches, ".*cannot perform operation on services without a list of services.*")
 }
 
 func (s *appsSuite) TestPostAppsBadSnap(c *check.C) {
 	req, err := http.NewRequest("POST", "/v2/apps", bytes.NewBufferString(`{"action": "stop", "names": ["snap-c"]}`))
 	c.Assert(err, check.IsNil)
-	rsp := s.req(c, req, nil).(*daemon.Resp)
-	c.Check(rsp.Status, check.Equals, 404)
-	c.Check(rsp.Type, check.Equals, daemon.ResponseTypeError)
-	c.Check(rsp.Result.(*daemon.ErrorResult).Message, check.Equals, `snap "snap-c" has no services`)
+	rspe := s.errorReq(c, req, nil)
+	c.Check(rspe.Status, check.Equals, 404)
+	c.Check(rspe.Message, check.Equals, `snap "snap-c" has no services`)
 }
 
 func (s *appsSuite) TestPostAppsBadApp(c *check.C) {
 	req, err := http.NewRequest("POST", "/v2/apps", bytes.NewBufferString(`{"action": "stop", "names": ["snap-a.what"]}`))
 	c.Assert(err, check.IsNil)
-	rsp := s.req(c, req, nil).(*daemon.Resp)
-	c.Check(rsp.Status, check.Equals, 404)
-	c.Check(rsp.Type, check.Equals, daemon.ResponseTypeError)
-	c.Check(rsp.Result.(*daemon.ErrorResult).Message, check.Equals, `snap "snap-a" has no service "what"`)
+	rspe := s.errorReq(c, req, nil)
+	c.Check(rspe.Status, check.Equals, 404)
+	c.Check(rspe.Message, check.Equals, `snap "snap-a" has no service "what"`)
 }
 
 func (s *appsSuite) TestPostAppsServiceControlError(c *check.C) {
 	req, err := http.NewRequest("POST", "/v2/apps", bytes.NewBufferString(`{"action": "start", "names": ["snap-a.svc1"]}`))
 	c.Assert(err, check.IsNil)
 	s.serviceControlError = fmt.Errorf("total failure")
-	rsp := s.req(c, req, nil).(*daemon.Resp)
-	c.Check(rsp.Status, check.Equals, 400)
-	c.Check(rsp.Type, check.Equals, daemon.ResponseTypeError)
-	c.Check(rsp.Result.(*daemon.ErrorResult).Message, check.Equals, `total failure`)
+	rspe := s.errorReq(c, req, nil)
+	c.Check(rspe.Status, check.Equals, 400)
+	c.Check(rspe.Message, check.Equals, `total failure`)
 }
 
 func (s *appsSuite) TestPostAppsConflict(c *check.C) {
 	req, err := http.NewRequest("POST", "/v2/apps", bytes.NewBufferString(`{"action": "start", "names": ["snap-a.svc1"]}`))
 	c.Assert(err, check.IsNil)
 	s.serviceControlError = &snapstate.ChangeConflictError{Snap: "snap-a", ChangeKind: "enable"}
-	rsp := s.req(c, req, nil).(*daemon.Resp)
-	c.Check(rsp.Status, check.Equals, 400)
-	c.Check(rsp.Type, check.Equals, daemon.ResponseTypeError)
-	c.Check(rsp.Result.(*daemon.ErrorResult).Message, check.Equals, `snap "snap-a" has "enable" change in progress`)
+	rspe := s.errorReq(c, req, nil)
+	c.Check(rspe.Status, check.Equals, 400)
+	c.Check(rspe.Message, check.Equals, `snap "snap-a" has "enable" change in progress`)
+}
+
+func (s *appsSuite) expectLogsAccess() {
+	s.expectReadAccess(daemon.AuthenticatedAccess{Polkit: "io.snapcraft.snapd.manage"})
 }
 
 func (s *appsSuite) TestLogs(c *check.C) {
+	s.expectLogsAccess()
+
 	s.jctlRCs = []io.ReadCloser{ioutil.NopCloser(strings.NewReader(`
 {"MESSAGE": "hello1", "SYSLOG_IDENTIFIER": "xyzzy", "_PID": "42", "__REALTIME_TIMESTAMP": "42"}
 {"MESSAGE": "hello2", "SYSLOG_IDENTIFIER": "xyzzy", "_PID": "42", "__REALTIME_TIMESTAMP": "44"}
@@ -636,6 +659,8 @@ func (s *appsSuite) TestLogs(c *check.C) {
 }
 
 func (s *appsSuite) TestLogsN(c *check.C) {
+	s.expectLogsAccess()
+
 	type T struct {
 		in  string
 		out int
@@ -663,15 +688,18 @@ func (s *appsSuite) TestLogsN(c *check.C) {
 }
 
 func (s *appsSuite) TestLogsBadN(c *check.C) {
+	s.expectLogsAccess()
+
 	req, err := http.NewRequest("GET", "/v2/logs?n=hello", nil)
 	c.Assert(err, check.IsNil)
 
-	rsp := s.req(c, req, nil).(*daemon.Resp)
-	c.Assert(rsp.Status, check.Equals, 400)
-	c.Assert(rsp.Type, check.Equals, daemon.ResponseTypeError)
+	rspe := s.errorReq(c, req, nil)
+	c.Assert(rspe.Status, check.Equals, 400)
 }
 
 func (s *appsSuite) TestLogsFollow(c *check.C) {
+	s.expectLogsAccess()
+
 	s.jctlRCs = []io.ReadCloser{
 		ioutil.NopCloser(strings.NewReader("")),
 		ioutil.NopCloser(strings.NewReader("")),
@@ -694,29 +722,48 @@ func (s *appsSuite) TestLogsFollow(c *check.C) {
 }
 
 func (s *appsSuite) TestLogsBadFollow(c *check.C) {
+	s.expectLogsAccess()
+
 	req, err := http.NewRequest("GET", "/v2/logs?follow=hello", nil)
 	c.Assert(err, check.IsNil)
 
-	rsp := s.req(c, req, nil).(*daemon.Resp)
-	c.Assert(rsp.Status, check.Equals, 400)
-	c.Assert(rsp.Type, check.Equals, daemon.ResponseTypeError)
+	rspe := s.errorReq(c, req, nil)
+	c.Assert(rspe.Status, check.Equals, 400)
 }
 
 func (s *appsSuite) TestLogsBadName(c *check.C) {
+	s.expectLogsAccess()
+
 	req, err := http.NewRequest("GET", "/v2/logs?names=hello", nil)
 	c.Assert(err, check.IsNil)
 
-	rsp := s.req(c, req, nil).(*daemon.Resp)
-	c.Assert(rsp.Status, check.Equals, 404)
-	c.Assert(rsp.Type, check.Equals, daemon.ResponseTypeError)
+	rspe := s.errorReq(c, req, nil)
+	c.Assert(rspe.Status, check.Equals, 404)
 }
 
 func (s *appsSuite) TestLogsSad(c *check.C) {
+	s.expectLogsAccess()
+
 	s.jctlErrs = []error{errors.New("potato")}
 	req, err := http.NewRequest("GET", "/v2/logs", nil)
 	c.Assert(err, check.IsNil)
 
-	rsp := s.req(c, req, nil).(*daemon.Resp)
-	c.Assert(rsp.Status, check.Equals, 500)
-	c.Assert(rsp.Type, check.Equals, daemon.ResponseTypeError)
+	rspe := s.errorReq(c, req, nil)
+	c.Assert(rspe.Status, check.Equals, 500)
+}
+
+func (s *appsSuite) TestLogsNoServices(c *check.C) {
+	s.expectLogsAccess()
+
+	// no installed snaps with services
+	st := s.d.Overlord().State()
+	st.Lock()
+	st.Set("snaps", nil)
+	st.Unlock()
+
+	req, err := http.NewRequest("GET", "/v2/logs", nil)
+	c.Assert(err, check.IsNil)
+
+	rspe := s.errorReq(c, req, nil)
+	c.Assert(rspe.Status, check.Equals, 404)
 }
