@@ -163,18 +163,27 @@ func (b *Backend) Initialize(opts *interfaces.SecurityBackendOptions) error {
 	// not.  If we do then we prefer the file ending with the name .real as
 	// that is the more recent name we use.
 	var profilePath string
+	// TODO: fix for distros using /usr/libexec/snapd
 	for _, profileFname := range []string{"usr.lib.snapd.snap-confine.real", "usr.lib.snapd.snap-confine"} {
-		profilePath = filepath.Join(apparmor_sandbox.ConfDir, profileFname)
-		if _, err := os.Stat(profilePath); err != nil {
+		maybeProfilePath := filepath.Join(apparmor_sandbox.ConfDir, profileFname)
+		if _, err := os.Stat(maybeProfilePath); err != nil {
 			if os.IsNotExist(err) {
 				continue
 			}
 			return err
 		}
+		profilePath = maybeProfilePath
 		break
 	}
 	if profilePath == "" {
-		return fmt.Errorf("cannot find system apparmor profile for snap-confine")
+		// XXX: is profile mandatory on some distros?
+
+		// There is no AppArmor profile for snap-confine, quite likely
+		// AppArmor support is enabled in the kernel and relevant
+		// userspace tools exist, but snap-confine was built without it,
+		// nothing we need to update then.
+		logger.Noticef("snap-confine apparmor profile is absent, nothing to update")
+		return nil
 	}
 
 	aaFlags := skipReadCache
@@ -196,6 +205,9 @@ func (b *Backend) Initialize(opts *interfaces.SecurityBackendOptions) error {
 // snapConfineFromSnapProfile returns the apparmor profile for
 // snap-confine in the given core/snapd snap.
 func snapConfineFromSnapProfile(info *snap.Info) (dir, glob string, content map[string]osutil.FileState, err error) {
+	// TODO: fix this for distros using /usr/libexec/snapd when those start
+	// to use reexec
+
 	// Find the vanilla apparmor profile for snap-confine as present in the given core snap.
 
 	// We must test the ".real" suffix first, this is a workaround for
@@ -403,10 +415,8 @@ func (b *Backend) prepareProfiles(snapInfo *snap.Info, opts interfaces.Confineme
 	}
 
 	// Get the files that this snap should have
-	content, err := b.deriveContent(spec.(*Specification), snapInfo, opts)
-	if err != nil {
-		return nil, fmt.Errorf("cannot obtain expected security files for snap %q: %s", snapName, err)
-	}
+	content := b.deriveContent(spec.(*Specification), snapInfo, opts)
+
 	dir := dirs.SnapAppArmorDir
 	globs := profileGlobs(snapInfo.InstanceName())
 	if err := os.MkdirAll(dir, 0755); err != nil {
@@ -564,6 +574,7 @@ func (b *Backend) Remove(snapName string) error {
 	globs := profileGlobs(snapName)
 	cache := apparmor_sandbox.CacheDir
 	_, removed, errEnsure := osutil.EnsureDirStateGlobs(dir, globs, nil)
+	// always try to unload affected profiles
 	errUnload := unloadProfiles(removed, cache)
 	if errEnsure != nil {
 		return fmt.Errorf("cannot synchronize security files for snap %q: %s", snapName, errEnsure)
@@ -572,7 +583,7 @@ func (b *Backend) Remove(snapName string) error {
 }
 
 func (b *Backend) RemoveLate(snapName string, rev snap.Revision, typ snap.Type) error {
-	logger.Noticef("remove late for snap %v (%s) type %v", snapName, rev, typ)
+	logger.Debugf("remove late for snap %v (%s) type %v", snapName, rev, typ)
 	if typ != snap.TypeSnapd {
 		// late remove is relevant only for snap confine profiles
 		return nil
@@ -582,6 +593,7 @@ func (b *Backend) RemoveLate(snapName string, rev snap.Revision, typ snap.Type)
 	_, removed, errEnsure := osutil.EnsureDirStateGlobs(dirs.SnapAppArmorDir, globs, nil)
 	// XXX: unloadProfiles() does not unload profiles from the kernel, but
 	// only removes profiles from the cache
+	// always try to unload the affected profile
 	errUnload := unloadProfiles(removed, apparmor_sandbox.CacheDir)
 	if errEnsure != nil {
 		return fmt.Errorf("cannot remove security profiles for snap %q (%s): %s", snapName, rev, errEnsure)
@@ -599,7 +611,7 @@ const (
 	attachComplain = "(attach_disconnected,mediate_deleted,complain)"
 )
 
-func (b *Backend) deriveContent(spec *Specification, snapInfo *snap.Info, opts interfaces.ConfinementOptions) (content map[string]osutil.FileState, err error) {
+func (b *Backend) deriveContent(spec *Specification, snapInfo *snap.Info, opts interfaces.ConfinementOptions) (content map[string]osutil.FileState) {
 	content = make(map[string]osutil.FileState, len(snapInfo.Apps)+len(snapInfo.Hooks)+1)
 
 	// Add profile for each app.
@@ -617,10 +629,10 @@ func (b *Backend) deriveContent(spec *Specification, snapInfo *snap.Info, opts i
 	// This applies to, for example, kernel snaps or gadget snaps (unless they have hooks).
 	if len(content) > 0 {
 		snippets := strings.Join(spec.UpdateNS(), "\n")
-		addUpdateNSProfile(snapInfo, opts, snippets, content)
+		addUpdateNSProfile(snapInfo, snippets, content)
 	}
 
-	return content, nil
+	return content
 }
 
 // addUpdateNSProfile adds an apparmor profile for snap-update-ns, tailored to a specific snap.
@@ -628,7 +640,7 @@ func (b *Backend) deriveContent(spec *Specification, snapInfo *snap.Info, opts i
 // This profile exists so that snap-update-ns doens't need to carry very wide, open permissions
 // that are suitable for poking holes (and writing) in nearly arbitrary places. Instead the profile
 // contains just the permissions needed to poke a hole and write to the layout-specific paths.
-func addUpdateNSProfile(snapInfo *snap.Info, opts interfaces.ConfinementOptions, snippets string, content map[string]osutil.FileState) {
+func addUpdateNSProfile(snapInfo *snap.Info, snippets string, content map[string]osutil.FileState) {
 	// Compute the template by injecting special updateNS snippets.
 	policy := templatePattern.ReplaceAllStringFunc(updateNSTemplate, func(placeholder string) string {
 		switch placeholder {
