@@ -169,6 +169,8 @@ func (s *backendSuite) SetUpTest(c *C) {
 	s.parserCmd = testutil.MockCommand(c, "apparmor_parser", fakeAppArmorParser)
 
 	apparmor.MockRuntimeNumCPU(func() int { return 99 })
+	restore := release.MockReleaseInfo(&release.OS{ID: "ubuntu"})
+	s.AddCleanup(restore)
 }
 
 func (s *backendSuite) TearDownTest(c *C) {
@@ -1076,36 +1078,6 @@ profile "snap.foo.hook.configure" (attach_disconnected,mediate_deleted) {
 	s.RemoveSnap(c, snapInfo)
 }
 
-func mockPartalAppArmorOnDistro(c *C, kernelVersion string, releaseID string, releaseIDLike ...string) (restore func()) {
-	restore1 := apparmor_sandbox.MockLevel(apparmor_sandbox.Partial)
-	restore2 := release.MockReleaseInfo(&release.OS{ID: releaseID, IDLike: releaseIDLike})
-	restore3 := osutil.MockKernelVersion(kernelVersion)
-	restore4 := apparmor.MockIsHomeUsingNFS(func() (bool, error) { return false, nil })
-	restore5 := apparmor.MockIsRootWritableOverlay(func() (string, error) { return "", nil })
-	// Replace the real template with a shorter variant that is easier to test.
-	restore6 := apparmor.MockTemplate("\n" +
-		"###VAR###\n" +
-		"###PROFILEATTACH### (attach_disconnected) {\n" +
-		"###SNIPPETS###\n" +
-		"}\n")
-	restore7 := apparmor.MockClassicTemplate("\n" +
-		"#classic\n" +
-		"###VAR###\n" +
-		"###PROFILEATTACH### (attach_disconnected) {\n" +
-		"###SNIPPETS###\n" +
-		"}\n")
-
-	return func() {
-		restore1()
-		restore2()
-		restore3()
-		restore4()
-		restore5()
-		restore6()
-		restore7()
-	}
-}
-
 const coreYaml = `name: core
 version: 1
 type: os
@@ -1279,6 +1251,8 @@ func (s *backendSuite) TestSnapConfineProfileDiscardedLateSnapd(c *C) {
 	c.Assert(err, IsNil)
 	// sanity
 	c.Assert(filepath.Join(dirs.SnapAppArmorDir, "snap-confine.snapd.222"), testutil.FilePresent)
+	// place a canary
+	c.Assert(ioutil.WriteFile(filepath.Join(dirs.SnapAppArmorDir, "snap-confine.snapd.111"), nil, 0644), IsNil)
 
 	// backed implements the right interface
 	late, ok := s.Backend.(interfaces.SecurityBackendDiscardingLate)
@@ -1286,6 +1260,8 @@ func (s *backendSuite) TestSnapConfineProfileDiscardedLateSnapd(c *C) {
 	err = late.RemoveLate(snapdInfo.InstanceName(), snapdInfo.Revision, snapdInfo.Type())
 	c.Assert(err, IsNil)
 	c.Check(filepath.Join(dirs.SnapAppArmorDir, "snap-confine.snapd.222"), testutil.FileAbsent)
+	// but the canary is still present
+	c.Assert(filepath.Join(dirs.SnapAppArmorDir, "snap-confine.snapd.111"), testutil.FilePresent)
 }
 
 func (s *backendSuite) TestCoreOnCoreCleansApparmorCache(c *C) {
@@ -1397,6 +1373,28 @@ func (s *backendSuite) TestSetupSnapConfineGeneratedPolicyWithNFS2(c *C) {
 	s.testSetupSnapConfineGeneratedPolicyWithNFS(c, "usr.lib.snapd.snap-confine.real")
 }
 
+func (s *backendSuite) TestSetupSnapConfineGeneratedPolicyWithNFSNoProfileFiles(c *C) {
+	// Make it appear as if NFS workaround was needed.
+	restore := apparmor.MockIsHomeUsingNFS(func() (bool, error) { return true, nil })
+	defer restore()
+	// Make it appear as if overlay was not used.
+	restore = apparmor.MockIsRootWritableOverlay(func() (string, error) { return "", nil })
+	defer restore()
+
+	// Intercept interaction with apparmor_parser
+	cmd := testutil.MockCommand(c, "apparmor_parser", "")
+	defer cmd.Restore()
+	// Set up apparmor profiles directory, but no profile for snap-confine
+	c.Assert(os.MkdirAll(apparmor_sandbox.ConfDir, 0755), IsNil)
+
+	// The apparmor backend should not fail if the apparmor profile of
+	// snap-confine is not present
+	err := (&apparmor.Backend{}).Initialize(nil)
+	c.Assert(err, IsNil)
+	// Since there is no profile file, no call to apparmor were made
+	c.Assert(cmd.Calls(), HasLen, 0)
+}
+
 // snap-confine policy when NFS is used and snapd has not re-executed.
 func (s *backendSuite) testSetupSnapConfineGeneratedPolicyWithNFS(c *C, profileFname string) {
 	// Make it appear as if NFS workaround was needed.
@@ -1722,6 +1720,28 @@ func (s *backendSuite) TestSetupSnapConfineGeneratedPolicyWithOverlay2(c *C) {
 	s.testSetupSnapConfineGeneratedPolicyWithOverlay(c, "usr.lib.snapd.snap-confine.real")
 }
 
+func (s *backendSuite) TestSetupSnapConfineGeneratedPolicyWithOverlayNoProfileFiles(c *C) {
+	// Make it appear as if overlay workaround was needed.
+	restore := apparmor.MockIsRootWritableOverlay(func() (string, error) { return "/upper", nil })
+	defer restore()
+	// No NFS workaround
+	restore = apparmor.MockIsHomeUsingNFS(func() (bool, error) { return false, nil })
+	defer restore()
+
+	// Intercept interaction with apparmor_parser
+	cmd := testutil.MockCommand(c, "apparmor_parser", "")
+	defer cmd.Restore()
+	// Set up apparmor profiles directory, but no profile for snap-confine
+	c.Assert(os.MkdirAll(apparmor_sandbox.ConfDir, 0755), IsNil)
+
+	// The apparmor backend should not fail if the apparmor profile of
+	// snap-confine is not present
+	err := (&apparmor.Backend{}).Initialize(nil)
+	c.Assert(err, IsNil)
+	// Since there is no profile file, no call to apparmor were made
+	c.Assert(cmd.Calls(), HasLen, 0)
+}
+
 // snap-confine policy when overlay is used and snapd has not re-executed.
 func (s *backendSuite) testSetupSnapConfineGeneratedPolicyWithOverlay(c *C, profileFname string) {
 	// Make it appear as if overlay workaround was needed.
