@@ -20,8 +20,8 @@
 package builtin_test
 
 import (
-	"fmt"
 	"path/filepath"
+	"strings"
 
 	. "gopkg.in/check.v1"
 
@@ -225,13 +225,19 @@ apps:
 
 func (s *ContentSuite) TestResolveSpecialVariable(c *C) {
 	info := snaptest.MockInfo(c, "{name: name, version: 0}", &snap.SideInfo{Revision: snap.R(42)})
-	c.Check(builtin.ResolveSpecialVariable("foo", info), Equals, filepath.Join(dirs.CoreSnapMountDir, "name/42/foo"))
 	c.Check(builtin.ResolveSpecialVariable("$SNAP/foo", info), Equals, filepath.Join(dirs.CoreSnapMountDir, "name/42/foo"))
 	c.Check(builtin.ResolveSpecialVariable("$SNAP_DATA/foo", info), Equals, "/var/snap/name/42/foo")
 	c.Check(builtin.ResolveSpecialVariable("$SNAP_COMMON/foo", info), Equals, "/var/snap/name/common/foo")
 	c.Check(builtin.ResolveSpecialVariable("$SNAP", info), Equals, filepath.Join(dirs.CoreSnapMountDir, "name/42"))
 	c.Check(builtin.ResolveSpecialVariable("$SNAP_DATA", info), Equals, "/var/snap/name/42")
 	c.Check(builtin.ResolveSpecialVariable("$SNAP_COMMON", info), Equals, "/var/snap/name/common")
+	c.Check(builtin.ResolveSpecialVariable("$SNAP_DATA/", info), Equals, "/var/snap/name/42/")
+	// automatically prefixed with $SNAP
+	c.Check(builtin.ResolveSpecialVariable("foo", info), Equals, filepath.Join(dirs.CoreSnapMountDir, "name/42/foo"))
+	c.Check(builtin.ResolveSpecialVariable("foo/snap/bar", info), Equals, "/snap/name/42/foo/snap/bar")
+	// contain invalid variables
+	c.Check(builtin.ResolveSpecialVariable("$PRUNE/bar", info), Equals, "/snap/name/42//bar")
+	c.Check(builtin.ResolveSpecialVariable("bar/$PRUNE/foo", info), Equals, "/snap/name/42/bar//foo")
 }
 
 // Check that legacy syntax works and allows sharing read-only snap content
@@ -243,7 +249,7 @@ plugs:
   target: import
 `
 	consumerInfo := snaptest.MockInfo(c, consumerYaml, &snap.SideInfo{Revision: snap.R(7)})
-	plug := interfaces.NewConnectedPlug(consumerInfo.Plugs["content"], nil)
+	plug := interfaces.NewConnectedPlug(consumerInfo.Plugs["content"], nil, nil)
 	const producerYaml = `name: producer
 version: 0
 slots:
@@ -252,7 +258,7 @@ slots:
    - export
 `
 	producerInfo := snaptest.MockInfo(c, producerYaml, &snap.SideInfo{Revision: snap.R(5)})
-	slot := interfaces.NewConnectedSlot(producerInfo.Slots["content"], nil)
+	slot := interfaces.NewConnectedSlot(producerInfo.Slots["content"], nil, nil)
 
 	spec := &mount.Specification{}
 	c.Assert(spec.AddConnectedPlug(s.iface, plug, slot), IsNil)
@@ -276,7 +282,7 @@ apps:
   command: foo
 `
 	consumerInfo := snaptest.MockInfo(c, consumerYaml, &snap.SideInfo{Revision: snap.R(7)})
-	plug := interfaces.NewConnectedPlug(consumerInfo.Plugs["content"], nil)
+	plug := interfaces.NewConnectedPlug(consumerInfo.Plugs["content"], nil, nil)
 	const producerYaml = `name: producer
 version: 0
 slots:
@@ -285,7 +291,7 @@ slots:
    - $SNAP/export
 `
 	producerInfo := snaptest.MockInfo(c, producerYaml, &snap.SideInfo{Revision: snap.R(5)})
-	slot := interfaces.NewConnectedSlot(producerInfo.Slots["content"], nil)
+	slot := interfaces.NewConnectedSlot(producerInfo.Slots["content"], nil, nil)
 
 	spec := &mount.Specification{}
 	c.Assert(spec.AddConnectedPlug(s.iface, plug, slot), IsNil)
@@ -310,42 +316,139 @@ slots:
 
 	updateNS := apparmorSpec.UpdateNS()
 	profile0 := `  # Read-only content sharing consumer:content -> producer:content (r#0)
-  mount options=(bind) /snap/producer/5/export/ -> /snap/consumer/7/import/,
-  remount options=(bind, ro) /snap/consumer/7/import/,
-  umount /snap/consumer/7/import/,
+  mount options=(bind) /snap/producer/5/export/ -> /snap/consumer/7/import{,-[0-9]*}/,
+  remount options=(bind, ro) /snap/consumer/7/import{,-[0-9]*}/,
+  mount options=(rprivate) -> /snap/consumer/7/import{,-[0-9]*}/,
+  umount /snap/consumer/7/import{,-[0-9]*}/,
   # Writable mimic /snap/producer/5
+  # .. permissions for traversing the prefix that is assumed to exist
+  # .. variant with mimic at /
+  # Allow reading the mimic directory, it must exist in the first place.
+  / r,
+  # Allow setting the read-only directory aside via a bind mount.
+  /tmp/.snap/ rw,
+  mount options=(rbind, rw) / -> /tmp/.snap/,
+  # Allow mounting tmpfs over the read-only directory.
+  mount fstype=tmpfs options=(rw) tmpfs -> /,
+  # Allow creating empty files and directories for bind mounting things
+  # to reconstruct the now-writable parent directory.
+  /tmp/.snap/*/ rw,
+  /*/ rw,
+  mount options=(rbind, rw) /tmp/.snap/*/ -> /*/,
+  /tmp/.snap/* rw,
+  /* rw,
+  mount options=(bind, rw) /tmp/.snap/* -> /*,
+  # Allow unmounting the auxiliary directory.
+  # TODO: use fstype=tmpfs here for more strictness (LP: #1613403)
+  mount options=(rprivate) -> /tmp/.snap/,
+  umount /tmp/.snap/,
+  # Allow unmounting the destination directory as well as anything
+  # inside.  This lets us perform the undo plan in case the writable
+  # mimic fails.
+  mount options=(rprivate) -> /,
+  mount options=(rprivate) -> /*,
+  mount options=(rprivate) -> /*/,
+  umount /,
+  umount /*,
+  umount /*/,
+  # .. variant with mimic at /snap/
+  /snap/ r,
+  /tmp/.snap/snap/ rw,
+  mount options=(rbind, rw) /snap/ -> /tmp/.snap/snap/,
+  mount fstype=tmpfs options=(rw) tmpfs -> /snap/,
+  /tmp/.snap/snap/*/ rw,
+  /snap/*/ rw,
+  mount options=(rbind, rw) /tmp/.snap/snap/*/ -> /snap/*/,
+  /tmp/.snap/snap/* rw,
+  /snap/* rw,
+  mount options=(bind, rw) /tmp/.snap/snap/* -> /snap/*,
+  mount options=(rprivate) -> /tmp/.snap/snap/,
+  umount /tmp/.snap/snap/,
+  mount options=(rprivate) -> /snap/,
+  mount options=(rprivate) -> /snap/*,
+  mount options=(rprivate) -> /snap/*/,
+  umount /snap/,
+  umount /snap/*,
+  umount /snap/*/,
+  # .. variant with mimic at /snap/producer/
+  /snap/producer/ r,
+  /tmp/.snap/snap/producer/ rw,
+  mount options=(rbind, rw) /snap/producer/ -> /tmp/.snap/snap/producer/,
+  mount fstype=tmpfs options=(rw) tmpfs -> /snap/producer/,
+  /tmp/.snap/snap/producer/*/ rw,
+  /snap/producer/*/ rw,
+  mount options=(rbind, rw) /tmp/.snap/snap/producer/*/ -> /snap/producer/*/,
+  /tmp/.snap/snap/producer/* rw,
+  /snap/producer/* rw,
+  mount options=(bind, rw) /tmp/.snap/snap/producer/* -> /snap/producer/*,
+  mount options=(rprivate) -> /tmp/.snap/snap/producer/,
+  umount /tmp/.snap/snap/producer/,
+  mount options=(rprivate) -> /snap/producer/,
+  mount options=(rprivate) -> /snap/producer/*,
+  mount options=(rprivate) -> /snap/producer/*/,
+  umount /snap/producer/,
+  umount /snap/producer/*,
+  umount /snap/producer/*/,
+  # .. variant with mimic at /snap/producer/5/
+  /snap/producer/5/ r,
+  /tmp/.snap/snap/producer/5/ rw,
   mount options=(rbind, rw) /snap/producer/5/ -> /tmp/.snap/snap/producer/5/,
   mount fstype=tmpfs options=(rw) tmpfs -> /snap/producer/5/,
-  mount options=(rbind, rw) /tmp/.snap/snap/producer/5/** -> /snap/producer/5/**,
+  /tmp/.snap/snap/producer/5/*/ rw,
+  /snap/producer/5/*/ rw,
+  mount options=(rbind, rw) /tmp/.snap/snap/producer/5/*/ -> /snap/producer/5/*/,
+  /tmp/.snap/snap/producer/5/* rw,
+  /snap/producer/5/* rw,
   mount options=(bind, rw) /tmp/.snap/snap/producer/5/* -> /snap/producer/5/*,
+  mount options=(rprivate) -> /tmp/.snap/snap/producer/5/,
   umount /tmp/.snap/snap/producer/5/,
-  umount /snap/producer/5{,/**},
-  /snap/producer/5/** rw,
-  /snap/producer/5/ rw,
-  /snap/producer/ rw,
-  /tmp/.snap/snap/producer/5/** rw,
-  /tmp/.snap/snap/producer/5/ rw,
-  /tmp/.snap/snap/producer/ rw,
-  /tmp/.snap/snap/ rw,
-  /tmp/.snap/ rw,
+  mount options=(rprivate) -> /snap/producer/5/,
+  mount options=(rprivate) -> /snap/producer/5/*,
+  mount options=(rprivate) -> /snap/producer/5/*/,
+  umount /snap/producer/5/,
+  umount /snap/producer/5/*,
+  umount /snap/producer/5/*/,
   # Writable mimic /snap/consumer/7
+  # .. variant with mimic at /snap/consumer/
+  /snap/consumer/ r,
+  /tmp/.snap/snap/consumer/ rw,
+  mount options=(rbind, rw) /snap/consumer/ -> /tmp/.snap/snap/consumer/,
+  mount fstype=tmpfs options=(rw) tmpfs -> /snap/consumer/,
+  /tmp/.snap/snap/consumer/*/ rw,
+  /snap/consumer/*/ rw,
+  mount options=(rbind, rw) /tmp/.snap/snap/consumer/*/ -> /snap/consumer/*/,
+  /tmp/.snap/snap/consumer/* rw,
+  /snap/consumer/* rw,
+  mount options=(bind, rw) /tmp/.snap/snap/consumer/* -> /snap/consumer/*,
+  mount options=(rprivate) -> /tmp/.snap/snap/consumer/,
+  umount /tmp/.snap/snap/consumer/,
+  mount options=(rprivate) -> /snap/consumer/,
+  mount options=(rprivate) -> /snap/consumer/*,
+  mount options=(rprivate) -> /snap/consumer/*/,
+  umount /snap/consumer/,
+  umount /snap/consumer/*,
+  umount /snap/consumer/*/,
+  # .. variant with mimic at /snap/consumer/7/
+  /snap/consumer/7/ r,
+  /tmp/.snap/snap/consumer/7/ rw,
   mount options=(rbind, rw) /snap/consumer/7/ -> /tmp/.snap/snap/consumer/7/,
   mount fstype=tmpfs options=(rw) tmpfs -> /snap/consumer/7/,
-  mount options=(rbind, rw) /tmp/.snap/snap/consumer/7/** -> /snap/consumer/7/**,
+  /tmp/.snap/snap/consumer/7/*/ rw,
+  /snap/consumer/7/*/ rw,
+  mount options=(rbind, rw) /tmp/.snap/snap/consumer/7/*/ -> /snap/consumer/7/*/,
+  /tmp/.snap/snap/consumer/7/* rw,
+  /snap/consumer/7/* rw,
   mount options=(bind, rw) /tmp/.snap/snap/consumer/7/* -> /snap/consumer/7/*,
+  mount options=(rprivate) -> /tmp/.snap/snap/consumer/7/,
   umount /tmp/.snap/snap/consumer/7/,
-  umount /snap/consumer/7{,/**},
-  /snap/consumer/7/** rw,
-  /snap/consumer/7/ rw,
-  /snap/consumer/ rw,
-  /tmp/.snap/snap/consumer/7/** rw,
-  /tmp/.snap/snap/consumer/7/ rw,
-  /tmp/.snap/snap/consumer/ rw,
-  /tmp/.snap/snap/ rw,
-  /tmp/.snap/ rw,
+  mount options=(rprivate) -> /snap/consumer/7/,
+  mount options=(rprivate) -> /snap/consumer/7/*,
+  mount options=(rprivate) -> /snap/consumer/7/*/,
+  umount /snap/consumer/7/,
+  umount /snap/consumer/7/*,
+  umount /snap/consumer/7/*/,
 `
-	c.Assert(updateNS[0], Equals, profile0)
-	c.Assert(updateNS, DeepEquals, []string{profile0})
+	c.Assert(strings.Join(updateNS[:], ""), Equals, profile0)
 }
 
 // Check that sharing of writable data is possible
@@ -360,7 +463,7 @@ apps:
   command: foo
 `
 	consumerInfo := snaptest.MockInfo(c, consumerYaml, &snap.SideInfo{Revision: snap.R(7)})
-	plug := interfaces.NewConnectedPlug(consumerInfo.Plugs["content"], nil)
+	plug := interfaces.NewConnectedPlug(consumerInfo.Plugs["content"], nil, nil)
 	const producerYaml = `name: producer
 version: 0
 slots:
@@ -369,7 +472,7 @@ slots:
    - $SNAP_DATA/export
 `
 	producerInfo := snaptest.MockInfo(c, producerYaml, &snap.SideInfo{Revision: snap.R(5)})
-	slot := interfaces.NewConnectedSlot(producerInfo.Slots["content"], nil)
+	slot := interfaces.NewConnectedSlot(producerInfo.Slots["content"], nil, nil)
 
 	spec := &mount.Specification{}
 	c.Assert(spec.AddConnectedPlug(s.iface, plug, slot), IsNil)
@@ -396,8 +499,9 @@ slots:
 
 	updateNS := apparmorSpec.UpdateNS()
 	profile0 := `  # Read-write content sharing consumer:content -> producer:content (w#0)
-  mount options=(bind, rw) /var/snap/producer/5/export/ -> /var/snap/consumer/7/import/,
-  umount /var/snap/consumer/7/import/,
+  mount options=(bind, rw) /var/snap/producer/5/export/ -> /var/snap/consumer/7/import{,-[0-9]*}/,
+  mount options=(rprivate) -> /var/snap/consumer/7/import{,-[0-9]*}/,
+  umount /var/snap/consumer/7/import{,-[0-9]*}/,
   # Writable directory /var/snap/producer/5/export
   /var/snap/producer/5/export/ rw,
   /var/snap/producer/5/ rw,
@@ -406,9 +510,10 @@ slots:
   /var/snap/consumer/7/import/ rw,
   /var/snap/consumer/7/ rw,
   /var/snap/consumer/ rw,
+  # Writable directory /var/snap/consumer/7/import-[0-9]*
+  /var/snap/consumer/7/import-[0-9]*/ rw,
 `
-	c.Assert(updateNS[0], Equals, profile0)
-	c.Assert(updateNS, DeepEquals, []string{profile0})
+	c.Assert(strings.Join(updateNS[:], ""), Equals, profile0)
 }
 
 // Check that sharing of writable common data is possible
@@ -423,7 +528,7 @@ apps:
   command: foo
 `
 	consumerInfo := snaptest.MockInfo(c, consumerYaml, &snap.SideInfo{Revision: snap.R(7)})
-	plug := interfaces.NewConnectedPlug(consumerInfo.Plugs["content"], nil)
+	plug := interfaces.NewConnectedPlug(consumerInfo.Plugs["content"], nil, nil)
 	const producerYaml = `name: producer
 version: 0
 slots:
@@ -432,7 +537,7 @@ slots:
    - $SNAP_COMMON/export
 `
 	producerInfo := snaptest.MockInfo(c, producerYaml, &snap.SideInfo{Revision: snap.R(5)})
-	slot := interfaces.NewConnectedSlot(producerInfo.Slots["content"], nil)
+	slot := interfaces.NewConnectedSlot(producerInfo.Slots["content"], nil, nil)
 
 	spec := &mount.Specification{}
 	c.Assert(spec.AddConnectedPlug(s.iface, plug, slot), IsNil)
@@ -459,8 +564,9 @@ slots:
 
 	updateNS := apparmorSpec.UpdateNS()
 	profile0 := `  # Read-write content sharing consumer:content -> producer:content (w#0)
-  mount options=(bind, rw) /var/snap/producer/common/export/ -> /var/snap/consumer/common/import/,
-  umount /var/snap/consumer/common/import/,
+  mount options=(bind, rw) /var/snap/producer/common/export/ -> /var/snap/consumer/common/import{,-[0-9]*}/,
+  mount options=(rprivate) -> /var/snap/consumer/common/import{,-[0-9]*}/,
+  umount /var/snap/consumer/common/import{,-[0-9]*}/,
   # Writable directory /var/snap/producer/common/export
   /var/snap/producer/common/export/ rw,
   /var/snap/producer/common/ rw,
@@ -469,9 +575,10 @@ slots:
   /var/snap/consumer/common/import/ rw,
   /var/snap/consumer/common/ rw,
   /var/snap/consumer/ rw,
+  # Writable directory /var/snap/consumer/common/import-[0-9]*
+  /var/snap/consumer/common/import-[0-9]*/ rw,
 `
-	c.Assert(updateNS[0], Equals, profile0)
-	c.Assert(updateNS, DeepEquals, []string{profile0})
+	c.Assert(strings.Join(updateNS[:], ""), Equals, profile0)
 }
 
 func (s *ContentSuite) TestInterfaces(c *C) {
@@ -488,7 +595,7 @@ apps:
  app:
   command: foo
 `, &snap.SideInfo{Revision: snap.R(1)}, "content")
-	connectedPlug := interfaces.NewConnectedPlug(plug, nil)
+	connectedPlug := interfaces.NewConnectedPlug(plug, nil, nil)
 
 	slot := MockSlot(c, `name: producer
 version: 0
@@ -503,7 +610,7 @@ slots:
      - $SNAP_COMMON/write-common
      - $SNAP_DATA/write-data
 `, &snap.SideInfo{Revision: snap.R(2)}, "content")
-	connectedSlot := interfaces.NewConnectedSlot(slot, nil)
+	connectedSlot := interfaces.NewConnectedSlot(slot, nil, nil)
 
 	// Create the mount and apparmor specifications.
 	mountSpec := &mount.Specification{}
@@ -554,11 +661,12 @@ slots:
 /snap/producer/2/read-snap/** mrkix,
 `
 	c.Assert(apparmorSpec.SnippetForTag("snap.consumer.app"), Equals, expected)
-	fmt.Printf("")
+
 	updateNS := apparmorSpec.UpdateNS()
 	profile0 := `  # Read-write content sharing consumer:content -> producer:content (w#0)
-  mount options=(bind, rw) /var/snap/producer/common/write-common/ -> /var/snap/consumer/common/import/write-common/,
-  umount /var/snap/consumer/common/import/write-common/,
+  mount options=(bind, rw) /var/snap/producer/common/write-common/ -> /var/snap/consumer/common/import/write-common{,-[0-9]*}/,
+  mount options=(rprivate) -> /var/snap/consumer/common/import/write-common{,-[0-9]*}/,
+  umount /var/snap/consumer/common/import/write-common{,-[0-9]*}/,
   # Writable directory /var/snap/producer/common/write-common
   /var/snap/producer/common/write-common/ rw,
   /var/snap/producer/common/ rw,
@@ -568,79 +676,171 @@ slots:
   /var/snap/consumer/common/import/ rw,
   /var/snap/consumer/common/ rw,
   /var/snap/consumer/ rw,
+  # Writable directory /var/snap/consumer/common/import/write-common-[0-9]*
+  /var/snap/consumer/common/import/write-common-[0-9]*/ rw,
 `
+	// Find the slice that describes profile0 by looking for the first unique
+	// line of the next profile.
+	start := 0
+	end, _ := apparmorSpec.UpdateNSIndexOf("  # Read-write content sharing consumer:content -> producer:content (w#1)\n")
+	c.Assert(strings.Join(updateNS[start:end], ""), Equals, profile0)
+
 	profile1 := `  # Read-write content sharing consumer:content -> producer:content (w#1)
-  mount options=(bind, rw) /var/snap/producer/2/write-data/ -> /var/snap/consumer/common/import/write-data/,
-  umount /var/snap/consumer/common/import/write-data/,
+  mount options=(bind, rw) /var/snap/producer/2/write-data/ -> /var/snap/consumer/common/import/write-data{,-[0-9]*}/,
+  mount options=(rprivate) -> /var/snap/consumer/common/import/write-data{,-[0-9]*}/,
+  umount /var/snap/consumer/common/import/write-data{,-[0-9]*}/,
   # Writable directory /var/snap/producer/2/write-data
   /var/snap/producer/2/write-data/ rw,
   /var/snap/producer/2/ rw,
-  /var/snap/producer/ rw,
   # Writable directory /var/snap/consumer/common/import/write-data
   /var/snap/consumer/common/import/write-data/ rw,
-  /var/snap/consumer/common/import/ rw,
-  /var/snap/consumer/common/ rw,
-  /var/snap/consumer/ rw,
+  # Writable directory /var/snap/consumer/common/import/write-data-[0-9]*
+  /var/snap/consumer/common/import/write-data-[0-9]*/ rw,
 `
+	// Find the slice that describes profile1 by looking for the first unique
+	// line of the next profile.
+	start = end
+	end, _ = apparmorSpec.UpdateNSIndexOf("  # Read-only content sharing consumer:content -> producer:content (r#0)\n")
+	c.Assert(strings.Join(updateNS[start:end], ""), Equals, profile1)
+
 	profile2 := `  # Read-only content sharing consumer:content -> producer:content (r#0)
-  mount options=(bind) /var/snap/producer/common/read-common/ -> /var/snap/consumer/common/import/read-common/,
-  remount options=(bind, ro) /var/snap/consumer/common/import/read-common/,
-  umount /var/snap/consumer/common/import/read-common/,
+  mount options=(bind) /var/snap/producer/common/read-common/ -> /var/snap/consumer/common/import/read-common{,-[0-9]*}/,
+  remount options=(bind, ro) /var/snap/consumer/common/import/read-common{,-[0-9]*}/,
+  mount options=(rprivate) -> /var/snap/consumer/common/import/read-common{,-[0-9]*}/,
+  umount /var/snap/consumer/common/import/read-common{,-[0-9]*}/,
   # Writable directory /var/snap/producer/common/read-common
   /var/snap/producer/common/read-common/ rw,
-  /var/snap/producer/common/ rw,
-  /var/snap/producer/ rw,
   # Writable directory /var/snap/consumer/common/import/read-common
   /var/snap/consumer/common/import/read-common/ rw,
-  /var/snap/consumer/common/import/ rw,
-  /var/snap/consumer/common/ rw,
-  /var/snap/consumer/ rw,
+  # Writable directory /var/snap/consumer/common/import/read-common-[0-9]*
+  /var/snap/consumer/common/import/read-common-[0-9]*/ rw,
 `
+	// Find the slice that describes profile2 by looking for the first unique
+	// line of the next profile.
+	start = end
+	end, _ = apparmorSpec.UpdateNSIndexOf("  # Read-only content sharing consumer:content -> producer:content (r#1)\n")
+	c.Assert(strings.Join(updateNS[start:end], ""), Equals, profile2)
+
 	profile3 := `  # Read-only content sharing consumer:content -> producer:content (r#1)
-  mount options=(bind) /var/snap/producer/2/read-data/ -> /var/snap/consumer/common/import/read-data/,
-  remount options=(bind, ro) /var/snap/consumer/common/import/read-data/,
-  umount /var/snap/consumer/common/import/read-data/,
+  mount options=(bind) /var/snap/producer/2/read-data/ -> /var/snap/consumer/common/import/read-data{,-[0-9]*}/,
+  remount options=(bind, ro) /var/snap/consumer/common/import/read-data{,-[0-9]*}/,
+  mount options=(rprivate) -> /var/snap/consumer/common/import/read-data{,-[0-9]*}/,
+  umount /var/snap/consumer/common/import/read-data{,-[0-9]*}/,
   # Writable directory /var/snap/producer/2/read-data
   /var/snap/producer/2/read-data/ rw,
-  /var/snap/producer/2/ rw,
-  /var/snap/producer/ rw,
   # Writable directory /var/snap/consumer/common/import/read-data
   /var/snap/consumer/common/import/read-data/ rw,
-  /var/snap/consumer/common/import/ rw,
-  /var/snap/consumer/common/ rw,
-  /var/snap/consumer/ rw,
+  # Writable directory /var/snap/consumer/common/import/read-data-[0-9]*
+  /var/snap/consumer/common/import/read-data-[0-9]*/ rw,
 `
+	// Find the slice that describes profile3 by looking for the first unique
+	// line of the next profile.
+	start = end
+	end, _ = apparmorSpec.UpdateNSIndexOf("  # Read-only content sharing consumer:content -> producer:content (r#2)\n")
+	c.Assert(strings.Join(updateNS[start:end], ""), Equals, profile3)
+
 	profile4 := `  # Read-only content sharing consumer:content -> producer:content (r#2)
-  mount options=(bind) /snap/producer/2/read-snap/ -> /var/snap/consumer/common/import/read-snap/,
-  remount options=(bind, ro) /var/snap/consumer/common/import/read-snap/,
-  umount /var/snap/consumer/common/import/read-snap/,
+  mount options=(bind) /snap/producer/2/read-snap/ -> /var/snap/consumer/common/import/read-snap{,-[0-9]*}/,
+  remount options=(bind, ro) /var/snap/consumer/common/import/read-snap{,-[0-9]*}/,
+  mount options=(rprivate) -> /var/snap/consumer/common/import/read-snap{,-[0-9]*}/,
+  umount /var/snap/consumer/common/import/read-snap{,-[0-9]*}/,
   # Writable mimic /snap/producer/2
+  # .. permissions for traversing the prefix that is assumed to exist
+  # .. variant with mimic at /
+  # Allow reading the mimic directory, it must exist in the first place.
+  / r,
+  # Allow setting the read-only directory aside via a bind mount.
+  /tmp/.snap/ rw,
+  mount options=(rbind, rw) / -> /tmp/.snap/,
+  # Allow mounting tmpfs over the read-only directory.
+  mount fstype=tmpfs options=(rw) tmpfs -> /,
+  # Allow creating empty files and directories for bind mounting things
+  # to reconstruct the now-writable parent directory.
+  /tmp/.snap/*/ rw,
+  /*/ rw,
+  mount options=(rbind, rw) /tmp/.snap/*/ -> /*/,
+  /tmp/.snap/* rw,
+  /* rw,
+  mount options=(bind, rw) /tmp/.snap/* -> /*,
+  # Allow unmounting the auxiliary directory.
+  # TODO: use fstype=tmpfs here for more strictness (LP: #1613403)
+  mount options=(rprivate) -> /tmp/.snap/,
+  umount /tmp/.snap/,
+  # Allow unmounting the destination directory as well as anything
+  # inside.  This lets us perform the undo plan in case the writable
+  # mimic fails.
+  mount options=(rprivate) -> /,
+  mount options=(rprivate) -> /*,
+  mount options=(rprivate) -> /*/,
+  umount /,
+  umount /*,
+  umount /*/,
+  # .. variant with mimic at /snap/
+  /snap/ r,
+  /tmp/.snap/snap/ rw,
+  mount options=(rbind, rw) /snap/ -> /tmp/.snap/snap/,
+  mount fstype=tmpfs options=(rw) tmpfs -> /snap/,
+  /tmp/.snap/snap/*/ rw,
+  /snap/*/ rw,
+  mount options=(rbind, rw) /tmp/.snap/snap/*/ -> /snap/*/,
+  /tmp/.snap/snap/* rw,
+  /snap/* rw,
+  mount options=(bind, rw) /tmp/.snap/snap/* -> /snap/*,
+  mount options=(rprivate) -> /tmp/.snap/snap/,
+  umount /tmp/.snap/snap/,
+  mount options=(rprivate) -> /snap/,
+  mount options=(rprivate) -> /snap/*,
+  mount options=(rprivate) -> /snap/*/,
+  umount /snap/,
+  umount /snap/*,
+  umount /snap/*/,
+  # .. variant with mimic at /snap/producer/
+  /snap/producer/ r,
+  /tmp/.snap/snap/producer/ rw,
+  mount options=(rbind, rw) /snap/producer/ -> /tmp/.snap/snap/producer/,
+  mount fstype=tmpfs options=(rw) tmpfs -> /snap/producer/,
+  /tmp/.snap/snap/producer/*/ rw,
+  /snap/producer/*/ rw,
+  mount options=(rbind, rw) /tmp/.snap/snap/producer/*/ -> /snap/producer/*/,
+  /tmp/.snap/snap/producer/* rw,
+  /snap/producer/* rw,
+  mount options=(bind, rw) /tmp/.snap/snap/producer/* -> /snap/producer/*,
+  mount options=(rprivate) -> /tmp/.snap/snap/producer/,
+  umount /tmp/.snap/snap/producer/,
+  mount options=(rprivate) -> /snap/producer/,
+  mount options=(rprivate) -> /snap/producer/*,
+  mount options=(rprivate) -> /snap/producer/*/,
+  umount /snap/producer/,
+  umount /snap/producer/*,
+  umount /snap/producer/*/,
+  # .. variant with mimic at /snap/producer/2/
+  /snap/producer/2/ r,
+  /tmp/.snap/snap/producer/2/ rw,
   mount options=(rbind, rw) /snap/producer/2/ -> /tmp/.snap/snap/producer/2/,
   mount fstype=tmpfs options=(rw) tmpfs -> /snap/producer/2/,
-  mount options=(rbind, rw) /tmp/.snap/snap/producer/2/** -> /snap/producer/2/**,
+  /tmp/.snap/snap/producer/2/*/ rw,
+  /snap/producer/2/*/ rw,
+  mount options=(rbind, rw) /tmp/.snap/snap/producer/2/*/ -> /snap/producer/2/*/,
+  /tmp/.snap/snap/producer/2/* rw,
+  /snap/producer/2/* rw,
   mount options=(bind, rw) /tmp/.snap/snap/producer/2/* -> /snap/producer/2/*,
+  mount options=(rprivate) -> /tmp/.snap/snap/producer/2/,
   umount /tmp/.snap/snap/producer/2/,
-  umount /snap/producer/2{,/**},
-  /snap/producer/2/** rw,
-  /snap/producer/2/ rw,
-  /snap/producer/ rw,
-  /tmp/.snap/snap/producer/2/** rw,
-  /tmp/.snap/snap/producer/2/ rw,
-  /tmp/.snap/snap/producer/ rw,
-  /tmp/.snap/snap/ rw,
-  /tmp/.snap/ rw,
+  mount options=(rprivate) -> /snap/producer/2/,
+  mount options=(rprivate) -> /snap/producer/2/*,
+  mount options=(rprivate) -> /snap/producer/2/*/,
+  umount /snap/producer/2/,
+  umount /snap/producer/2/*,
+  umount /snap/producer/2/*/,
   # Writable directory /var/snap/consumer/common/import/read-snap
   /var/snap/consumer/common/import/read-snap/ rw,
-  /var/snap/consumer/common/import/ rw,
-  /var/snap/consumer/common/ rw,
-  /var/snap/consumer/ rw,
+  # Writable directory /var/snap/consumer/common/import/read-snap-[0-9]*
+  /var/snap/consumer/common/import/read-snap-[0-9]*/ rw,
 `
-	c.Assert(updateNS[0], Equals, profile0)
-	c.Assert(updateNS[1], Equals, profile1)
-	c.Assert(updateNS[2], Equals, profile2)
-	c.Assert(updateNS[3], Equals, profile3)
-	c.Assert(updateNS[4], Equals, profile4)
-	c.Assert(updateNS, DeepEquals, []string{profile0, profile1, profile2, profile3, profile4})
+	// Find the slice that describes profile4 by looking till the end of the list.
+	start = end
+	c.Assert(strings.Join(updateNS[start:], ""), Equals, profile4)
+	c.Assert(strings.Join(updateNS, ""), DeepEquals, strings.Join([]string{profile0, profile1, profile2, profile3, profile4}, ""))
 }
 
 func (s *ContentSuite) TestModernContentInterfacePlugins(c *C) {
@@ -657,7 +857,7 @@ apps:
   command: foo
 
 `, &snap.SideInfo{Revision: snap.R(1)}, "plugins")
-	connectedPlug := interfaces.NewConnectedPlug(plug, nil)
+	connectedPlug := interfaces.NewConnectedPlug(plug, nil, nil)
 
 	// XXX: realistically the plugin may be a single file and we don't support
 	// those very well.
@@ -669,7 +869,7 @@ slots:
   source:
     read: [$SNAP/plugin]
 `, &snap.SideInfo{Revision: snap.R(1)}, "plugin-for-app")
-	connectedSlotOne := interfaces.NewConnectedSlot(slotOne, nil)
+	connectedSlotOne := interfaces.NewConnectedSlot(slotOne, nil, nil)
 
 	slotTwo := MockSlot(c, `name: plugin-two
 version: 0
@@ -679,7 +879,7 @@ slots:
   source:
     read: [$SNAP/plugin]
 `, &snap.SideInfo{Revision: snap.R(1)}, "plugin-for-app")
-	connectedSlotTwo := interfaces.NewConnectedSlot(slotTwo, nil)
+	connectedSlotTwo := interfaces.NewConnectedSlot(slotTwo, nil, nil)
 
 	// Create the mount and apparmor specifications.
 	mountSpec := &mount.Specification{}
@@ -732,7 +932,7 @@ apps:
  app:
   command: foo
 `, &snap.SideInfo{Revision: snap.R(1)}, "content")
-	connectedPlug := interfaces.NewConnectedPlug(plug, nil)
+	connectedPlug := interfaces.NewConnectedPlug(plug, nil, nil)
 
 	slot := MockSlot(c, `name: producer
 version: 0
@@ -744,7 +944,7 @@ slots:
     write:
      - $SNAP_DATA/directory
 `, &snap.SideInfo{Revision: snap.R(2)}, "content")
-	connectedSlot := interfaces.NewConnectedSlot(slot, nil)
+	connectedSlot := interfaces.NewConnectedSlot(slot, nil, nil)
 
 	// Create the mount and apparmor specifications.
 	mountSpec := &mount.Specification{}
@@ -794,7 +994,7 @@ plugs:
   target: $SNAP_COMMON/import
 `
 	consumerInfo := snaptest.MockInfo(c, consumerYaml, &snap.SideInfo{Revision: snap.R(7)})
-	plug := interfaces.NewConnectedPlug(consumerInfo.Plugs["content"], nil)
+	plug := interfaces.NewConnectedPlug(consumerInfo.Plugs["content"], nil, nil)
 	const producerYaml = `name: producer
 version: 0
 slots:
@@ -806,7 +1006,7 @@ apps:
     command: bar
 `
 	producerInfo := snaptest.MockInfo(c, producerYaml, &snap.SideInfo{Revision: snap.R(5)})
-	slot := interfaces.NewConnectedSlot(producerInfo.Slots["content"], nil)
+	slot := interfaces.NewConnectedSlot(producerInfo.Slots["content"], nil, nil)
 
 	apparmorSpec := &apparmor.Specification{}
 	err := apparmorSpec.AddConnectedSlot(s.iface, plug, slot)
@@ -821,3 +1021,12 @@ apps:
 `
 	c.Assert(apparmorSpec.SnippetForTag("snap.producer.app"), Equals, expected)
 }
+
+func (s *ContentSuite) TestStaticInfo(c *C) {
+	si := interfaces.StaticInfoOf(s.iface)
+	c.Assert(si.ImplicitOnCore, Equals, false)
+	c.Assert(si.ImplicitOnClassic, Equals, false)
+	c.Assert(si.Summary, Equals, `allows sharing code and data with other snaps`)
+	c.Assert(si.BaseDeclarationSlots, testutil.Contains, "content: $SLOT(content)")
+	c.Assert(si.AffectsPlugOnRefresh, Equals, true)
+}
