@@ -1,7 +1,7 @@
 // -*- Mode: Go; indent-tabs-mode: t -*-
 
 /*
- * Copyright (C) 2016-2020 Canonical Ltd
+ * Copyright (C) 2016-2021 Canonical Ltd
  *
  * This program is free software: you can redistribute it and/or modify
  * it under the terms of the GNU General Public License version 3 as
@@ -34,13 +34,14 @@ import (
 	"github.com/snapcore/snapd/asserts/sysdb"
 	"github.com/snapcore/snapd/boot"
 	"github.com/snapcore/snapd/dirs"
+	"github.com/snapcore/snapd/gadget"
 	"github.com/snapcore/snapd/i18n"
+	"github.com/snapcore/snapd/kernel/fde"
 	"github.com/snapcore/snapd/logger"
 	"github.com/snapcore/snapd/osutil"
 	"github.com/snapcore/snapd/overlord/assertstate"
 	"github.com/snapcore/snapd/overlord/auth"
 	"github.com/snapcore/snapd/overlord/configstate/config"
-	"github.com/snapcore/snapd/overlord/devicestate/fde"
 	"github.com/snapcore/snapd/overlord/devicestate/internal"
 	"github.com/snapcore/snapd/overlord/hookstate"
 	"github.com/snapcore/snapd/overlord/snapstate"
@@ -48,8 +49,12 @@ import (
 	"github.com/snapcore/snapd/overlord/storecontext"
 	"github.com/snapcore/snapd/progress"
 	"github.com/snapcore/snapd/release"
+	"github.com/snapcore/snapd/secboot"
+	"github.com/snapcore/snapd/seed"
 	"github.com/snapcore/snapd/snap"
+	"github.com/snapcore/snapd/snap/snapfile"
 	"github.com/snapcore/snapd/snapdenv"
+	"github.com/snapcore/snapd/strutil"
 	"github.com/snapcore/snapd/sysconfig"
 	"github.com/snapcore/snapd/systemd"
 	"github.com/snapcore/snapd/timings"
@@ -60,10 +65,16 @@ var (
 	restrictCloudInit = sysconfig.RestrictCloudInit
 )
 
+// EarlyConfig is a hook set by configstate that can process early configuration
+// during managers' startup.
+var EarlyConfig func(st *state.State, preloadGadget func() (sysconfig.Device, *gadget.Info, error)) error
+
 // DeviceManager is responsible for managing the device identity and device
 // policies.
 type DeviceManager struct {
-	systemMode string
+	// sysMode is the system mode from modeenv or "" on pre-UC20,
+	// use SystemMode instead
+	sysMode string
 	// saveAvailable keeps track whether /var/lib/snapd/save
 	// is available, i.e. exists and is mounted from ubuntu-save
 	// if the latter exists.
@@ -82,10 +93,14 @@ type DeviceManager struct {
 	bootOkRan            bool
 	bootRevisionsUpdated bool
 
+	seedTimings *timings.Timings
+
 	ensureSeedInConfigRan bool
 
 	ensureInstalledRan bool
 
+	ensureTriedRecoverySystemRan bool
+
 	cloudInitAlreadyRestricted           bool
 	cloudInitErrorAttemptStart           *time.Time
 	cloudInitEnabledInactiveAttemptStart *time.Time
@@ -115,7 +130,7 @@ func Manager(s *state.State, hookManager *hookstate.HookManager, runner *state.T
 		return nil, err
 	}
 	if modeEnv != nil {
-		m.systemMode = modeEnv.Mode
+		m.sysMode = modeEnv.Mode
 	}
 
 	s.Lock()
@@ -126,13 +141,15 @@ func Manager(s *state.State, hookManager *hookstate.HookManager, runner *state.T
 		return nil, err
 	}
 
-	hookManager.Register(regexp.MustCompile("^prepare-device$"), newPrepareDeviceHandler)
+	hookManager.Register(regexp.MustCompile("^prepare-device$"), newBasicHookStateHandler)
+	hookManager.Register(regexp.MustCompile("^install-device$"), newBasicHookStateHandler)
 
 	runner.AddHandler("generate-device-key", m.doGenerateDeviceKey, nil)
 	runner.AddHandler("request-serial", m.doRequestSerial, nil)
 	runner.AddHandler("mark-preseeded", m.doMarkPreseeded, nil)
 	runner.AddHandler("mark-seeded", m.doMarkSeeded, nil)
 	runner.AddHandler("setup-run-system", m.doSetupRunSystem, nil)
+	runner.AddHandler("restart-system-to-run-mode", m.doRestartSystemToRunMode, nil)
 	runner.AddHandler("prepare-remodeling", m.doPrepareRemodeling, nil)
 	runner.AddCleanup("prepare-remodeling", m.cleanupRemodel)
 	// this *must* always run last and finalizes a remodel
@@ -145,6 +162,15 @@ func Manager(s *state.State, hookManager *hookstate.HookManager, runner *state.T
 	// or gadget snaps. There are no further changes to the boot assets,
 	// unless a new gadget update is deployed.
 	runner.AddHandler("update-gadget-assets", m.doUpdateGadgetAssets, nil)
+	// There is no undo handler for successful boot config update. The
+	// config assets are assumed to be always backwards compatible.
+	runner.AddHandler("update-managed-boot-config", m.doUpdateManagedBootConfig, nil)
+	// kernel command line updates from a gadget supplied file
+	runner.AddHandler("update-gadget-cmdline", m.doUpdateGadgetCommandLine, m.undoUpdateGadgetCommandLine)
+	// recovery systems
+	runner.AddHandler("create-recovery-system", m.doCreateRecoverySystem, m.undoCreateRecoverySystem)
+	runner.AddHandler("finalize-recovery-system", m.doFinalizeTriedRecoverySystem, m.undoFinalizeTriedRecoverySystem)
+	runner.AddCleanup("finalize-recovery-system", m.cleanupRecoverySystem)
 
 	runner.AddBlocked(gadgetUpdateBlocked)
 
@@ -156,6 +182,16 @@ func Manager(s *state.State, hookManager *hookstate.HookManager, runner *state.T
 	return m, nil
 }
 
+type genericHook struct{}
+
+func (h genericHook) Before() error                 { return nil }
+func (h genericHook) Done() error                   { return nil }
+func (h genericHook) Error(err error) (bool, error) { return false, nil }
+
+func newBasicHookStateHandler(context *hookstate.Context) hookstate.Handler {
+	return genericHook{}
+}
+
 func maybeReadModeenv() (*boot.Modeenv, error) {
 	modeEnv, err := boot.ReadModeenv("")
 	if err != nil && !os.IsNotExist(err) {
@@ -164,17 +200,60 @@ func maybeReadModeenv() (*boot.Modeenv, error) {
 	return modeEnv, nil
 }
 
+// ReloadModeenv is only useful for integration testing
+func (m *DeviceManager) ReloadModeenv() error {
+	osutil.MustBeTestBinary("ReloadModeenv can only be called from tests")
+	modeEnv, err := maybeReadModeenv()
+	if err != nil {
+		return err
+	}
+	if modeEnv != nil {
+		m.sysMode = modeEnv.Mode
+	}
+	return nil
+}
+
+type SysExpectation int
+
+const (
+	// SysAny indicates any system is appropriate.
+	SysAny SysExpectation = iota
+	// SysHasModeenv indicates only systems with modeenv are appropriate.
+	SysHasModeenv
+)
+
+// SystemMode returns the current mode of the system.
+// An expectation about the system controls the returned mode when
+// none is set explicitly, as it's the case on pre-UC20 systems. In
+// which case, with SysAny, the mode defaults to implicit "run", thus
+// covering pre-UC20 systems. With SysHasModeeenv, as there is always
+// an explicit mode in systems that use modeenv, no implicit default
+// is used and thus "" is returned for pre-UC20 systems.
+func (m *DeviceManager) SystemMode(sysExpect SysExpectation) string {
+	if m.sysMode == "" {
+		if sysExpect == SysHasModeenv {
+			return ""
+		}
+		return "run"
+	}
+	return m.sysMode
+}
+
 // StartUp implements StateStarterUp.Startup.
 func (m *DeviceManager) StartUp() error {
-	// system mode is explicitly set on UC20
 	// TODO:UC20: ubuntu-save needs to be mounted for recover too
-	if !release.OnClassic && m.systemMode == "run" {
+	if !release.OnClassic && m.SystemMode(SysHasModeenv) == "run" {
+		// UC20 and run mode => setup /var/lib/snapd/save
 		if err := m.maybeSetupUbuntuSave(); err != nil {
 			return fmt.Errorf("cannot set up ubuntu-save: %v", err)
 		}
 	}
 
-	return nil
+	// TODO: setup proper timings measurements for this
+
+	m.state.Lock()
+	defer m.state.Unlock()
+	return EarlyConfig(m.state, m.preloadGadget)
 }
 
 func (m *DeviceManager) maybeSetupUbuntuSave() error {
@@ -258,37 +337,18 @@ func gadgetUpdateBlocked(cand *state.Task, running []*state.Task) bool {
 	if cand.Kind() == "update-gadget-assets" && len(running) != 0 {
 		// update-gadget-assets must be the only task running
 		return true
-	} else {
-		for _, other := range running {
-			if other.Kind() == "update-gadget-assets" {
-				// no other task can be started when
-				// update-gadget-assets is running
-				return true
-			}
+	}
+	for _, other := range running {
+		if other.Kind() == "update-gadget-assets" {
+			// no other task can be started when
+			// update-gadget-assets is running
+			return true
 		}
 	}
 
 	return false
 }
 
-type prepareDeviceHandler struct{}
-
-func newPrepareDeviceHandler(context *hookstate.Context) hookstate.Handler {
-	return prepareDeviceHandler{}
-}
-
-func (h prepareDeviceHandler) Before() error {
-	return nil
-}
-
-func (h prepareDeviceHandler) Done() error {
-	return nil
-}
-
-func (h prepareDeviceHandler) Error(err error) error {
-	return nil
-}
-
 func (m *DeviceManager) changeInFlight(kind string) bool {
 	for _, chg := range m.state.Changes() {
 		if chg.Kind() == kind && !chg.Status().Ready() {
@@ -347,18 +407,11 @@ func setClassicFallbackModel(st *state.State, device *auth.DeviceState) error {
 	return nil
 }
 
-func (m *DeviceManager) SystemMode() string {
-	if m.systemMode == "" {
-		return "run"
-	}
-	return m.systemMode
-}
-
 func (m *DeviceManager) ensureOperational() error {
 	m.state.Lock()
 	defer m.state.Unlock()
 
-	if m.SystemMode() != "run" {
+	if m.SystemMode(SysAny) != "run" {
 		// avoid doing registration in ephemeral mode
 		// note: this also stop auto-refreshes indirectly
 		return nil
@@ -476,9 +529,6 @@ func (m *DeviceManager) ensureOperational() error {
 		}
 		prepareDevice = hookstate.HookTask(m.state, summary, hooksup, nil)
 		tasks = append(tasks, prepareDevice)
-		// hooks are under a different manager, make sure we consider
-		// it immediately
-		m.state.EnsureBefore(0)
 	}
 
 	genKey := m.state.NewTask("generate-device-key", i18n.G("Generate device key"))
@@ -519,6 +569,101 @@ func (m *DeviceManager) setTimeOnce(name string, t time.Time) error {
 	return nil
 }
 
+func (m *DeviceManager) seedStart() (*timings.Timings, error) {
+	if m.seedTimings != nil {
+		// reuse the early cached one
+		return m.seedTimings, nil
+	}
+	perfTimings := timings.New(map[string]string{"ensure": "seed"})
+
+	var recordedStart string
+	var start time.Time
+	if m.preseed {
+		recordedStart = "preseed-start-time"
+		start = timeNow()
+	} else {
+		recordedStart = "seed-start-time"
+		start = startTime
+	}
+	if err := m.setTimeOnce(recordedStart, start); err != nil {
+		return nil, err
+	}
+	return perfTimings, nil
+}
+
+func (m *DeviceManager) preloadGadget() (sysconfig.Device, *gadget.Info, error) {
+	var sysLabel string
+	modeEnv, err := maybeReadModeenv()
+	if err != nil {
+		return nil, nil, err
+	}
+	if modeEnv != nil {
+		sysLabel = modeEnv.RecoverySystem
+	}
+
+	// we time preloadGadget + first ensureSeeded together
+	// under --ensure=seed
+	tm, err := m.seedStart()
+	if err != nil {
+		return nil, nil, err
+	}
+	// cached for first ensureSeeded
+	m.seedTimings = tm
+
+	// Here we behave as if there was no gadget if we encounter
+	// errors, under the assumption that those will be resurfaced
+	// in ensureSeed. This preserves having a failing to seed
+	// snapd continuing running.
+	//
+	// TODO: consider changing that again but we need consider the
+	// effect of the different failure mode.
+	//
+	// We also assume that anything sensitive will not be guarded
+	// just by option flags. For example automatic user creation
+	// also requires the model to be known/set. Otherwise ignoring
+	// errors here would be problematic.
+	var deviceSeed seed.Seed
+	timings.Run(tm, "import-assertions[early]", "early import assertions from seed", func(nested timings.Measurer) {
+		deviceSeed, err = loadDeviceSeed(m.state, sysLabel)
+	})
+	if err != nil {
+		// this same error will be resurfaced in ensureSeed later
+		if err != seed.ErrNoAssertions {
+			logger.Debugf("early import assertions from seed failed: %v", err)
+		}
+		return nil, nil, state.ErrNoState
+	}
+	model := deviceSeed.Model()
+	if model.Gadget() == "" {
+		// no gadget
+		return nil, nil, state.ErrNoState
+	}
+	var gi *gadget.Info
+	timings.Run(tm, "preload-verified-gadget-metadata", "preload verified gadget metadata from seed", func(nested timings.Measurer) {
+		gi, err = func() (*gadget.Info, error) {
+			if err := deviceSeed.LoadEssentialMeta([]snap.Type{snap.TypeGadget}, nested); err != nil {
+				return nil, err
+			}
+			essGadget := deviceSeed.EssentialSnaps()
+			if len(essGadget) != 1 {
+				return nil, fmt.Errorf("multiple gadgets among essential snaps are unexpected")
+			}
+			snapf, err := snapfile.Open(essGadget[0].Path)
+			if err != nil {
+				return nil, err
+			}
+			return gadget.ReadInfoFromSnapFile(snapf, model)
+		}()
+	})
+	if err != nil {
+		logger.Noticef("preload verified gadget metadata from seed failed: %v", err)
+		return nil, nil, state.ErrNoState
+	}
+
+	dev := newModelDeviceContext(m, model)
+	return dev, gi, nil
+}
+
 var populateStateFromSeed = populateStateFromSeedImpl
 
 // ensureSeeded makes sure that the snaps from seed.yaml get installed
@@ -536,24 +681,17 @@ func (m *DeviceManager) ensureSeeded() error {
 		return nil
 	}
 
-	perfTimings := timings.New(map[string]string{"ensure": "seed"})
-
 	if m.changeInFlight("seed") {
 		return nil
 	}
 
-	var recordedStart string
-	var start time.Time
-	if m.preseed {
-		recordedStart = "preseed-start-time"
-		start = timeNow()
-	} else {
-		recordedStart = "seed-start-time"
-		start = startTime
-	}
-	if err := m.setTimeOnce(recordedStart, start); err != nil {
+	perfTimings, err := m.seedStart()
+	if err != nil {
 		return err
 	}
+	// we time preloadGadget + first ensureSeeded together
+	// succcessive ensureSeeded should be timed separately
+	m.seedTimings = nil
 
 	var opts *populateStateFromSeedOptions
 	if m.preseed {
@@ -565,7 +703,7 @@ func (m *DeviceManager) ensureSeeded() error {
 		}
 		if modeEnv != nil {
 			opts = &populateStateFromSeedOptions{
-				Mode:  m.systemMode,
+				Mode:  modeEnv.Mode,
 				Label: modeEnv.RecoverySystem,
 			}
 		}
@@ -593,12 +731,6 @@ func (m *DeviceManager) ensureSeeded() error {
 	return nil
 }
 
-// ResetBootOk is only useful for integration testing
-func (m *DeviceManager) ResetBootOk() {
-	m.bootOkRan = false
-	m.bootRevisionsUpdated = false
-}
-
 func (m *DeviceManager) ensureBootOk() error {
 	m.state.Lock()
 	defer m.state.Unlock()
@@ -608,7 +740,7 @@ func (m *DeviceManager) ensureBootOk() error {
 	}
 
 	// boot-ok/update-boot-revision is only relevant in run-mode
-	if m.SystemMode() != "run" {
+	if m.SystemMode(SysAny) != "run" {
 		return nil
 	}
 
@@ -649,6 +781,18 @@ func (m *DeviceManager) ensureCloudInitRestricted() error {
 		return err
 	}
 
+	if !seeded {
+		// we need to wait until we are seeded
+		return nil
+	}
+
+	if release.OnClassic {
+		// don't re-run on classic since classic devices don't get subject to
+		// the cloud-init restricting that core devices do
+		m.cloudInitAlreadyRestricted = true
+		return nil
+	}
+
 	// On Ubuntu Core devices that have been seeded, we want to restrict
 	// cloud-init so that its more dangerous (for an IoT device at least)
 	// features are not exploitable after a device has been seeded. This allows
@@ -658,135 +802,136 @@ func (m *DeviceManager) ensureCloudInitRestricted() error {
 	// boots but disallows arbitrary cloud-init {user,meta,vendor}-data to be
 	// attached to a device via a USB drive and inject code onto the device.
 
-	if seeded && !release.OnClassic {
-		opts := &sysconfig.CloudInitRestrictOptions{}
+	opts := &sysconfig.CloudInitRestrictOptions{}
 
-		// check the current state of cloud-init, if it is disabled or already
-		// restricted then we have nothing to do
-		cloudInitStatus, err := cloudInitStatus()
-		if err != nil {
-			return err
-		}
-		statusMsg := ""
-
-		switch cloudInitStatus {
-		case sysconfig.CloudInitDisabledPermanently, sysconfig.CloudInitRestrictedBySnapd:
-			// already been permanently disabled, nothing to do
-			m.cloudInitAlreadyRestricted = true
-			return nil
-		case sysconfig.CloudInitUntriggered:
-			// hasn't been used
-			statusMsg = "reported to be in disabled state"
-		case sysconfig.CloudInitDone:
-			// is done being used
-			statusMsg = "reported to be done"
-		case sysconfig.CloudInitErrored:
-			// cloud-init errored, so we give the device admin / developer a few
-			// minutes to reboot the machine to re-run cloud-init and try again,
-			// otherwise we will disable cloud-init permanently
-
-			// initialize the time we first saw cloud-init in error state
-			if m.cloudInitErrorAttemptStart == nil {
-				// save the time we started the attempt to restrict
-				now := timeNow()
-				m.cloudInitErrorAttemptStart = &now
-				logger.Noticef("System initialized, cloud-init reported to be in error state, will disable in 3 minutes")
-			}
-
-			// check if 3 minutes have elapsed since we first saw cloud-init in
-			// error state
-			timeSinceFirstAttempt := timeNow().Sub(*m.cloudInitErrorAttemptStart)
-			if timeSinceFirstAttempt <= 3*time.Minute {
-				// we need to keep waiting for cloud-init, up to 3 minutes
-				nextCheck := 3*time.Minute - timeSinceFirstAttempt
-				m.state.EnsureBefore(nextCheck)
-				return nil
-			}
-			// otherwise, we timed out waiting for cloud-init to be fixed or
-			// rebooted and should restrict cloud-init
-			// we will restrict cloud-init below, but we need to force the
-			// disable, as by default RestrictCloudInit will error on state
-			// CloudInitErrored
-			opts.ForceDisable = true
-			statusMsg = "reported to be in error state after 3 minutes"
-		default:
-			// in unknown states we are conservative and let the device run for
-			// a while to see if it transitions to a known state, but eventually
-			// will disable anyways
-			fallthrough
-		case sysconfig.CloudInitEnabled:
-			// we will give cloud-init up to 5 minutes to try and run, if it
-			// still has not transitioned to some other known state, then we
-			// will give up waiting for it and disable it anyways
-
-			// initialize the first time we saw cloud-init in enabled state
-			if m.cloudInitEnabledInactiveAttemptStart == nil {
-				// save the time we started the attempt to restrict
-				now := timeNow()
-				m.cloudInitEnabledInactiveAttemptStart = &now
-			}
-
-			// keep re-scheduling again in 10 seconds until we hit 5 minutes
-			timeSinceFirstAttempt := timeNow().Sub(*m.cloudInitEnabledInactiveAttemptStart)
-			if timeSinceFirstAttempt <= 5*time.Minute {
-				// TODO: should we log a message here about waiting for cloud-init
-				//       to be in a "known state"?
-				m.state.EnsureBefore(10 * time.Second)
-				return nil
-			}
-
-			// otherwise, we gave cloud-init 5 minutes to run, if it's still not
-			// done disable it anyways
-			// note we we need to force the disable, as by default
-			// RestrictCloudInit will error on state CloudInitEnabled
-			opts.ForceDisable = true
-			statusMsg = "failed to transition to done or error state after 5 minutes"
-		}
-
-		// we should always have a model if we are seeded and are not on classic
-		model, err := m.Model()
-		if err != nil {
-			return err
-		}
-
-		// For UC20, we want to always disable cloud-init after it has run on
-		// first boot unless we are in a "real cloud", i.e. not using NoCloud,
-		// or if we installed cloud-init configuration from the gadget
-		if model.Grade() != asserts.ModelGradeUnset {
-			// always disable NoCloud/local datasources after first boot on
-			// uc20, this is because even if the gadget has a cloud.conf
-			// configuring NoCloud, the config installed by cloud-init should
-			// not work differently for later boots, so it's sufficient that
-			// NoCloud runs on first-boot and never again
-			opts.DisableAfterLocalDatasourcesRun = true
-		}
-
-		// now restrict/disable cloud-init
-		res, err := restrictCloudInit(cloudInitStatus, opts)
-		if err != nil {
-			return err
-		}
-
-		// log a message about what we did
-		actionMsg := ""
-		switch res.Action {
-		case "disable":
-			actionMsg = "disabled permanently"
-		case "restrict":
-			// log different messages depending on what datasource was used
-			if res.DataSource == "NoCloud" {
-				actionMsg = "set datasource_list to [ NoCloud ] and disabled auto-import by filesystem label"
-			} else {
-				// all other datasources just log that we limited it to that datasource
-				actionMsg = fmt.Sprintf("set datasource_list to [ %s ]", res.DataSource)
-			}
-		default:
-			return fmt.Errorf("internal error: unexpected action %s taken while restricting cloud-init", res.Action)
-		}
-		logger.Noticef("System initialized, cloud-init %s, %s", statusMsg, actionMsg)
-
-		m.cloudInitAlreadyRestricted = true
+	// check the current state of cloud-init, if it is disabled or already
+	// restricted then we have nothing to do
+	cloudInitStatus, err := cloudInitStatus()
+	if err != nil {
+		return err
 	}
+	statusMsg := ""
+
+	switch cloudInitStatus {
+	case sysconfig.CloudInitDisabledPermanently, sysconfig.CloudInitRestrictedBySnapd:
+		// already been permanently disabled, nothing to do
+		m.cloudInitAlreadyRestricted = true
+		return nil
+	case sysconfig.CloudInitNotFound:
+		// no cloud init at all
+		statusMsg = "not found"
+	case sysconfig.CloudInitUntriggered:
+		// hasn't been used
+		statusMsg = "reported to be in disabled state"
+	case sysconfig.CloudInitDone:
+		// is done being used
+		statusMsg = "reported to be done"
+	case sysconfig.CloudInitErrored:
+		// cloud-init errored, so we give the device admin / developer a few
+		// minutes to reboot the machine to re-run cloud-init and try again,
+		// otherwise we will disable cloud-init permanently
+
+		// initialize the time we first saw cloud-init in error state
+		if m.cloudInitErrorAttemptStart == nil {
+			// save the time we started the attempt to restrict
+			now := timeNow()
+			m.cloudInitErrorAttemptStart = &now
+			logger.Noticef("System initialized, cloud-init reported to be in error state, will disable in 3 minutes")
+		}
+
+		// check if 3 minutes have elapsed since we first saw cloud-init in
+		// error state
+		timeSinceFirstAttempt := timeNow().Sub(*m.cloudInitErrorAttemptStart)
+		if timeSinceFirstAttempt <= 3*time.Minute {
+			// we need to keep waiting for cloud-init, up to 3 minutes
+			nextCheck := 3*time.Minute - timeSinceFirstAttempt
+			m.state.EnsureBefore(nextCheck)
+			return nil
+		}
+		// otherwise, we timed out waiting for cloud-init to be fixed or
+		// rebooted and should restrict cloud-init
+		// we will restrict cloud-init below, but we need to force the
+		// disable, as by default RestrictCloudInit will error on state
+		// CloudInitErrored
+		opts.ForceDisable = true
+		statusMsg = "reported to be in error state after 3 minutes"
+	default:
+		// in unknown states we are conservative and let the device run for
+		// a while to see if it transitions to a known state, but eventually
+		// will disable anyways
+		fallthrough
+	case sysconfig.CloudInitEnabled:
+		// we will give cloud-init up to 5 minutes to try and run, if it
+		// still has not transitioned to some other known state, then we
+		// will give up waiting for it and disable it anyways
+
+		// initialize the first time we saw cloud-init in enabled state
+		if m.cloudInitEnabledInactiveAttemptStart == nil {
+			// save the time we started the attempt to restrict
+			now := timeNow()
+			m.cloudInitEnabledInactiveAttemptStart = &now
+		}
+
+		// keep re-scheduling again in 10 seconds until we hit 5 minutes
+		timeSinceFirstAttempt := timeNow().Sub(*m.cloudInitEnabledInactiveAttemptStart)
+		if timeSinceFirstAttempt <= 5*time.Minute {
+			// TODO: should we log a message here about waiting for cloud-init
+			//       to be in a "known state"?
+			m.state.EnsureBefore(10 * time.Second)
+			return nil
+		}
+
+		// otherwise, we gave cloud-init 5 minutes to run, if it's still not
+		// done disable it anyways
+		// note we we need to force the disable, as by default
+		// RestrictCloudInit will error on state CloudInitEnabled
+		opts.ForceDisable = true
+		statusMsg = "failed to transition to done or error state after 5 minutes"
+	}
+
+	// we should always have a model if we are seeded and are not on classic
+	model, err := m.Model()
+	if err != nil {
+		return err
+	}
+
+	// For UC20, we want to always disable cloud-init after it has run on
+	// first boot unless we are in a "real cloud", i.e. not using NoCloud,
+	// or if we installed cloud-init configuration from the gadget
+	if model.Grade() != asserts.ModelGradeUnset {
+		// always disable NoCloud/local datasources after first boot on
+		// uc20, this is because even if the gadget has a cloud.conf
+		// configuring NoCloud, the config installed by cloud-init should
+		// not work differently for later boots, so it's sufficient that
+		// NoCloud runs on first-boot and never again
+		opts.DisableAfterLocalDatasourcesRun = true
+	}
+
+	// now restrict/disable cloud-init
+	res, err := restrictCloudInit(cloudInitStatus, opts)
+	if err != nil {
+		return err
+	}
+
+	// log a message about what we did
+	actionMsg := ""
+	switch res.Action {
+	case "disable":
+		actionMsg = "disabled permanently"
+	case "restrict":
+		// log different messages depending on what datasource was used
+		if res.DataSource == "NoCloud" {
+			actionMsg = "set datasource_list to [ NoCloud ] and disabled auto-import by filesystem label"
+		} else {
+			// all other datasources just log that we limited it to that datasource
+			actionMsg = fmt.Sprintf("set datasource_list to [ %s ]", res.DataSource)
+		}
+	default:
+		return fmt.Errorf("internal error: unexpected action %s taken while restricting cloud-init", res.Action)
+	}
+	logger.Noticef("System initialized, cloud-init %s, %s", statusMsg, actionMsg)
+
+	m.cloudInitAlreadyRestricted = true
 
 	return nil
 }
@@ -803,7 +948,7 @@ func (m *DeviceManager) ensureInstalled() error {
 		return nil
 	}
 
-	if m.SystemMode() != "install" {
+	if m.SystemMode(SysHasModeenv) != "install" {
 		return nil
 	}
 
@@ -820,15 +965,66 @@ func (m *DeviceManager) ensureInstalled() error {
 		return nil
 	}
 
+	perfTimings := timings.New(map[string]string{"ensure": "install-system"})
+
+	model, err := m.Model()
+	if err != nil && err != state.ErrNoState {
+		return err
+	}
+	if err != nil {
+		return fmt.Errorf("internal error: core device brand and model are set but there is no model assertion")
+	}
+
+	// check if the gadget has an install-device hook
+	var hasInstallDeviceHook bool
+
+	gadgetInfo, err := snapstate.CurrentInfo(m.state, model.Gadget())
+	if err != nil {
+		return fmt.Errorf("internal error: device is seeded in install mode but has no gadget snap: %v", err)
+	}
+	hasInstallDeviceHook = (gadgetInfo.Hooks["install-device"] != nil)
+
 	m.ensureInstalledRan = true
 
-	tasks := []*state.Task{}
+	var prev *state.Task
 	setupRunSystem := m.state.NewTask("setup-run-system", i18n.G("Setup system for run mode"))
-	tasks = append(tasks, setupRunSystem)
+
+	tasks := []*state.Task{setupRunSystem}
+	addTask := func(t *state.Task) {
+		t.WaitFor(prev)
+		tasks = append(tasks, t)
+		prev = t
+	}
+	prev = setupRunSystem
+
+	// add the install-device hook before ensure-next-boot-to-run-mode if it
+	// exists in the snap
+	var installDevice *state.Task
+	if hasInstallDeviceHook {
+		summary := i18n.G("Run install-device hook")
+		hooksup := &hookstate.HookSetup{
+			// TODO: what's a reasonable timeout for the install-device hook?
+			Snap: model.Gadget(),
+			Hook: "install-device",
+		}
+		installDevice = hookstate.HookTask(m.state, summary, hooksup, nil)
+		addTask(installDevice)
+	}
+
+	restartSystem := m.state.NewTask("restart-system-to-run-mode", i18n.G("Ensure next boot to run mode"))
+	addTask(restartSystem)
+
+	if installDevice != nil {
+		// reference used by snapctl reboot
+		installDevice.Set("restart-task", restartSystem.ID())
+	}
 
 	chg := m.state.NewChange("install-system", i18n.G("Install the system"))
 	chg.AddAll(state.NewTaskSet(tasks...))
 
+	state.TagTimingsWithChange(perfTimings, chg)
+	perfTimings.Save(m.state)
+
 	return nil
 }
 
@@ -911,6 +1107,73 @@ func (m *DeviceManager) ensureSeedInConfig() error {
 
 }
 
+func (m *DeviceManager) appendTriedRecoverySystem(label string) error {
+	// state is locked by the caller
+
+	var triedSystems []string
+	if err := m.state.Get("tried-systems", &triedSystems); err != nil && err != state.ErrNoState {
+		return err
+	}
+	if strutil.ListContains(triedSystems, label) {
+		// system already recorded as tried?
+		return nil
+	}
+	triedSystems = append(triedSystems, label)
+	m.state.Set("tried-systems", triedSystems)
+	return nil
+}
+
+func (m *DeviceManager) ensureTriedRecoverySystem() error {
+	if release.OnClassic {
+		return nil
+	}
+	// nothing to do if not UC20 and run mode
+	if m.SystemMode(SysHasModeenv) != "run" {
+		return nil
+	}
+	if m.ensureTriedRecoverySystemRan {
+		return nil
+	}
+
+	m.state.Lock()
+	defer m.state.Unlock()
+
+	deviceCtx, err := DeviceCtx(m.state, nil, nil)
+	if err != nil && err != state.ErrNoState {
+		return err
+	}
+	outcome, label, err := boot.InspectTryRecoverySystemOutcome(deviceCtx)
+	if err != nil {
+		if !boot.IsInconsistentRecoverySystemState(err) {
+			return err
+		}
+		// boot variables state was inconsistent
+		logger.Noticef("tried recovery system outcome error: %v", err)
+	}
+	switch outcome {
+	case boot.TryRecoverySystemOutcomeSuccess:
+		logger.Noticef("tried recovery system %q was successful", label)
+		if err := m.appendTriedRecoverySystem(label); err != nil {
+			return err
+		}
+	case boot.TryRecoverySystemOutcomeFailure:
+		logger.Noticef("tried recovery system %q failed", label)
+	case boot.TryRecoverySystemOutcomeInconsistent:
+		logger.Noticef("inconsistent outcome of a tried recovery system")
+	case boot.TryRecoverySystemOutcomeNoneTried:
+		// no system was tried
+	}
+	if outcome != boot.TryRecoverySystemOutcomeNoneTried {
+		if err := boot.ClearTryRecoverySystem(deviceCtx, label); err != nil {
+			logger.Noticef("cannot clear tried recovery system status: %v", err)
+			return err
+		}
+	}
+
+	m.ensureTriedRecoverySystemRan = true
+	return nil
+}
+
 type ensureError struct {
 	errs []error
 }
@@ -960,6 +1223,10 @@ func (m *DeviceManager) Ensure() error {
 		if err := m.ensureInstalled(); err != nil {
 			errs = append(errs, err)
 		}
+
+		if err := m.ensureTriedRecoverySystem(); err != nil {
+			errs = append(errs, err)
+		}
 	}
 
 	if len(errs) > 0 {
@@ -969,6 +1236,14 @@ func (m *DeviceManager) Ensure() error {
 	return nil
 }
 
+// ResetToPostBootState is only useful for integration testing.
+func (m *DeviceManager) ResetToPostBootState() {
+	osutil.MustBeTestBinary("ResetToPostBootState can only be called from tests")
+	m.bootOkRan = false
+	m.bootRevisionsUpdated = false
+	m.ensureTriedRecoverySystemRan = false
+}
+
 var errNoSaveSupport = errors.New("no save directory before UC20")
 
 // withSaveDir invokes a function making sure save dir is available.
@@ -1105,6 +1380,52 @@ func (m *DeviceManager) Serial() (*asserts.Serial, error) {
 	return findSerial(m.state, nil)
 }
 
+type SystemModeInfo struct {
+	Mode              string
+	HasModeenv        bool
+	Seeded            bool
+	BootFlags         []string
+	HostDataLocations []string
+}
+
+// SystemModeInfo returns details about the current system mode the device is in.
+func (m *DeviceManager) SystemModeInfo() (*SystemModeInfo, error) {
+	deviceCtx, err := DeviceCtx(m.state, nil, nil)
+	if err == state.ErrNoState {
+		return nil, fmt.Errorf("cannot report system mode information before device model is acknowledged")
+	}
+	if err != nil {
+		return nil, err
+	}
+
+	var seeded bool
+	err = m.state.Get("seeded", &seeded)
+	if err != nil && err != state.ErrNoState {
+		return nil, err
+	}
+
+	mode := deviceCtx.SystemMode()
+	smi := SystemModeInfo{
+		Mode:       mode,
+		HasModeenv: deviceCtx.HasModeenv(),
+		Seeded:     seeded,
+	}
+	if smi.HasModeenv {
+		bootFlags, err := boot.BootFlags(deviceCtx)
+		if err != nil {
+			return nil, err
+		}
+		smi.BootFlags = bootFlags
+
+		hostDataLocs, err := boot.HostUbuntuDataForMode(mode)
+		if err != nil {
+			return nil, err
+		}
+		smi.HostDataLocations = hostDataLocs
+	}
+	return &smi, nil
+}
+
 type SystemAction struct {
 	Title string
 	Mode  string
@@ -1143,7 +1464,7 @@ var ErrNoSystems = errors.New("no systems seeds")
 // systems, ErrNoSystems when no systems seeds were found or other error.
 func (m *DeviceManager) Systems() ([]*System, error) {
 	// it's tough luck when we cannot determine the current system seed
-	systemMode := m.SystemMode()
+	systemMode := m.SystemMode(SysAny)
 	currentSys, _ := currentSystemForMode(m.state, systemMode)
 
 	systemLabels, err := filepath.Glob(filepath.Join(dirs.SnapSeedDir, "systems", "*"))
@@ -1199,7 +1520,7 @@ func (m *DeviceManager) Reboot(systemLabel, mode string) error {
 
 	// no systemLabel means "current" so get the current system label
 	if systemLabel == "" {
-		systemMode := m.SystemMode()
+		systemMode := m.SystemMode(SysAny)
 		currentSys, err := currentSystemForMode(m.state, systemMode)
 		if err != nil {
 			return fmt.Errorf("cannot get current system: %v", err)
@@ -1243,7 +1564,7 @@ func (m *DeviceManager) switchToSystemAndMode(systemLabel, mode string, sameSyst
 		return err
 	}
 
-	systemMode := m.SystemMode()
+	systemMode := m.SystemMode(SysAny)
 	// ignore the error to be robust in scenarios that
 	// dont' stricly require currentSys to be carried through.
 	// make sure that currentSys == nil does not break
@@ -1381,7 +1702,7 @@ func (m *DeviceManager) hasFDESetupHook() (bool, error) {
 	return hasFDESetupHookInKernel(kernelInfo), nil
 }
 
-func (m *DeviceManager) runFDESetupHook(op string, params *boot.FDESetupHookParams) ([]byte, error) {
+func (m *DeviceManager) runFDESetupHook(req *fde.SetupRequest) ([]byte, error) {
 	// TODO:UC20: when this runs on refresh we need to be very careful
 	// that we never run this when the kernel is not fully configured
 	// i.e. when there are no security profiles for the hook
@@ -1404,21 +1725,6 @@ func (m *DeviceManager) runFDESetupHook(op string, params *boot.FDESetupHookPara
 		// XXX: should this be configurable somehow?
 		Timeout: 5 * time.Minute,
 	}
-	req := &fde.SetupRequest{
-		Op:      op,
-		Key:     params.Key[:],
-		KeyName: params.KeyName,
-		// TODO: include boot chains
-	}
-	for _, model := range params.Models {
-		req.Models = append(req.Models, map[string]string{
-			"series":     model.Series(),
-			"brand-id":   model.BrandID(),
-			"model":      model.Model(),
-			"grade":      string(model.Grade()),
-			"signkey-id": model.SignKeyID(),
-		})
-	}
 	contextData := map[string]interface{}{
 		"fde-setup-request": req,
 	}
@@ -1426,44 +1732,53 @@ func (m *DeviceManager) runFDESetupHook(op string, params *boot.FDESetupHookPara
 	defer st.Lock()
 	context, err := m.hookMgr.EphemeralRunHook(context.Background(), hooksup, contextData)
 	if err != nil {
-		return nil, fmt.Errorf("cannot run hook for %q: %v", op, err)
+		return nil, fmt.Errorf("cannot run hook for %q: %v", req.Op, err)
 	}
 	// the hook is expected to call "snapctl fde-setup-result" which
-	// wil set the "fde-setup-result" value on the task
-	var hookResult []byte
+	// will set the "fde-setup-result" value on the task
+	var hookOutput []byte
 	context.Lock()
-	err = context.Get("fde-setup-result", &hookResult)
+	err = context.Get("fde-setup-result", &hookOutput)
 	context.Unlock()
 	if err != nil {
-		return nil, fmt.Errorf("cannot get result from fde-setup hook %q: %v", op, err)
+		return nil, fmt.Errorf("cannot get result from fde-setup hook %q: %v", req.Op, err)
 	}
-
-	return hookResult, nil
+	return hookOutput, nil
 }
 
-func (m *DeviceManager) checkFDEFeatures(st *state.State) error {
+func (m *DeviceManager) checkFDEFeatures() (et secboot.EncryptionType, err error) {
+	// TODO: move most of this to kernel/fde.Features
 	// Run fde-setup hook with "op":"features". If the hook
 	// returns any {"features":[...]} reply we consider the
 	// hardware supported. If the hook errors or if it returns
 	// {"error":"hardware-unsupported"} we don't.
-	output, err := m.runFDESetupHook("features", &boot.FDESetupHookParams{})
+	req := &fde.SetupRequest{
+		Op: "features",
+	}
+	output, err := m.runFDESetupHook(req)
 	if err != nil {
-		return err
+		return et, err
 	}
 	var res struct {
 		Features []string `json:"features"`
 		Error    string   `json:"error"`
 	}
 	if err := json.Unmarshal(output, &res); err != nil {
-		return fmt.Errorf("cannot parse hook output %q: %v", output, err)
+		return et, fmt.Errorf("cannot parse hook output %q: %v", output, err)
 	}
 	if res.Features == nil && res.Error == "" {
-		return fmt.Errorf(`cannot use hook: neither "features" nor "error" returned`)
+		return et, fmt.Errorf(`cannot use hook: neither "features" nor "error" returned`)
 	}
 	if res.Error != "" {
-		return fmt.Errorf("cannot use hook: it returned error: %v", res.Error)
+		return et, fmt.Errorf("cannot use hook: it returned error: %v", res.Error)
 	}
-	return nil
+	if strutil.ListContains(res.Features, "device-setup") {
+		et = secboot.EncryptionTypeDeviceSetupHook
+	} else {
+		et = secboot.EncryptionTypeLUKS
+	}
+
+	return et, nil
 }
 
 func hasFDESetupHookInKernel(kernelInfo *snap.Info) bool {
@@ -1487,6 +1802,6 @@ func (h fdeSetupHandler) Done() error {
 	return nil
 }
 
-func (h fdeSetupHandler) Error(err error) error {
-	return nil
+func (h fdeSetupHandler) Error(err error) (bool, error) {
+	return false, nil
 }
