@@ -20,17 +20,17 @@
 package hookstate_test
 
 import (
+	"context"
 	"encoding/json"
 	"fmt"
-	"os"
 	"path/filepath"
 	"regexp"
-	"sort"
 	"sync/atomic"
 	"testing"
 	"time"
 
 	. "gopkg.in/check.v1"
+	"gopkg.in/tomb.v2"
 
 	"github.com/snapcore/snapd/dirs"
 	"github.com/snapcore/snapd/overlord"
@@ -46,11 +46,12 @@ import (
 
 func TestHookManager(t *testing.T) { TestingT(t) }
 
-type hookManagerSuite struct {
+type baseHookManagerSuite struct {
 	testutil.BaseTest
 
 	o           *overlord.Overlord
 	state       *state.State
+	se          *overlord.StateEngine
 	manager     *hookstate.HookManager
 	context     *hookstate.Context
 	mockHandler *hooktest.MockHandler
@@ -59,6 +60,87 @@ type hookManagerSuite struct {
 	command     *testutil.MockCmd
 }
 
+var (
+	settleTimeout = testutil.HostScaledTimeout(15 * time.Second)
+)
+
+func (s *baseHookManagerSuite) commonSetUpTest(c *C) {
+	s.BaseTest.SetUpTest(c)
+
+	hooktype1 := snap.NewHookType(regexp.MustCompile("^do-something$"))
+	hooktype2 := snap.NewHookType(regexp.MustCompile("^undo-something$"))
+	s.AddCleanup(snap.MockAppendSupportedHookTypes([]*snap.HookType{hooktype1, hooktype2}))
+
+	dirs.SetRootDir(c.MkDir())
+	s.o = overlord.Mock()
+	s.state = s.o.State()
+	manager, err := hookstate.Manager(s.state, s.o.TaskRunner())
+	c.Assert(err, IsNil)
+	s.manager = manager
+	s.se = s.o.StateEngine()
+	s.o.AddManager(s.manager)
+	s.o.AddManager(s.o.TaskRunner())
+	c.Assert(s.o.StartUp(), IsNil)
+
+	s.AddCleanup(snap.MockSanitizePlugsSlots(func(snapInfo *snap.Info) {}))
+
+	s.command = testutil.MockCommand(c, "snap", "")
+	s.AddCleanup(s.command.Restore)
+
+	s.context = nil
+	s.mockHandler = hooktest.NewMockHandler()
+	s.manager.Register(regexp.MustCompile("configure"), func(context *hookstate.Context) hookstate.Handler {
+		s.context = context
+		return s.mockHandler
+	})
+	s.AddCleanup(hookstate.MockErrtrackerReport(func(string, string, string, map[string]string) (string, error) {
+		return "", nil
+	}))
+}
+
+func (s *baseHookManagerSuite) commonTearDownTest(c *C) {
+	s.BaseTest.TearDownTest(c)
+
+	s.manager.StopHooks()
+	s.se.Stop()
+	dirs.SetRootDir("")
+}
+
+func (s *baseHookManagerSuite) setUpSnap(c *C, instanceName string, yaml string) {
+	hooksup := &hookstate.HookSetup{
+		Snap:     instanceName,
+		Hook:     "configure",
+		Revision: snap.R(1),
+	}
+
+	initialContext := map[string]interface{}{
+		"test-key": "test-value",
+	}
+
+	s.state.Lock()
+	defer s.state.Unlock()
+	s.task = hookstate.HookTask(s.state, "test summary", hooksup, initialContext)
+	c.Assert(s.task, NotNil, Commentf("Expected HookTask to return a task"))
+
+	s.change = s.state.NewChange("kind", "summary")
+	s.change.AddTask(s.task)
+
+	snapName, instanceKey := snap.SplitInstanceName(instanceName)
+
+	sideInfo := &snap.SideInfo{RealName: snapName, SnapID: "some-snap-id", Revision: snap.R(1)}
+	snaptest.MockSnapInstance(c, instanceName, yaml, sideInfo)
+	snapstate.Set(s.state, instanceName, &snapstate.SnapState{
+		Active:      true,
+		Sequence:    []*snap.SideInfo{sideInfo},
+		Current:     snap.R(1),
+		InstanceKey: instanceKey,
+	})
+}
+
+type hookManagerSuite struct {
+	baseHookManagerSuite
+}
+
 var _ = Suite(&hookManagerSuite{})
 
 var snapYaml = `
@@ -67,6 +149,8 @@ version: 1.0
 hooks:
     configure:
     prepare-device:
+    do-something:
+    undo-something:
 `
 
 var snapYaml1 = `
@@ -84,79 +168,23 @@ hooks:
 `
 
 func (s *hookManagerSuite) SetUpTest(c *C) {
-	s.BaseTest.SetUpTest(c)
+	s.commonSetUpTest(c)
 
-	dirs.SetRootDir(c.MkDir())
-	s.o = overlord.Mock()
-	s.state = s.o.State()
-	manager, err := hookstate.Manager(s.state)
-	c.Assert(err, IsNil)
-	s.manager = manager
-	s.o.AddManager(s.manager)
-
-	s.AddCleanup(snap.MockSanitizePlugsSlots(func(snapInfo *snap.Info) {}))
-
-	hooksup := &hookstate.HookSetup{
-		Snap:     "test-snap",
-		Hook:     "configure",
-		Revision: snap.R(1),
-	}
-
-	initialContext := map[string]interface{}{
-		"test-key": "test-value",
-	}
-
-	s.state.Lock()
-	s.task = hookstate.HookTask(s.state, "test summary", hooksup, initialContext)
-	c.Assert(s.task, NotNil, Commentf("Expected HookTask to return a task"))
-
-	s.change = s.state.NewChange("kind", "summary")
-	s.change.AddTask(s.task)
-
-	sideInfo := &snap.SideInfo{RealName: "test-snap", SnapID: "some-snap-id", Revision: snap.R(1)}
-	snaptest.MockSnap(c, snapYaml, sideInfo)
-	snapstate.Set(s.state, "test-snap", &snapstate.SnapState{
-		Active:   true,
-		Sequence: []*snap.SideInfo{sideInfo},
-		Current:  snap.R(1),
-	})
-	s.state.Unlock()
-
-	s.command = testutil.MockCommand(c, "snap", "")
-	s.AddCleanup(s.command.Restore)
-
-	s.context = nil
-	s.mockHandler = hooktest.NewMockHandler()
-	s.manager.Register(regexp.MustCompile("configure"), func(context *hookstate.Context) hookstate.Handler {
-		s.context = context
-		return s.mockHandler
-	})
-	s.AddCleanup(hookstate.MockErrtrackerReport(func(string, string, string, map[string]string) (string, error) {
-		return "", nil
-	}))
+	s.setUpSnap(c, "test-snap", snapYaml)
 }
 
 func (s *hookManagerSuite) TearDownTest(c *C) {
-	s.BaseTest.TearDownTest(c)
-
-	s.manager.Stop()
-	dirs.SetRootDir("")
+	s.commonTearDownTest(c)
 }
 
 func (s *hookManagerSuite) settle(c *C) {
-	err := s.o.Settle(5 * time.Second)
+	err := s.o.Settle(settleTimeout)
 	c.Assert(err, IsNil)
 }
 
 func (s *hookManagerSuite) TestSmoke(c *C) {
-	s.manager.Ensure()
-	s.manager.Wait()
-}
-
-func (s *hookManagerSuite) TestKnownTaskKinds(c *C) {
-	kinds := s.manager.KnownTaskKinds()
-	sort.Strings(kinds)
-	c.Assert(kinds, DeepEquals, []string{"configure-snapd", "run-hook"})
+	s.se.Ensure()
+	s.se.Wait()
 }
 
 func (s *hookManagerSuite) TestHookSetupJsonMarshal(c *C) {
@@ -207,20 +235,20 @@ func (s *hookManagerSuite) TestHookTaskEnsure(c *C) {
 			didRun <- s.manager.GracefullyWaitRunningHooks()
 		}()
 	}
-	s.manager.Ensure()
+	s.se.Ensure()
 	select {
 	case ok := <-didRun:
 		c.Check(ok, Equals, true)
 	case <-time.After(5 * time.Second):
 		c.Fatal("hook run should have been done by now")
 	}
-	s.manager.Wait()
+	s.se.Wait()
 
 	s.state.Lock()
 	defer s.state.Unlock()
 
 	c.Assert(s.context, NotNil, Commentf("Expected handler generator to be called with a valid context"))
-	c.Check(s.context.SnapName(), Equals, "test-snap")
+	c.Check(s.context.InstanceName(), Equals, "test-snap")
 	c.Check(s.context.SnapRevision(), Equals, snap.R(1))
 	c.Check(s.context.HookName(), Equals, "configure")
 
@@ -243,8 +271,8 @@ func (s *hookManagerSuite) TestHookTaskEnsureRestarting(c *C) {
 	// we do no start new hooks runs if we are restarting
 	s.state.RequestRestart(state.RestartDaemon)
 
-	s.manager.Ensure()
-	s.manager.Wait()
+	s.se.Ensure()
+	s.se.Wait()
 
 	s.state.Lock()
 	defer s.state.Unlock()
@@ -268,8 +296,8 @@ func (s *hookManagerSuite) TestHookSnapMissing(c *C) {
 	snapstate.Set(s.state, "test-snap", nil)
 	s.state.Unlock()
 
-	s.manager.Ensure()
-	s.manager.Wait()
+	s.se.Ensure()
+	s.se.Wait()
 
 	s.state.Lock()
 	defer s.state.Unlock()
@@ -289,8 +317,8 @@ func (s *hookManagerSuite) TestHookHijackingHappy(c *C) {
 		return nil
 	})
 
-	s.manager.Ensure()
-	s.manager.Wait()
+	s.se.Ensure()
+	s.se.Wait()
 
 	s.state.Lock()
 	defer s.state.Unlock()
@@ -299,7 +327,7 @@ func (s *hookManagerSuite) TestHookHijackingHappy(c *C) {
 	c.Check(s.command.Calls(), HasLen, 0)
 
 	c.Assert(s.context, NotNil)
-	c.Check(s.context.SnapName(), Equals, "test-snap")
+	c.Check(s.context.InstanceName(), Equals, "test-snap")
 	c.Check(s.context.SnapRevision(), Equals, snap.R(1))
 	c.Check(s.context.HookName(), Equals, "configure")
 
@@ -317,8 +345,8 @@ func (s *hookManagerSuite) TestHookHijackingUnHappy(c *C) {
 		return fmt.Errorf("not-happy-at-all")
 	})
 
-	s.manager.Ensure()
-	s.manager.Wait()
+	s.se.Ensure()
+	s.se.Wait()
 
 	s.state.Lock()
 	defer s.state.Unlock()
@@ -326,7 +354,7 @@ func (s *hookManagerSuite) TestHookHijackingUnHappy(c *C) {
 	c.Check(s.command.Calls(), HasLen, 0)
 
 	c.Assert(s.context, NotNil)
-	c.Check(s.context.SnapName(), Equals, "test-snap")
+	c.Check(s.context.InstanceName(), Equals, "test-snap")
 	c.Check(s.context.SnapRevision(), Equals, snap.R(1))
 	c.Check(s.context.HookName(), Equals, "configure")
 
@@ -348,8 +376,8 @@ func (s *hookManagerSuite) TestHookHijackingVeryUnHappy(c *C) {
 }
 
 func (s *hookManagerSuite) TestHookTaskInitializesContext(c *C) {
-	s.manager.Ensure()
-	s.manager.Wait()
+	s.se.Ensure()
+	s.se.Wait()
 
 	var value string
 	c.Assert(s.context, NotNil, Commentf("Expected handler generator to be called with a valid context"))
@@ -365,8 +393,8 @@ func (s *hookManagerSuite) TestHookTaskHandlesHookError(c *C) {
 		c, "snap", ">&2 echo 'hook failed at user request'; exit 1")
 	defer cmd.Restore()
 
-	s.manager.Ensure()
-	s.manager.Wait()
+	s.se.Ensure()
+	s.se.Wait()
 
 	s.state.Lock()
 	defer s.state.Unlock()
@@ -396,8 +424,8 @@ func (s *hookManagerSuite) TestHookTaskHandleIgnoreErrorWorks(c *C) {
 		c, "snap", ">&2 echo 'hook failed at user request'; exit 1")
 	defer cmd.Restore()
 
-	s.manager.Ensure()
-	s.manager.Wait()
+	s.se.Ensure()
+	s.se.Wait()
 
 	s.state.Lock()
 	defer s.state.Unlock()
@@ -412,6 +440,33 @@ func (s *hookManagerSuite) TestHookTaskHandleIgnoreErrorWorks(c *C) {
 	checkTaskLogContains(c, s.task, ".*ignoring failure in hook.*")
 }
 
+func (s *hookManagerSuite) TestHookTaskHandlesHookErrorAndIgnoresIt(c *C) {
+	// tell the mock handler to return 'true' from its Error() handler,
+	// indicating to the hookmgr to ignore the original hook error.
+	s.mockHandler.IgnoreOriginalErr = true
+
+	// Simulate hook error
+	cmd := testutil.MockCommand(
+		c, "snap", ">&2 echo 'hook failed at user request'; exit 1")
+	defer cmd.Restore()
+
+	s.se.Ensure()
+	s.se.Wait()
+
+	s.state.Lock()
+	defer s.state.Unlock()
+
+	c.Check(s.mockHandler.BeforeCalled, Equals, true)
+	c.Check(s.mockHandler.DoneCalled, Equals, false)
+	c.Check(s.mockHandler.ErrorCalled, Equals, true)
+
+	c.Check(s.task.Kind(), Equals, "run-hook")
+	c.Check(s.task.Status(), Equals, state.DoneStatus)
+	c.Check(s.change.Status(), Equals, state.DoneStatus)
+
+	c.Check(s.manager.NumRunningHooks(), Equals, 0)
+}
+
 func (s *hookManagerSuite) TestHookTaskEnforcesTimeout(c *C) {
 	var hooksup hookstate.HookSetup
 
@@ -425,17 +480,8 @@ func (s *hookManagerSuite) TestHookTaskEnforcesTimeout(c *C) {
 	cmd := testutil.MockCommand(c, "snap", "while true; do sleep 1; done")
 	defer cmd.Restore()
 
-	s.manager.Ensure()
-	completed := make(chan struct{})
-	go func() {
-		s.manager.Wait()
-		close(completed)
-	}()
-
-	s.state.Lock()
-	s.state.Unlock()
-	s.manager.Ensure()
-	<-completed
+	s.se.Ensure()
+	s.se.Wait()
 
 	s.state.Lock()
 	defer s.state.Unlock()
@@ -459,8 +505,8 @@ func (s *hookManagerSuite) TestHookTaskEnforcesDefaultTimeout(c *C) {
 	cmd := testutil.MockCommand(c, "snap", "while true; do sleep 1; done")
 	defer cmd.Restore()
 
-	s.manager.Ensure()
-	s.manager.Wait()
+	s.se.Ensure()
+	s.se.Wait()
 
 	s.state.Lock()
 	defer s.state.Unlock()
@@ -490,17 +536,8 @@ func (s *hookManagerSuite) TestHookTaskEnforcedTimeoutWithIgnoreError(c *C) {
 	cmd := testutil.MockCommand(c, "snap", "while true; do sleep 1; done")
 	defer cmd.Restore()
 
-	s.manager.Ensure()
-	completed := make(chan struct{})
-	go func() {
-		s.manager.Wait()
-		close(completed)
-	}()
-
-	s.state.Lock()
-	s.state.Unlock()
-	s.manager.Ensure()
-	<-completed
+	s.se.Ensure()
+	s.se.Wait()
 
 	s.state.Lock()
 	defer s.state.Unlock()
@@ -521,20 +558,15 @@ func (s *hookManagerSuite) TestHookTaskCanKillHook(c *C) {
 	cmd := testutil.MockCommand(c, "snap", "while true; do sleep 1; done")
 	defer cmd.Restore()
 
-	s.manager.Ensure()
-	completed := make(chan struct{})
-	go func() {
-		s.manager.Wait()
-		close(completed)
-	}()
+	s.se.Ensure()
 
-	// Abort the change, which should kill the hanging hook, and wait for the
-	// task to complete.
+	// Abort the change, which should kill the hanging hook, and
+	// wait for the task to complete.
 	s.state.Lock()
 	s.change.Abort()
 	s.state.Unlock()
-	s.manager.Ensure()
-	<-completed
+	s.se.Ensure()
+	s.se.Wait()
 
 	s.state.Lock()
 	defer s.state.Unlock()
@@ -559,8 +591,8 @@ func (s *hookManagerSuite) TestHookTaskCorrectlyIncludesContext(c *C) {
 		c, "snap", ">&2 echo \"SNAP_COOKIE=$SNAP_COOKIE\"; exit 1")
 	defer cmd.Restore()
 
-	s.manager.Ensure()
-	s.manager.Wait()
+	s.se.Ensure()
+	s.se.Wait()
 
 	s.state.Lock()
 	defer s.state.Unlock()
@@ -578,8 +610,8 @@ func (s *hookManagerSuite) TestHookTaskCorrectlyIncludesContext(c *C) {
 func (s *hookManagerSuite) TestHookTaskHandlerBeforeError(c *C) {
 	s.mockHandler.BeforeError = true
 
-	s.manager.Ensure()
-	s.manager.Wait()
+	s.se.Ensure()
+	s.se.Wait()
 
 	s.state.Lock()
 	defer s.state.Unlock()
@@ -597,8 +629,8 @@ func (s *hookManagerSuite) TestHookTaskHandlerBeforeError(c *C) {
 func (s *hookManagerSuite) TestHookTaskHandlerDoneError(c *C) {
 	s.mockHandler.DoneError = true
 
-	s.manager.Ensure()
-	s.manager.Wait()
+	s.se.Ensure()
+	s.se.Wait()
 
 	s.state.Lock()
 	defer s.state.Unlock()
@@ -620,8 +652,8 @@ func (s *hookManagerSuite) TestHookTaskHandlerErrorError(c *C) {
 	cmd := testutil.MockCommand(c, "snap", "exit 1")
 	defer cmd.Restore()
 
-	s.manager.Ensure()
-	s.manager.Wait()
+	s.se.Ensure()
+	s.se.Wait()
 
 	s.state.Lock()
 	defer s.state.Unlock()
@@ -636,6 +668,67 @@ func (s *hookManagerSuite) TestHookTaskHandlerErrorError(c *C) {
 	checkTaskLogContains(c, s.task, `.*Error failed at user request.*`)
 }
 
+func (s *hookManagerSuite) TestHookUndoRunsOnError(c *C) {
+	handler := hooktest.NewMockHandler()
+	undoHandler := hooktest.NewMockHandler()
+
+	s.manager.Register(regexp.MustCompile("^do-something$"), func(context *hookstate.Context) hookstate.Handler {
+		return handler
+	})
+	s.manager.Register(regexp.MustCompile("^undo-something$"), func(context *hookstate.Context) hookstate.Handler {
+		return undoHandler
+	})
+
+	hooksup := &hookstate.HookSetup{
+		Snap:     "test-snap",
+		Hook:     "do-something",
+		Revision: snap.R(1),
+	}
+	undohooksup := &hookstate.HookSetup{
+		Snap:     "test-snap",
+		Hook:     "undo-something",
+		Revision: snap.R(1),
+	}
+
+	// use unknown hook to fail the change
+	failinghooksup := &hookstate.HookSetup{
+		Snap:     "test-snap",
+		Hook:     "unknown-hook",
+		Revision: snap.R(1),
+	}
+
+	initialContext := map[string]interface{}{}
+
+	s.state.Lock()
+	task := hookstate.HookTaskWithUndo(s.state, "test summary", hooksup, undohooksup, initialContext)
+	c.Assert(task, NotNil)
+	failtask := hookstate.HookTask(s.state, "test summary", failinghooksup, initialContext)
+	failtask.WaitFor(task)
+
+	change := s.state.NewChange("kind", "summary")
+	change.AddTask(task)
+	change.AddTask(failtask)
+	s.state.Unlock()
+
+	s.settle(c)
+
+	s.state.Lock()
+	defer s.state.Unlock()
+
+	c.Check(handler.BeforeCalled, Equals, true)
+	c.Check(handler.DoneCalled, Equals, true)
+	c.Check(handler.ErrorCalled, Equals, false)
+
+	c.Check(undoHandler.BeforeCalled, Equals, true)
+	c.Check(undoHandler.DoneCalled, Equals, true)
+	c.Check(undoHandler.ErrorCalled, Equals, false)
+
+	c.Check(task.Status(), Equals, state.UndoneStatus)
+	c.Check(change.Status(), Equals, state.ErrorStatus)
+
+	c.Check(s.manager.NumRunningHooks(), Equals, 0)
+}
+
 func (s *hookManagerSuite) TestHookWithoutHandlerIsError(c *C) {
 	hooksup := &hookstate.HookSetup{
 		Snap:     "test-snap",
@@ -646,8 +739,8 @@ func (s *hookManagerSuite) TestHookWithoutHandlerIsError(c *C) {
 	s.task.Set("hook-setup", hooksup)
 	s.state.Unlock()
 
-	s.manager.Ensure()
-	s.manager.Wait()
+	s.se.Ensure()
+	s.se.Wait()
 
 	s.state.Lock()
 	defer s.state.Unlock()
@@ -664,8 +757,8 @@ func (s *hookManagerSuite) TestHookWithMultipleHandlersIsError(c *C) {
 		return hooktest.NewMockHandler()
 	})
 
-	s.manager.Ensure()
-	s.manager.Wait()
+	s.se.Ensure()
+	s.se.Wait()
 
 	s.state.Lock()
 	defer s.state.Unlock()
@@ -686,8 +779,8 @@ func (s *hookManagerSuite) TestHookWithoutHookIsError(c *C) {
 	s.task.Set("hook-setup", hooksup)
 	s.state.Unlock()
 
-	s.manager.Ensure()
-	s.manager.Wait()
+	s.se.Ensure()
+	s.se.Wait()
 
 	s.state.Lock()
 	defer s.state.Unlock()
@@ -712,8 +805,42 @@ func (s *hookManagerSuite) TestHookWithoutHookOptional(c *C) {
 	s.task.Set("hook-setup", hooksup)
 	s.state.Unlock()
 
-	s.manager.Ensure()
-	s.manager.Wait()
+	s.se.Ensure()
+	s.se.Wait()
+
+	c.Check(s.mockHandler.BeforeCalled, Equals, false)
+	c.Check(s.mockHandler.DoneCalled, Equals, false)
+	c.Check(s.mockHandler.ErrorCalled, Equals, false)
+
+	c.Check(s.command.Calls(), IsNil)
+
+	s.state.Lock()
+	defer s.state.Unlock()
+
+	c.Check(s.task.Kind(), Equals, "run-hook")
+	c.Check(s.task.Status(), Equals, state.DoneStatus)
+	c.Check(s.change.Status(), Equals, state.DoneStatus)
+
+	c.Logf("Task log:\n%s\n", s.task.Log())
+}
+
+func (s *hookManagerSuite) TestHookWithoutHookAlways(c *C) {
+	s.manager.Register(regexp.MustCompile("missing-hook"), func(context *hookstate.Context) hookstate.Handler {
+		return s.mockHandler
+	})
+
+	hooksup := &hookstate.HookSetup{
+		Snap:     "test-snap",
+		Hook:     "missing-hook",
+		Optional: true,
+		Always:   true,
+	}
+	s.state.Lock()
+	s.task.Set("hook-setup", hooksup)
+	s.state.Unlock()
+
+	s.se.Ensure()
+	s.se.Wait()
 
 	c.Check(s.mockHandler.BeforeCalled, Equals, true)
 	c.Check(s.mockHandler.DoneCalled, Equals, true)
@@ -741,8 +868,8 @@ func (s *hookManagerSuite) TestOptionalHookWithMissingHandler(c *C) {
 	s.task.Set("hook-setup", hooksup)
 	s.state.Unlock()
 
-	s.manager.Ensure()
-	s.manager.Wait()
+	s.se.Ensure()
+	s.se.Wait()
 
 	c.Check(s.command.Calls(), IsNil)
 
@@ -770,8 +897,6 @@ func checkTaskLogContains(c *C, task *state.Task, pattern string) {
 
 func (s *hookManagerSuite) TestHookTaskRunsRightSnapCmd(c *C) {
 	coreSnapCmdPath := filepath.Join(dirs.SnapMountDir, "core/12/usr/bin/snap")
-	err := os.MkdirAll(filepath.Dir(coreSnapCmdPath), 0755)
-	c.Assert(err, IsNil)
 	cmd := testutil.MockCommand(c, coreSnapCmdPath, "")
 	defer cmd.Restore()
 
@@ -781,8 +906,8 @@ func (s *hookManagerSuite) TestHookTaskRunsRightSnapCmd(c *C) {
 	})
 	defer r()
 
-	s.manager.Ensure()
-	s.manager.Wait()
+	s.se.Ensure()
+	s.se.Wait()
 
 	s.state.Lock()
 	defer s.state.Unlock()
@@ -817,8 +942,8 @@ func (s *hookManagerSuite) TestHookTaskHandlerReportsErrorIfRequested(c *C) {
 		c, "snap", ">&2 echo 'hook failed at user request'; exit 1")
 	defer cmd.Restore()
 
-	s.manager.Ensure()
-	s.manager.Wait()
+	s.se.Ensure()
+	s.se.Wait()
 
 	s.state.Lock()
 	defer s.state.Unlock()
@@ -848,8 +973,8 @@ func (s *hookManagerSuite) TestHookTaskHandlerReportsErrorDisabled(c *C) {
 		c, "snap", ">&2 echo 'hook failed at user request'; exit 1")
 	defer cmd.Restore()
 
-	s.manager.Ensure()
-	s.manager.Wait()
+	s.se.Ensure()
+	s.se.Wait()
 
 	s.state.Lock()
 	defer s.state.Unlock()
@@ -922,8 +1047,8 @@ func (h *MockConcurrentHandler) Done() error {
 	return nil
 }
 
-func (h *MockConcurrentHandler) Error(err error) error {
-	return nil
+func (h *MockConcurrentHandler) Error(err error) (bool, error) {
+	return false, nil
 }
 
 func NewMockConcurrentHandler(onDone func()) *MockConcurrentHandler {
@@ -972,13 +1097,13 @@ func (s *hookManagerSuite) TestHookTasksForDifferentSnapsRunConcurrently(c *C) {
 		testSnap2HookCalls++
 	})
 	s.manager.Register(regexp.MustCompile("prepare-device"), func(context *hookstate.Context) hookstate.Handler {
-		if context.SnapName() == "test-snap-1" {
+		if context.InstanceName() == "test-snap-1" {
 			return mockHandler1
 		}
-		if context.SnapName() == "test-snap-2" {
+		if context.InstanceName() == "test-snap-2" {
 			return mockHandler2
 		}
-		c.Fatalf("unknown snap: %s", context.SnapName())
+		c.Fatalf("unknown snap: %s", context.InstanceName())
 		return nil
 	})
 
@@ -1018,8 +1143,8 @@ func (s *hookManagerSuite) TestCompatForConfigureSnapd(c *C) {
 	chg.AddTask(task)
 
 	st.Unlock()
-	s.manager.Ensure()
-	s.manager.Wait()
+	s.se.Ensure()
+	s.se.Wait()
 	st.Lock()
 
 	c.Check(chg.Status(), Equals, state.DoneStatus)
@@ -1052,7 +1177,7 @@ func (s *hookManagerSuite) TestGracefullyWaitRunningHooksTimeout(c *C) {
 		return nil
 	})
 
-	s.manager.Ensure()
+	s.se.Ensure()
 	select {
 	case noPending := <-didRun:
 		c.Check(noPending, Equals, false)
@@ -1060,3 +1185,181 @@ func (s *hookManagerSuite) TestGracefullyWaitRunningHooksTimeout(c *C) {
 		c.Fatal("timeout should have expired")
 	}
 }
+
+func (s *hookManagerSuite) TestSnapstateOpConflict(c *C) {
+	s.state.Lock()
+	defer s.state.Unlock()
+	_, err := snapstate.Disable(s.state, "test-snap")
+	c.Assert(err, ErrorMatches, `snap "test-snap" has "kind" change in progress`)
+}
+
+func (s *hookManagerSuite) TestHookHijackingNoConflict(c *C) {
+	s.state.Lock()
+	defer s.state.Unlock()
+
+	s.manager.RegisterHijack("configure", "test-snap", func(ctx *hookstate.Context) error {
+		return nil
+	})
+
+	// no conflict on hijacked hooks
+	_, err := snapstate.Disable(s.state, "test-snap")
+	c.Assert(err, IsNil)
+}
+
+func (s *hookManagerSuite) TestEphemeralRunHook(c *C) {
+	contextData := map[string]interface{}{
+		"key":  "value",
+		"key2": "value2",
+	}
+	s.testEphemeralRunHook(c, contextData)
+}
+
+func (s *hookManagerSuite) TestEphemeralRunHookNoContextData(c *C) {
+	var contextData map[string]interface{} = nil
+	s.testEphemeralRunHook(c, contextData)
+}
+
+func (s *hookManagerSuite) testEphemeralRunHook(c *C, contextData map[string]interface{}) {
+	var hookInvokeCalled []string
+	hookInvoke := func(ctx *hookstate.Context, tomb *tomb.Tomb) ([]byte, error) {
+		c.Check(ctx.HookName(), Equals, "configure")
+		hookInvokeCalled = append(hookInvokeCalled, ctx.HookName())
+
+		// check that context data was set correctly
+		var s string
+		ctx.Lock()
+		defer ctx.Unlock()
+		for k, v := range contextData {
+			ctx.Get(k, &s)
+			c.Check(s, Equals, v)
+		}
+		ctx.Set("key-set-from-hook", "value-set-from-hook")
+
+		return []byte("some output"), nil
+	}
+	restore := hookstate.MockRunHook(hookInvoke)
+	defer restore()
+
+	hooksup := &hookstate.HookSetup{
+		Snap:     "test-snap",
+		Revision: snap.R(1),
+		Hook:     "configure",
+	}
+	context, err := s.manager.EphemeralRunHook(context.Background(), hooksup, contextData)
+	c.Assert(err, IsNil)
+	c.Check(hookInvokeCalled, DeepEquals, []string{"configure"})
+
+	var value string
+	context.Lock()
+	context.Get("key-set-from-hook", &value)
+	context.Unlock()
+	c.Check(value, Equals, "value-set-from-hook")
+}
+
+func (s *hookManagerSuite) TestEphemeralRunHookNoSnap(c *C) {
+	hookInvoke := func(ctx *hookstate.Context, tomb *tomb.Tomb) ([]byte, error) {
+		c.Fatalf("hook should not be invoked in this test")
+		return nil, nil
+	}
+	restore := hookstate.MockRunHook(hookInvoke)
+	defer restore()
+
+	hooksup := &hookstate.HookSetup{
+		Snap:     "not-installed-snap",
+		Revision: snap.R(1),
+		Hook:     "configure",
+	}
+	contextData := map[string]interface{}{
+		"key": "value",
+	}
+	_, err := s.manager.EphemeralRunHook(context.Background(), hooksup, contextData)
+	c.Assert(err, ErrorMatches, `cannot run ephemeral hook "configure" for snap "not-installed-snap": no state entry for key`)
+}
+
+func (s *hookManagerSuite) TestEphemeralRunHookContextCanCancel(c *C) {
+	tombDying := 0
+	hookRunning := make(chan struct{})
+
+	hookInvoke := func(_ *hookstate.Context, tomb *tomb.Tomb) ([]byte, error) {
+		close(hookRunning)
+
+		select {
+		case <-tomb.Dying():
+			tombDying++
+		case <-time.After(10 * time.Second):
+			c.Fatalf("hook not canceled after 10s")
+		}
+		return nil, nil
+	}
+	restore := hookstate.MockRunHook(hookInvoke)
+	defer restore()
+
+	hooksup := &hookstate.HookSetup{
+		Snap:     "test-snap",
+		Revision: snap.R(1),
+		Hook:     "configure",
+	}
+
+	ctx, cancelFunc := context.WithCancel(context.Background())
+	go func() {
+		<-hookRunning
+		cancelFunc()
+	}()
+	_, err := s.manager.EphemeralRunHook(ctx, hooksup, nil)
+	c.Assert(err, IsNil)
+	c.Check(tombDying, Equals, 1)
+}
+
+type parallelInstancesHookManagerSuite struct {
+	baseHookManagerSuite
+}
+
+var _ = Suite(&parallelInstancesHookManagerSuite{})
+
+func (s *parallelInstancesHookManagerSuite) SetUpTest(c *C) {
+	s.commonSetUpTest(c)
+	s.setUpSnap(c, "test-snap_instance", snapYaml)
+}
+
+func (s *parallelInstancesHookManagerSuite) TearDownTest(c *C) {
+	s.commonTearDownTest(c)
+}
+
+func (s *parallelInstancesHookManagerSuite) TestHookTaskEnsureHookRan(c *C) {
+	didRun := make(chan bool)
+	s.mockHandler.BeforeCallback = func() {
+		c.Check(s.manager.NumRunningHooks(), Equals, 1)
+		go func() {
+			didRun <- s.manager.GracefullyWaitRunningHooks()
+		}()
+	}
+	s.se.Ensure()
+	select {
+	case ok := <-didRun:
+		c.Check(ok, Equals, true)
+	case <-time.After(5 * time.Second):
+		c.Fatal("hook run should have been done by now")
+	}
+	s.se.Wait()
+
+	s.state.Lock()
+	defer s.state.Unlock()
+
+	c.Check(s.context.InstanceName(), Equals, "test-snap_instance")
+	c.Check(s.context.SnapRevision(), Equals, snap.R(1))
+	c.Check(s.context.HookName(), Equals, "configure")
+
+	c.Check(s.command.Calls(), DeepEquals, [][]string{{
+		"snap", "run", "--hook", "configure", "-r", "1", "test-snap_instance",
+	}})
+
+	c.Check(s.mockHandler.BeforeCalled, Equals, true)
+	c.Check(s.mockHandler.DoneCalled, Equals, true)
+	c.Check(s.mockHandler.ErrorCalled, Equals, false)
+
+	c.Check(s.task.Kind(), Equals, "run-hook")
+	c.Check(s.task.Status(), Equals, state.DoneStatus)
+	c.Check(s.change.Status(), Equals, state.DoneStatus)
+
+	c.Check(s.manager.NumRunningHooks(), Equals, 0)
+}
