@@ -435,10 +435,21 @@ distro_install_build_snapd(){
                 # to create sockets with incorrect context, this installation of
                 # socket activated snaps fails, see:
                 # https://bugzilla.redhat.com/show_bug.cgi?id=1660141
+                # https://bugzilla.redhat.com/show_bug.cgi?id=1197886
                 # https://github.com/systemd/systemd/issues/9997
                 systemctl daemon-reexec
                 ;;
         esac
+        case "$SPREAD_SYSTEM" in
+            fedora-*)
+                # the problem with SELinux policy also affects the user instance
+                # in 248, see:
+                # https://bugzilla.redhat.com/show_bug.cgi?id=1960576
+                # note, this fixes it for the root user only, the test user
+                # session is created dynamically as needed
+                systemctl --user daemon-reexec
+                ;;
+        esac
 
         if os.query is-arch-linux; then
             # Arch policy does not allow calling daemon-reloads in package
@@ -451,7 +462,7 @@ distro_install_build_snapd(){
             fi
         fi
 
-        if [[ "$SPREAD_SYSTEM" == opensuse-tumbleweed-* ]]; then
+        if os.query is-opensuse-tumbleweed; then
             # Package installation applies vendor presets only, which leaves
             # snapd.apparmor disabled.
             systemctl enable --now snapd.apparmor.service
@@ -485,6 +496,7 @@ distro_get_package_extension() {
 
 pkg_dependencies_ubuntu_generic(){
     echo "
+        python3
         autoconf
         automake
         autotools-dev
@@ -539,17 +551,6 @@ pkg_dependencies_ubuntu_classic(){
         ubuntu-14.04-*)
                 pkg_linux_image_extra
             ;;
-        ubuntu-16.04-32)
-            echo "
-                dbus-user-session
-                gccgo-6
-                evolution-data-server
-                fwupd
-                gnome-online-accounts
-                packagekit
-                "
-                pkg_linux_image_extra
-            ;;
         ubuntu-16.04-64)
             echo "
                 dbus-user-session
@@ -566,10 +567,22 @@ pkg_dependencies_ubuntu_classic(){
                 "
                 pkg_linux_image_extra
             ;;
+        ubuntu-18.04-32)
+            echo "
+                dbus-user-session
+                gccgo-6
+                evolution-data-server
+                fwupd
+                gnome-online-accounts
+                packagekit
+                "
+                pkg_linux_image_extra
+            ;;
         ubuntu-18.04-64)
             echo "
                 dbus-user-session
                 gccgo-8
+                gperf
                 evolution-data-server
                 fwupd
                 packagekit
@@ -579,21 +592,22 @@ pkg_dependencies_ubuntu_classic(){
         ubuntu-20.04-64)
             echo "
                 evolution-data-server
+                fwupd
                 gccgo-9
+                libvirt-daemon-system
                 packagekit
+                qemu-kvm
                 qemu-utils
                 shellcheck
                 "
             ;;
-        ubuntu-20.10-64)
-            echo "
-                qemu-utils
-                "
-            ;;
-        ubuntu-21.04-64)
+        ubuntu-21.04-64|ubuntu-21.10-64*)
+            # bpftool is part of linux-tools package
             echo "
                 dbus-user-session
+                fwupd
                 golang
+                linux-tools-$(uname -r)
                 qemu-utils
                 "
             ;;
@@ -620,6 +634,13 @@ pkg_dependencies_ubuntu_classic(){
                 "
             ;;
     esac
+    case "$SPREAD_SYSTEM" in
+        debian-11-*|debian-sid-*)
+            echo "
+                 bpftool
+                 "
+            ;;
+    esac
 }
 
 pkg_linux_image_extra (){
@@ -642,8 +663,10 @@ pkg_dependencies_ubuntu_core(){
         pkg_linux_image_extra
 }
 
-pkg_dependencies_fedora(){
+pkg_dependencies_fedora_centos_common(){
     echo "
+        python3
+        bpftool
         clang
         curl
         dbus-x11
@@ -674,8 +697,15 @@ pkg_dependencies_fedora(){
         "
 }
 
+pkg_dependencies_fedora(){
+    echo "
+         libcap-static
+        "
+}
+
 pkg_dependencies_amazon(){
     echo "
+        python3
         curl
         dbus-x11
         expect
@@ -686,6 +716,7 @@ pkg_dependencies_amazon(){
         grub2-tools
         jq
         iptables-services
+        libcap-static
         man
         nc
         net-tools
@@ -703,21 +734,30 @@ pkg_dependencies_amazon(){
 
 pkg_dependencies_opensuse(){
     echo "
+        python3
         apparmor-profiles
+        audit
+        bash-completion
+        bpftool
         clang
         curl
+        dbus-1-python3
         evolution-data-server
         expect
         fontconfig
         fwupd
         git
         golang-packaging
+        iptables
         jq
         lsb-release
         man
+        man-pages
         nfs-kernel-server
+        nss-mdns
         PackageKit
         python3-yaml
+        strace
         netcat-openbsd
         osc
         udisks2
@@ -731,8 +771,10 @@ pkg_dependencies_opensuse(){
 
 pkg_dependencies_arch(){
     echo "
+    apparmor
     base-devel
     bash-completion
+    bpf
     clang
     curl
     evolution-data-server
@@ -764,7 +806,6 @@ pkg_dependencies_arch(){
     xdg-user-dirs
     xdg-utils
     xfsprogs
-    apparmor
     zsh
     "
 }
@@ -782,7 +823,11 @@ pkg_dependencies(){
         amazon-*|centos-7-*)
             pkg_dependencies_amazon
             ;;
-        fedora-*|centos-*)
+        centos-*)
+            pkg_dependencies_fedora_centos_common
+            ;;
+        fedora-*)
+            pkg_dependencies_fedora_centos_common
             pkg_dependencies_fedora
             ;;
         opensuse-*)
