@@ -44,7 +44,7 @@ var snapdServiceStopTimeout = time.Duration(timeout.DefaultTimeout)
 var execStartRe = regexp.MustCompile(`(?m)^ExecStart=(/usr/bin/snap\s+.*|/usr/lib/snapd/.*)$`)
 
 // snapdToolingMountUnit is the name of the mount unit that makes the
-const snapdToolingMountUnit = "usr-lib-snapd.mount"
+const SnapdToolingMountUnit = "usr-lib-snapd.mount"
 
 func snapdSkipStart(content []byte) bool {
 	return bytes.Contains(content, []byte("X-Snapd-Snap: do-not-start"))
@@ -65,8 +65,12 @@ func snapdUnitSkipStart(unitPath string) (skip bool, err error) {
 }
 
 func writeSnapdToolingMountUnit(sysd systemd.Systemd, prefix string) error {
+
+	// TODO: the following comment is wrong, we don't need RequiredBy=snapd here?
+
 	// Not using AddMountUnitFile() because we need
 	// "RequiredBy=snapd.service"
+
 	content := []byte(fmt.Sprintf(`[Unit]
 Description=Make the snapd snap tooling available for the system
 Before=snapd.service
@@ -80,7 +84,7 @@ Options=bind
 [Install]
 WantedBy=snapd.service
 `, prefix))
-	fullPath := filepath.Join(dirs.SnapServicesDir, snapdToolingMountUnit)
+	fullPath := filepath.Join(dirs.SnapServicesDir, SnapdToolingMountUnit)
 
 	err := osutil.EnsureFileState(fullPath,
 		&osutil.MemoryFileState{
@@ -97,11 +101,14 @@ WantedBy=snapd.service
 	if err := sysd.DaemonReload(); err != nil {
 		return err
 	}
-	if err := sysd.Enable(snapdToolingMountUnit); err != nil {
+	if err := sysd.Enable(SnapdToolingMountUnit); err != nil {
 		return err
 	}
 
-	if err := sysd.Restart(snapdToolingMountUnit, 5*time.Second); err != nil {
+	// meh this is killing snap services that use Requires=<this-unit> because
+	// it doesn't use verbatim systemctl restart, it instead does it with
+	// a systemctl stop and then a systemctl start, which triggers LP #1924805
+	if err := sysd.Restart(SnapdToolingMountUnit, 5*time.Second); err != nil {
 		return err
 	}
 
@@ -292,6 +299,10 @@ func AddSnapdSnapServices(s *snap.Info, inter interacter) error {
 		return err
 	}
 
+	if err := writeSnapdDbusActivationOnCore(s); err != nil {
+		return err
+	}
+
 	return nil
 }
 
@@ -563,7 +574,7 @@ func writeSnapdDbusConfigOnCore(s *snap.Info) error {
 	return nil
 }
 
-func undoSnapdDbusConfigOnCore(s *snap.Info) error {
+func undoSnapdDbusConfigOnCore() error {
 	_, _, err := osutil.EnsureDirState(dirs.SnapDBusSystemPolicyDir, "snapd.*.conf", nil)
 	if err != nil {
 		return err
@@ -572,6 +583,33 @@ func undoSnapdDbusConfigOnCore(s *snap.Info) error {
 	return err
 }
 
+var dbusSessionServices = []string{
+	"io.snapcraft.Launcher.service",
+	"io.snapcraft.Settings.service",
+	"io.snapcraft.SessionAgent.service",
+}
+
+func writeSnapdDbusActivationOnCore(s *snap.Info) error {
+	if err := os.MkdirAll(dirs.SnapDBusSessionServicesDir, 0755); err != nil {
+		return err
+	}
+
+	content := make(map[string]osutil.FileState, len(dbusSessionServices)+1)
+	for _, service := range dbusSessionServices {
+		content[service] = &osutil.FileReference{
+			Path: filepath.Join(s.MountDir(), "usr/share/dbus-1/services", service),
+		}
+	}
+
+	_, _, err := osutil.EnsureDirStateGlobs(dirs.SnapDBusSessionServicesDir, dbusSessionServices, content)
+	return err
+}
+
+func undoSnapdDbusActivationOnCore() error {
+	_, _, err := osutil.EnsureDirStateGlobs(dirs.SnapDBusSessionServicesDir, dbusSessionServices, nil)
+	return err
+}
+
 // RemoveSnapdSnapServicesOnCore removes the snapd services generated by a prior
 // call to AddSnapdSnapServices. The core snap is used as the reference for
 // restoring the system state, making this undo helper suitable for use when
@@ -588,7 +626,10 @@ func RemoveSnapdSnapServicesOnCore(s *snap.Info, inter interacter) error {
 
 	sysd := systemd.NewUnderRoot(dirs.GlobalRootDir, systemd.SystemMode, inter)
 
-	if err := undoSnapdDbusConfigOnCore(s); err != nil {
+	if err := undoSnapdDbusActivationOnCore(); err != nil {
+		return err
+	}
+	if err := undoSnapdDbusConfigOnCore(); err != nil {
 		return err
 	}
 	if err := undoSnapdServicesOnCore(s, sysd); err != nil {
